@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: BuildMessage.py,v 1.47 2003/06/11 19:05:35 nickm Exp $
+# $Id: BuildMessage.py,v 1.65 2003/12/08 02:22:56 nickm Exp $
 
 """mixminion.BuildMessage
 
@@ -11,21 +11,88 @@ import sys
 import types
 
 import mixminion.Crypto as Crypto
+import mixminion.Fragments
 from mixminion.Packet import *
 from mixminion.Common import MixError, MixFatalError, LOG, UIError
+import mixminion.Packet
+import mixminion._minionlib
 
 if sys.version_info[:3] < (2,2,0):
     import mixminion._zlibutil as zlibutil
 
-__all__ = ['buildForwardMessage', 'buildEncryptedMessage', 'buildReplyMessage',
-           'buildReplyBlock', 'checkPathLength', 'decodePayload' ]
+__all__ = ['buildForwardPacket', 'buildEncryptedForwardPacket',
+           'buildReplyPacket', 'buildReplyBlock', 'checkPathLength',
+           'encodeMessage', 'decodePayload' ]
 
-def buildForwardMessage(payload, exitType, exitInfo, path1, path2,
+def encodeMessage(message, overhead, uncompressedFragmentPrefix="",
+                  paddingPRNG=None):
+    """Given a message, compress it, fragment it into individual payloads,
+       and add extra fields (size, hash, etc) as appropriate.  Return a list
+       of strings, each of which is a message payload suitable for use in
+       build*Message.
+
+              payload: the initial payload
+              overhead: number of bytes to omit from each payload,
+                        given the type ofthe message encoding.
+                        (0 or ENC_FWD_OVERHEAD)
+              uncompressedFragmentPrefix: If we fragment the message,
+                we add this string to the message after compression but
+                before whitening and fragmentation.
+              paddingPRNG: generator for padding.
+
+       Note: If multiple strings are returned, be sure to shuffle them
+       before transmitting them to the network.
+    """
+    assert overhead in (0, ENC_FWD_OVERHEAD)
+    if paddingPRNG is None:
+        paddingPRNG = Crypto.getCommonPRNG()
+    origLength = len(message)
+    payload = compressData(message)
+    length = len(payload)
+
+    if length > 1024 and length*20 <= origLength:
+        LOG.warn("Message is very compressible and will look like a zlib bomb")
+
+    paddingLen = PAYLOAD_LEN - SINGLETON_PAYLOAD_OVERHEAD - overhead - length
+
+    # If the compressed payload fits in 28K, we're set.
+    if paddingLen >= 0:
+        # We pad the payload, and construct a new SingletonPayload,
+        # including this payload's size and checksum.
+        payload += paddingPRNG.getBytes(paddingLen)
+        p = SingletonPayload(length, None, payload)
+        p.computeHash()
+        return [ p.pack() ]
+
+    # Okay, we need fo fragment the message.  First, add the prefix if needed.
+    if uncompressedFragmentPrefix:
+        payload = uncompressedFragmentPrefix+payload
+    # Now generate a message ID
+    messageid = Crypto.getCommonPRNG().getBytes(FRAGMENT_MESSAGEID_LEN)
+    # Figure out how many chunks to divide it into...
+    p = mixminion.Fragments.FragmentationParams(len(payload), overhead)
+    # ... fragment the payload into chunks...
+    rawFragments = p.getFragments(payload)
+    fragments = []
+    # ... and annotate each chunk with appropriate payload header info.
+    for i in xrange(len(rawFragments)):
+        pyld = FragmentPayload(i, None, messageid, p.length, rawFragments[i])
+        pyld.computeHash()
+        fragments.append(pyld.pack())
+        rawFragments[i] = None
+    return fragments
+
+def buildRandomPayload(paddingPRNG=None):
+    """Return a new random payload, suitable for use in a DROP packet."""
+    if not paddingPRNG:
+        paddingPRNG = Crypto.getCommonPRNG()
+    return paddingPRNG.getBytes(PAYLOAD_LEN)
+
+def buildForwardPacket(payload, exitType, exitInfo, path1, path2,
                         paddingPRNG=None):
     """Construct a forward message.
-            payload: The payload to deliver.  Must compress to under 28K-22b.
-                  If it does not, MixError is raised.  If the payload is
-                  None, 28K of random data is sent.
+            payload: The payload to deliver.  Must be exactly 28K.  If the
+                  payload is None, 28K of random data is sent.
             exitType: The routing type for the final node. (2 bytes, >=0x100)
             exitInfo: The routing info for the final node, not including tag.
             path1: Sequence of ServerInfo objects for the first leg of the path
@@ -43,17 +110,11 @@ def buildForwardMessage(payload, exitType, exitInfo, path1, path2,
         raise MixError("Second leg of path is empty")
 
     suppressTag = 0
-    if exitType == DROP_TYPE:
+    #XXXX refactor _TYPES_WITHOUT_TAGS
+    if exitType == DROP_TYPE or mixminion.Packet._TYPES_WITHOUT_TAGS.get(exitType):
         suppressTag = 1
-        payload = None
 
-    # Compress, pad, and checksum the payload.
-    if payload is not None:
-        payload = _encodePayload(payload, 0, paddingPRNG)
-        LOG.debug("Encoding forward message for %s-byte payload",len(payload))
-    else:
-        payload = paddingPRNG.getBytes(PAYLOAD_LEN)
-        LOG.debug("Generating DROP message with %s bytes", PAYLOAD_LEN)
+    assert len(payload) == PAYLOAD_LEN
 
     LOG.debug("  Using path %s:%s",
                    ",".join([s.getNickname() for s in path1]),
@@ -64,15 +125,15 @@ def buildForwardMessage(payload, exitType, exitInfo, path1, path2,
     if not suppressTag:
         tag = _getRandomTag(paddingPRNG)
         exitInfo = tag + exitInfo
-    return _buildMessage(payload, exitType, exitInfo, path1, path2,
-                         paddingPRNG)
+    return _buildPacket(payload, exitType, exitInfo, path1, path2,
+                        paddingPRNG)
 
-def buildEncryptedForwardMessage(payload, exitType, exitInfo, path1, path2,
+
+def buildEncryptedForwardPacket(payload, exitType, exitInfo, path1, path2,
                                  key, paddingPRNG=None, secretRNG=None):
     """Construct a forward message encrypted with the public key of a
        given user.
-            payload: The payload to deliver.  Must compress to under 28K-60b.
-                  If it does not, MixError is raised.
+            payload: The payload to deliver.  Must be 28K-42b long.
             exitType: The routing type for the final node. (2 bytes, >=0x100)
             exitInfo: The routing info for the final node, not including tag.
             path1: Sequence of ServerInfo objects for the first leg of the path
@@ -92,10 +153,9 @@ def buildEncryptedForwardMessage(payload, exitType, exitInfo, path1, path2,
                    [s.getNickname() for s in path2])
     LOG.debug("  Delivering to %04x:%r", exitType, exitInfo)
 
-    # Compress, pad, and checksum the payload.
     # (For encrypted-forward messages, we have overhead for OAEP padding
     #   and the session  key, but we save 20 bytes by spilling into the tag.)
-    payload = _encodePayload(payload, ENC_FWD_OVERHEAD, paddingPRNG)
+    assert len(payload) == PAYLOAD_LEN - ENC_FWD_OVERHEAD
 
     # Generate the session key, and prepend it to the payload.
     sessionKey = secretRNG.getBytes(SECRET_LEN)
@@ -127,11 +187,12 @@ def buildEncryptedForwardMessage(payload, exitType, exitInfo, path1, path2,
     assert len(payload) == 28*1024
 
     # And now, we can finally build the message.
-    return _buildMessage(payload, exitType, exitInfo, path1, path2,paddingPRNG)
+    return _buildPacket(payload, exitType, exitInfo, path1, path2,paddingPRNG)
 
-def buildReplyMessage(payload, path1, replyBlock, paddingPRNG=None):
+def buildReplyPacket(payload, path1, replyBlock, paddingPRNG=None):
     """Build a message using a reply block.  'path1' is a sequence of
-       ServerInfo for the nodes on the first leg of the path.
+       ServerInfo for the nodes on the first leg of the path.  'payload'
+       must be exactly 28K long.
     """
     if paddingPRNG is None:
         paddingPRNG = Crypto.getCommonPRNG()
@@ -140,8 +201,7 @@ def buildReplyMessage(payload, path1, replyBlock, paddingPRNG=None):
                    len(payload))
     LOG.debug("  Using path %s/??",[s.getNickname() for s in path1])
 
-    # Compress, pad, and checksum the payload.
-    payload = _encodePayload(payload, 0, paddingPRNG)
+    assert len(payload) == PAYLOAD_LEN
 
     # Encrypt the payload so that it won't appear as plaintext to the
     #  crossover note.  (We use 'decrypt' so that the message recipient can
@@ -150,7 +210,7 @@ def buildReplyMessage(payload, path1, replyBlock, paddingPRNG=None):
                          Crypto.PAYLOAD_ENCRYPT_MODE)
     payload = Crypto.lioness_decrypt(payload, k)
 
-    return _buildMessage(payload, None, None,
+    return _buildPacket(payload, None, None,
                          path1=path1, path2=replyBlock)
 
 def _buildReplyBlockImpl(path, exitType, exitInfo, expiryTime=0,
@@ -198,13 +258,15 @@ def _buildReplyBlockImpl(path, exitType, exitInfo, expiryTime=0,
     header = _buildHeader(path, headerSecrets, exitType, tag+exitInfo,
                           paddingPRNG=Crypto.getCommonPRNG())
 
+    # XXXX007 switch to Host info.  We need to use IPV4 for reply blocks
+    # XXXX007 for now, since we don't know which servers will support HOST.
     return ReplyBlock(header, expiryTime,
-                      SWAP_FWD_TYPE,
-                      path[0].getRoutingInfo().pack(), sharedKey), secrets, tag
+                      SWAP_FWD_IPV4_TYPE,
+                      path[0].getIPV4Info().pack(), sharedKey), secrets, tag
 
 # Maybe we shouldn't even allow this to be called with userKey==None.
 def buildReplyBlock(path, exitType, exitInfo, userKey,
-                    expiryTime=0, secretRNG=None):
+                    expiryTime=None, secretRNG=None):
     """Construct a 'state-carrying' reply block that does not require the
        reply-message recipient to remember a list of secrets.
        Instead, all secrets are generated from an AES counter-mode
@@ -247,11 +309,14 @@ def checkPathLength(path1, path2, exitType, exitInfo, explicitSwap=0):
     err = 0 # 0: no error. 1: 1st leg too big. 2: 1st leg okay, 2nd too big.
     if path1 is not None:
         try:
-            _getRouting(path1, SWAP_FWD_TYPE, path2[0].getRoutingInfo().pack())
+            rt,ri = path1[-1].getRoutingFor(path2[0],swap=1)
+            _getRouting(path1, rt, ri)
         except MixError:
             err = 1
-    # Add tag as needed to last exitinfo.
-    if exitType != DROP_TYPE and exitInfo is not None:
+    # Add a dummy tag as needed to last exitinfo.
+    if (exitType != DROP_TYPE
+        and not mixminion.Packet._TYPES_WITHOUT_TAGS.get(exitType)
+        and exitInfo is not None):
         exitInfo += "X"*20
     else:
         exitInfo = ""
@@ -263,51 +328,59 @@ def checkPathLength(path1, path2, exitType, exitInfo, explicitSwap=0):
     if err and not explicitSwap:
         raise UIError("Address and path will not fit in one header")
     elif err:
-        raise UIError("Address and %s leg of path will not fit in one header",
-                      ["first", "second"][err-1])
-    
+        raise UIError("Address and %s leg of path will not fit in one header"
+                      % ["first", "second"][err-1])
+
 #----------------------------------------------------------------------
 # MESSAGE DECODING
-
-def decodePayload(payload, tag, key=None,
-                  userKeys=None):
-    """Given a 28K payload and a 20-byte decoding tag, attempt to decode and
-       decompress the original message.
+def decodePayload(payload, tag, key=None, userKeys=()):
+    """Given a 28K payload and a 20-byte decoding tag, attempt to decode the
+       original message.  Returns either a SingletonPayload instance, a
+       FragmentPayload instance, or None.
 
            key: an RSA key to decode encrypted forward messages, or None
-           userKeys: a map from identity names to keys for reply blocks,
-                or None.
+           userKeys: a sequence of (name,key) tuples maping identity names to
+                SURB keys. For backward compatibility, 'userKeys' may also be
+                None (no SURBs known), a dict (from name to key), or a single
+                key (implied identity is "").
 
        If we can successfully decrypt the payload, we return it.  If we
        might be able to decrypt the payload given more/different keys,
        we return None.  If the payload is corrupt, we raise MixError.
     """
     if userKeys is None:
-        userKeys = {}
+        userKeys = []
     elif type(userKeys) is types.StringType:
-        userKeys = { "" : userKeys }
+        userKeys = [ ("", userKeys) ]
+    elif type(userKeys) is types.DictType:
+        userKeys = userKeys.items()
 
-    if len(payload) != PAYLOAD_LEN or len(tag) != TAG_LEN:
-        raise MixError("Wrong payload or tag length")
+    if len(payload) != PAYLOAD_LEN:
+        raise MixError("Wrong payload length")
+
+    if len(tag) not in (0, TAG_LEN):
+        raise MixError("Wrong tag length: %s"%len(tag))
 
     # If the payload already contains a valid checksum, it's a forward
     # message.
     if _checkPayload(payload):
-        return _decodeForwardPayload(payload)
+        return parsePayload(payload)
+
+    if not tag:
+        return None
 
     # If H(tag|userKey|"Validate") ends with 0, then the message _might_
     # be a reply message using H(tag|userKey|"Generate") as the seed for
     # its master secrets.  (There's a 1-in-256 chance that it isn't.)
-    if userKeys:
-        for name,userKey in userKeys.items():
-            if Crypto.sha1(tag+userKey+"Validate")[-1] == '\x00':
-                try:
-                    p = _decodeStatelessReplyPayload(payload, tag, userKey)
-                    if name:
-                        LOG.info("Decoded reply message to identity %r", name)
-                    return p
-                except MixError:
-                    pass
+    for name,userKey in userKeys:
+        if Crypto.sha1(tag+userKey+"Validate")[-1] == '\x00':
+            try:
+                p = _decodeStatelessReplyPayload(payload, tag, userKey)
+                if name:
+                    LOG.info("Decoded reply message to identity %r", name)
+                return p
+            except MixError:
+                pass
 
     # If we have an RSA key, and none of the above steps get us a good
     # payload, then we may as well try to decrypt the start of tag+key with
@@ -322,7 +395,10 @@ def decodePayload(payload, tag, key=None,
 def _decodeForwardPayload(payload):
     """Helper function: decode a non-encrypted forward payload. Return values
        are the same as decodePayload."""
-    return _decodePayloadImpl(payload)
+    if not _checkPayload(payload):
+        raise MixError("Hash doesn't match")
+
+    return parsePayload(payload)
 
 def _decodeEncryptedForwardPayload(payload, tag, key):
     """Helper function: decode an encrypted forward payload.  Return values
@@ -348,11 +424,14 @@ def _decodeEncryptedForwardPayload(payload, tag, key):
     rest = rsaPart[SECRET_LEN:] + Crypto.lioness_decrypt(rest, k)
 
     # ... and then, check the checksum and continue.
-    return _decodePayloadImpl(rest)
+    if not _checkPayload(rest):
+        raise MixError("Invalid checksum on encrypted forward payload")
+
+    return parsePayload(rest)
 
 def _decodeReplyPayload(payload, secrets, check=0):
     """Helper function: decode a reply payload, given a known list of packet
-         master secrets. If 'check' is true, then 'secerets' may be overlong.
+         master secrets. If 'check' is true, then 'secrets' may be overlong.
          Return values are the same as decodePayload.
       [secrets must be in _reverse_ order]
     """
@@ -362,10 +441,14 @@ def _decodeReplyPayload(payload, secrets, check=0):
         k = Crypto.Keyset(sec).getLionessKeys(Crypto.PAYLOAD_ENCRYPT_MODE)
         payload = Crypto.lioness_encrypt(payload, k)
         if check and _checkPayload(payload):
-            break
+            return parsePayload(payload)
 
-    # ... and then, check the checksum and continue.
-    return _decodePayloadImpl(payload)
+    # If 'check' is false, then we might still have a good payload.  If
+    # 'check' is true, we don't.
+    if check or not _checkPayload(payload):
+        raise MixError("Invalid checksum on reply payload")
+
+    return parsePayload(payload)
 
 def _decodeStatelessReplyPayload(payload, tag, userKey):
     """Decode a (state-carrying) reply payload."""
@@ -378,8 +461,8 @@ def _decodeStatelessReplyPayload(payload, tag, userKey):
     return _decodeReplyPayload(payload, secrets, check=1)
 
 #----------------------------------------------------------------------
-def _buildMessage(payload, exitType, exitInfo,
-                  path1, path2, paddingPRNG=None, paranoia=0):
+def _buildPacket(payload, exitType, exitInfo,
+                 path1, path2, paddingPRNG=None, paranoia=0):
     """Helper method to create a message.
 
     The following fields must be set:
@@ -411,7 +494,7 @@ def _buildMessage(payload, exitType, exitInfo,
         reply = path2
         path2 = None
     else:
-        if len(exitInfo) < TAG_LEN and exitType != DROP_TYPE:
+        if len(exitInfo) < TAG_LEN and exitType != DROP_TYPE and not mixminion.Packet._TYPES_WITHOUT_TAGS.get(exitType):
             raise MixError("Implausibly short exit info: %r"%exitInfo)
         if exitType < MIN_EXIT_TYPE and exitType != DROP_TYPE:
             raise MixError("Invalid exit type: %4x"%exitType)
@@ -433,8 +516,7 @@ def _buildMessage(payload, exitType, exitInfo,
         path1exittype = reply.routingType
         path1exitinfo = reply.routingInfo
     else:
-        path1exittype = SWAP_FWD_TYPE
-        path1exitinfo = path2[0].getRoutingInfo().pack()
+        path1exittype, path1exitinfo = path1[-1].getRoutingFor(path2[0],swap=1)
 
     # Generate secrets for path1.
     secrets1 = [ secretRNG.getBytes(SECRET_LEN) for _ in path1 ]
@@ -546,7 +628,7 @@ def _buildHeader(path,secrets,exitType,exitInfo,paddingPRNG):
         rsaPart = subhead.pack() + underflow
         esh = Crypto.pk_encrypt(rsaPart, pubkey)
 
-        # Concatentate the asymmetric and symmetric parts, to get the next
+        # Concatenate the asymmetric and symmetric parts, to get the next
         # header.
         header = esh + header
 
@@ -594,71 +676,22 @@ def _constructMessage(secrets1, secrets2, header1, header2, payload):
         header2 = Crypto.lioness_encrypt(header2,hkey)
         payload = Crypto.lioness_encrypt(payload,pkey)
 
-    return Message(header1, header2, payload).pack()
+    return Packet(header1, header2, payload).pack()
 
 #----------------------------------------------------------------------
 # Payload-related helpers
 
-def _encodePayload(payload, overhead, paddingPRNG):
-    """Helper: compress a payload, pad it, and add extra fields (size and hash)
-              payload: the initial payload
-              overhead: number of bytes to omit from result
-                        (0 or ENC_FWD_OVERHEAD)
-              paddingPRNG: generator for padding.
-
-       BUG: This should eventually support K-of-N.
-    """
-    assert overhead in (0, ENC_FWD_OVERHEAD)
-
-    # Compress the data, and figure out how much padding we'll need.
-    origLength = len(payload)
-    payload = compressData(payload)
-    length = len(payload)
-
-    if length > 1024 and length*20 <= origLength:
-        LOG.warn("Message is very compressible and will look like a zlib bomb")
-
-    paddingLen = PAYLOAD_LEN - SINGLETON_PAYLOAD_OVERHEAD - overhead - length
-
-    # If the compressed payload doesn't fit in 28K, then we need to bail out.
-    if paddingLen < 0:
-        raise MixError("Payload too long for singleton message")
-
-    # Otherwise, we pad the payload, and construct a new SingletonPayload,
-    # including this payload's size and checksum.
-    payload += paddingPRNG.getBytes(paddingLen)
-    return SingletonPayload(length, Crypto.sha1(payload), payload).pack()
-
 def _getRandomTag(rng):
     "Helper: Return a 20-byte string with the MSB of byte 0 set to 0."
     b = ord(rng.getBytes(1)) & 0x7f
     return chr(b) + rng.getBytes(TAG_LEN-1)
 
-def _decodePayloadImpl(payload):
-    """Helper: try to decode an encoded payload: checks only encoding,
-       not encryption."""
-    # Is the hash ok?
-    if not _checkPayload(payload):
-        raise MixError("Hash doesn't match")
-
-    # Parse the payload into its size, checksum, and body.
-    payload = parsePayload(payload)
-
-    if not payload.isSingleton():
-        raise MixError("Message fragments not yet supported")
-
-    # Uncompress the body.
-    contents = payload.getContents()
-    # If the payload would expand to be more than 20K long, and the
-    # compression factor is greater than 20, we warn of a possible zlib
-    # bomb.
-    maxLen = max(20*1024, 20*len(contents))
-
-    return uncompressData(contents, maxLength=maxLen)
-
 def _checkPayload(payload):
     'Return true iff the hash on the given payload seems valid'
-    return payload[2:22] == Crypto.sha1(payload[22:])
+    if ord(payload[0]) & 0x80:
+        return payload[3:23] == Crypto.sha1(payload[23:])
+    else:
+        return payload[2:22] == Crypto.sha1(payload[22:])
 
 def _getRouting(path, exitType, exitInfo):
     """Given a list of ServerInfo, and a final exitType and exitInfo,
@@ -671,8 +704,9 @@ def _getRouting(path, exitType, exitInfo):
        Raises MixError if the routing info is too big to fit into a single
        header. """
     # Construct a list 'routing' of exitType, exitInfo.
-    routing = [ (FWD_TYPE, node.getRoutingInfo().pack()) for
-                node in path[1:] ]
+    routing = []
+    for i in xrange(len(path)-1):
+        routing.append(path[i].getRoutingFor(path[i+1],swap=0))
     routing.append((exitType, exitInfo))
 
     # sizes[i] is number of bytes added to header for subheader i.
@@ -690,3 +724,4 @@ def _getRouting(path, exitType, exitInfo):
         raise MixError("Routing info won't fit in header")
 
     return routing, sizes, totalSize
+
