@@ -1,14 +1,16 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: Common.py,v 1.89 2003/06/06 06:04:57 nickm Exp $
+# $Id: Common.py,v 1.124.2.1 2003/12/18 23:13:55 nickm Exp $
 
 """mixminion.Common
 
    Common functionality and utility code for Mixminion"""
 
-__all__ = [ 'IntervalSet', 'Lockfile', 'LOG', 'LogStream', 'MixError',
+__all__ = [ 'IntervalSet', 'Lockfile', 'LockfileLocked', 'LOG', 'LogStream',
+            'MixError',
             'MixFatalError', 'MixProtocolError', 'UIError', 'UsageError',
             'armorText', 'ceilDiv', 'checkPrivateDir', 'checkPrivateFile',
-            'createPrivateDir', 'encodeBase64', 'floorDiv', 'formatBase64',
+            'createPrivateDir', 'disp64',
+            'encodeBase64', 'englishSequence', 'floorDiv', 'formatBase64',
             'formatDate', 'formatFnameTime', 'formatTime',
             'installSIGCHLDHandler', 'isSMTPMailbox', 'openUnique',
             'previousMidnight', 'readFile', 'readPickled',
@@ -21,11 +23,8 @@ import bisect
 import calendar
 import cPickle
 import errno
-import fcntl
-import grp
 import gzip
 import os
-import pwd
 import re
 import signal
 import stat
@@ -36,9 +35,27 @@ import time
 import traceback
 # Imported here so we can get it in mixminion.server without being shadowed
 # by the old Queue.py file.
-from Queue import Queue
+from Queue import Queue, Empty
 MessageQueue = Queue
+QueueEmpty = Empty
 del Queue
+del Empty
+
+O_BINARY = getattr(os, 'O_BINARY', 0)
+
+try:
+    import fcntl
+except ImportError:
+    fcntl = None
+try:
+    import msvcrt
+except ImportError:
+    mcvcrt = None
+
+try:
+    import pwd, grp
+except ImportError:
+    pwd = grp = None
 
 from types import StringType
 
@@ -54,6 +71,10 @@ class MixProtocolError(MixError):
     """Exception class for MMTP protocol violations"""
     pass
 
+class TimeoutError(MixProtocolError):
+    """Exception raised for protocol timeout."""
+    pass
+
 class MixProtocolReject(MixProtocolError):
     """Exception class for server-rejected packets."""
     pass
@@ -66,17 +87,22 @@ class UIError(MixError):
     """Exception raised for an error that should be reported to the user,
        not dumped as a stack trace."""
     def dump(self):
-        if str(self): print >>sys.stderr, "ERROR:", str(self)
+        if str(self):
+            print >>sys.stderr, _logtime(), "[ERROR]", str(self)
     def dumpAndExit(self):
         self.dump()
         sys.exit(1)
 
 class UsageError(UIError):
     """Exception raised for an error that should be reported to the user
-       along with a usage mesage.
+       along with a usage message.
     """
     pass
 
+class MixFilePermissionError(MixFatalError,UIError):
+    """Exception raised when a file has the wrong owner or permissions."""
+    pass
+
 #----------------------------------------------------------------------
 # Portability to future Python versions
 
@@ -139,7 +165,7 @@ if sys.version_info[0:3] >= (2,1,0):
         """Convert 's' to a one-line base-64 representation."""
         return binascii.b2a_base64(s).strip()
 else:
-    # Python 2.0 didin't allow a binascii to return more than one line.
+    # Python 2.0 didn't allow a binascii to return more than one line.
     def formatBase64(s):
         """Convert 's' to a one-line base-64 representation."""
         return encodeBase64(s, 64, 1)
@@ -159,6 +185,56 @@ def encodeBase64(s, lineWidth=64, oneline=0):
     else:
         return "".join(pieces)
 
+def disp64(s,n=-1):
+    """Return a 'beautified' base64 for use in log messages."""
+    s = formatBase64(s)
+    if n >= 0:
+        s = s[:n]
+    while s.endswith('='):
+        s = s[:-1]
+    return s
+
+def englishSequence(lst, empty="none", compound="and"):
+    """Given a sequence of items, return the sequence formatted
+       according to ordinary English conventions of punctuation.
+
+       If the list is empty, the value of 'empty' will be returned."""
+
+    if len(lst) == 0:
+        return empty
+    elif len(lst) == 1:
+        return lst[0]
+
+    punc = ", "
+    for item in lst:
+        if ("," in item or
+            stringContains(item, " and ") or
+            stringContains(item, " or ")):
+            punc = "; "
+            break
+
+    if len(lst) == 2:
+        if punc == ", ":
+            return "%s %s %s" % (lst[0], compound, lst[1])
+        else:
+            return "%s; %s %s" % (lst[0], compound, lst[1])
+    else:
+        return "%s%s%s %s" % (punc.join(lst[0:-1]), punc, compound, lst[-1])
+
+_HOST_CHARS = ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"+
+               "abcdefghijklmnopqrstuvwxyz"+
+               "0123456789.-")
+def isPlausibleHostname(s):
+    """Return true iff 's' is made up only of characters that sometimes
+       appear in hostnames, and has a plausible arrangement of dots."""
+    if not s:
+        return 0
+    if s.translate(_ALLCHARS, _HOST_CHARS):
+        return 0
+    if stringContains(s, "..") or s.startswith("."):
+        return 0
+    return 1
+
 #----------------------------------------------------------------------
 # Functions to generate and parse OpenPGP-style ASCII armor
 
@@ -170,7 +246,7 @@ def armorText(s, type, headers=(), base64=1):
        list of key-value pairs for headers, generates an OpenPGP-style
        ASCII-armored message of type 'type', with contents 's' and
        headers 'header'.
-       
+
        If base64 is false, uses cleartext armor."""
     result = []
     result.append("-----BEGIN %s-----\n" %type)
@@ -204,7 +280,7 @@ def unarmorText(s, findTypes, base64=1, base64fn=None):
           we do cleartext armor.
     """
     result = []
-    
+
     while 1:
         tp = None
         fields = []
@@ -247,6 +323,8 @@ def unarmorText(s, findTypes, base64=1, base64fn=None):
 
         if base64:
             try:
+                if stringContains(s[idx:endIdx], "\n[...]\n"):
+                    raise ValueError("Value seems to be truncated by a Mixminion-Mixmaster gateway")
                 value = binascii.a2b_base64(s[idx:endIdx])
             except (TypeError, binascii.Incomplete, binascii.Error), e:
                 raise ValueError(str(e))
@@ -258,16 +336,56 @@ def unarmorText(s, findTypes, base64=1, base64fn=None):
             value = "\n".join(v)
 
         result.append((tp, fields, value))
-        
+
         s = s[mEnd.end()+1:]
 
     raise MixFatalError("Unreachable code somehow reached.")
 
 #----------------------------------------------------------------------
+
+#----------------------------------------------------------------------
+
+# A set of directories we've issued warnings about -- we won't check
+# them again.
+_WARNED_DIRECTORIES = {}
+# A set of directories that have checked out -- we won't check them again.
+_VALID_DIRECTORIES = {}
+# A list of user IDs
+_TRUSTED_UIDS = [ 0 ]
+
+# Flags: what standard Unix access controls should we check for?
+_CHECK_UID = 1
+_CHECK_GID = 1
+_CHECK_MODE = 1
+
+if sys.platform in ('cygwin', 'win32'):
+    # Under windows, we can't do anything sensible with permissions AFAICT.
+    _CHECK_UID = _CHECK_GID = _CHECK_MODE = 0
+elif os.environ.get("MM_NO_FILE_PARANOIA"):
+    _CHECK_UID = _CHECK_GID = _CHECK_MODE = 0
+
+def _uidToName(uid):
+    """Helper function: given a uid, return a username or descriptive
+       string."""
+    try:
+        return pwd.getpwuid(uid)[0]
+    except (KeyError, AttributeError):
+        # KeyError: no such pwent.  AttributeError: pwd module not loaded.
+        return "user %s"%uid
+
+def _gidToName(gid):
+    """Helper function: given a gid, return a groupname or descriptive string
+       """
+    try:
+        return grp.getgrgid(gid)[0]
+    except (KeyError, AttributeError):
+        # KeyError: no such grpent.  AttributeError: grp module not loaded.
+        return "group %s"%gid
+
 def checkPrivateFile(fn, fix=1):
     """Checks whether f is a file owned by this uid, set to mode 0600 or
        0700, and all its parents pass checkPrivateDir.  Raises MixFatalError
-       if the assumtions are not met; else return None.  If 'fix' is true,
+       if the assumptions are not met; else return None.  If 'fix' is true,
        repair permissions on the file rather than raising MixFatalError."""
     parent, _ = os.path.split(fn)
     checkPrivateDir(parent)
@@ -275,20 +393,27 @@ def checkPrivateFile(fn, fix=1):
         st = os.stat(fn)
     except OSError, e:
         if e.errno == errno.EEXIST:
-            raise MixFatalError("Nonexistant file %s" % fn)
+            raise MixFatalError("Nonexistent file %s" % fn)
         else:
-            raise MixFatalError("Could't stat file %s: %s" % (fn, e))
+            raise MixFatalError("Couldn't stat file %s: %s" % (fn, e))
     if not st:
-        raise MixFatalError("Nonexistant file %s" % fn)
+        raise MixFatalError("Nonexistent file %s" % fn)
     if not os.path.isfile(fn):
         raise MixFatalError("%s is not a regular file" % fn)
-    me = os.getuid()
-    if st[stat.ST_UID] != me:
-        raise MixFatalError("File %s must have owner %s" % (fn, me))
+
+    if _CHECK_UID:
+        me = os.getuid()
+        if st[stat.ST_UID] != me:
+            ownerName = _uidToName(st[stat.ST_UID])
+            myName = _uidToName(me)
+            raise MixFilePermissionError(
+                "File %s is owned by %s, but Mixminion is running as %s"
+                % (fn, ownerName, myName))
     mode = st[stat.ST_MODE] & 0777
-    if mode not in (0700, 0600):
+    if _CHECK_MODE and mode not in (0700, 0600):
         if not fix:
-            raise MixFatalError("Bad mode %o on file %s" % (mode & 0777, fn))
+            raise MixFilePermissionError("Bad permissions (mode %o) on file %s"
+                                         % (mode & 0777, fn))
         newmode = {0:0600,0100:0700}[(mode & 0100)]
         LOG.warn("Repairing permissions on file %s" % fn)
         os.chmod(fn, newmode)
@@ -302,21 +427,19 @@ def createPrivateDir(d, nocreate=0):
         try:
             os.makedirs(d, 0700)
         except OSError, e:
-            raise MixFatalError("Unable to create directory %s: %s" % d, e)
+            raise MixFatalError(
+                "Unable to create directory %s: %s" % (d, e))
 
     checkPrivateDir(d)
 
-_WARNED_DIRECTORIES = {}
-_VALID_DIRECTORIES = {}
-_TRUSTED_UIDS = [ 0 ]
-
 def checkPrivateDir(d, recurse=1):
     """Check whether d is a directory owned by this uid, set to mode
        0700. All of d's parents must not be writable or owned by anybody but
        this uid and uid 0.  If any of these conditions are unmet, raise
        MixFatalErrror.  Otherwise, return None."""
-    me = os.getuid()
-    trusted_uids = _TRUSTED_UIDS + [ me ]
+    if _CHECK_UID:
+        me = os.getuid()
+        trusted_uids = _TRUSTED_UIDS + [ me ]
 
     if not os.path.isabs(d):
         d = os.path.abspath(d)
@@ -328,11 +451,15 @@ def checkPrivateDir(d, recurse=1):
 
     st = os.stat(d)
     # check permissions
-    if st[stat.ST_MODE] & 0777 != 0700:
-        raise MixFatalError("Directory %s must be mode 0700" % d)
+    if _CHECK_MODE and st[stat.ST_MODE] & 0777 != 0700:
+        raise MixFilePermissionError("Directory %s must be mode 0700" % d)
 
-    if st[stat.ST_UID] != me:
-        raise MixFatalError("Directory %s has must have owner %s" %(d, me))
+    if _CHECK_UID and st[stat.ST_UID] != me:
+        ownerName = _uidToName( st[stat.ST_UID])
+        myName = _uidToName(me)
+        raise MixFilePermissionError(
+            "Directory %s is owned by %s, but Mixminion is running as %s"
+            %(d,ownerName,myName))
 
     if not recurse:
         return
@@ -349,39 +476,64 @@ def checkPrivateDir(d, recurse=1):
         st = os.stat(d)
         mode = st[stat.ST_MODE]
         owner = st[stat.ST_UID]
-        if owner not in trusted_uids:
-            raise MixFatalError("Bad owner (uid=%s) on directory %s"
-                                % (owner, d))
-        if (mode & 02) and not (mode & stat.S_ISVTX):
-            raise MixFatalError("Bad mode (%o) on directory %s" %
-                                (mode&0777, d))
+        if _CHECK_UID and owner not in trusted_uids:
+            ownerName = _uidToName(owner)
+            trustedNames = map(_uidToName,trusted_uids)
+            raise MixFilePermissionError(
+                "Directory %s is owned by %s, but I only trust %s"
+                % (d, ownerName, englishSequence(trustedNames, "(nobody)")))
+        if _CHECK_MODE and (mode & 02) and not (mode & stat.S_ISVTX):
+            raise MixFilePermissionError(
+                "Bad permissions (mode %o) on directory %s" %
+                (mode&0777, d))
 
-        if (mode & 020) and not (mode & stat.S_ISVTX):
+        if _CHECK_MODE and (mode & 020) and not (mode & stat.S_ISVTX):
             # FFFF We may want to give an even stronger error here.
-            if not _WARNED_DIRECTORIES.has_key(d):
-                group = grp.getgrgid(st[stat.ST_GID])[0]
+            if _CHECK_GID and not _WARNED_DIRECTORIES.has_key(d):
+                groupName = _gidToName(st[stat.ST_GID])
                 LOG.warn("Directory %s is writable by group %s (mode %o)",
-                         d, group, mode&0777)
-            _WARNED_DIRECTORIES[d] = 1
+                         d, groupName, mode&0777)
+                _WARNED_DIRECTORIES[d] = 1
+
+def configureFileParanoia(config):
+    global _CHECK_UID
+    global _CHECK_GID
+    global _CHECK_MODE
+    paranoia = config['Host']['FileParanoia']
+    if not paranoia:
+        _CHECK_UID = _CHECK_GID = _CHECK_MODE = 0
 
-def configureTrustedUsers(config):
     users = config['Host']['TrustedUser']
     if not users:
         return
+    if not _CHECK_UID:
+        return
 
-    for u in users:
-        u = u.strip()
-        try:
-            ent = pwd.getpwnam(u)
-        except KeyError:
-            LOG.warn("TrustedUser: No such user as %s", u)
-            continue
-
-        uid = ent[2]
+    for uid in users:
         _TRUSTED_UIDS.append(uid)
 
 #----------------------------------------------------------------------
 # File helpers
+
+
+# On windows, rename(f1,f2) fails if f2 already exists.  These wrappers
+# handle replacing files.
+if sys.platform == 'win32':
+    def replaceFile(f1, f2):
+        """Move the file named 'f1' to a new name 'f2'.  Replace any file
+           already named 'f2'."""
+        # WWWW This isn't atomic.  Later versions of the windows API add
+        # WWWW functions named MoveFileEx and ReplaceFile that may do the
+        # WWWW right thing, but those don't exist in Windows Me/98/95.
+        if os.path.exists(f2):
+            os.unlink(f2)
+        os.rename(f1, f2)
+else:
+    def replaceFile(f1, f2):
+        """Move the file named 'f1' to a new name 'f2'.  Replace any file
+           already named 'f2'."""
+        os.rename(f1, f2)
+
 class AtomicFile:
     """Wrapper around open/write/rename to encapsulate writing to a temporary
        file, then moving to the final filename on close.
@@ -403,7 +555,7 @@ class AtomicFile:
 
     def close(self):
         """Close the underlying file and replace the destination file."""
-        os.rename(self.tmpname, self.fname)
+        replaceFile(self.tmpname, self.fname)
         self.f.close()
         self.f = None
 
@@ -445,7 +597,7 @@ def writeFile(fn, contents, mode=0600, binary=0):
         if os.path.exists(tmpname): os.unlink(tmpname)
         raise
 
-    os.rename(tmpname, fn)
+    replaceFile(tmpname, fn)
 
 def readPickled(fn):
     """Given the name of a file containing a pickled object, return the pickled
@@ -471,7 +623,7 @@ def writePickled(fn, obj, mode=0600):
         if os.path.exists(tmpname): os.unlink(tmpname)
         raise
 
-    os.rename(tmpname, fn)
+    replaceFile(tmpname, fn)
 
 def tryUnlink(fname):
     """Try to remove the file named fname.  If the file is erased, return 1.
@@ -530,19 +682,27 @@ def _overwriteFile(f):
         sz = _BLKSIZEMAP[parent]
     except KeyError:
         if hasattr(os, 'statvfs'):
-            sz = os.statvfs(f)[statvfs.F_BSIZE]
+            try:
+                sz = os.statvfs(parent)[statvfs.F_BSIZE]
+            except OSError:
+                sz = 8192 # Should be a safe guess? (????)
         else:
             sz = 8192 # Should be a safe guess? (????)
         _BLKSIZEMAP[parent] = sz
         if sz > len(_NILSTR):
             _NILSTR = '\x00' * sz
     nil = _NILSTR[:sz]
-    fd = os.open(f, os.O_WRONLY)
+    try:
+        fd = os.open(f, os.O_WRONLY|O_BINARY)
+    except OSError:
+        return
     try:
         size = os.fstat(fd)[stat.ST_SIZE]
         blocks = ceilDiv(size, sz)
         for _ in xrange(blocks):
             os.write(fd, nil)
+        if hasattr(os, 'fsync'):
+            os.fsync(fd)
     finally:
         os.close(fd)
 
@@ -565,7 +725,7 @@ def secureDelete(fnames, blocking=0):
        partition from the get-go... or to physically destroy and
        replace your hard drive every so often.)
 
-       So we don't even bother trying to make the data 'physcially
+       So we don't even bother trying to make the data 'physically
        irretrievable.'  We just zero it out, which should be good
        enough to stymie root for most purposes, and totally inadequate
        against a well-funded adversary with access to your hard drive
@@ -635,11 +795,11 @@ class _FileLogHandler:
         try:
             parent = os.path.split(self.fname)[0]
             if not os.path.exists(parent):
-                createPrivateDir(parent)
+                os.mkdirs(parent, 0700)
             self.file = open(self.fname, 'a')
         except OSError, e:
             self.file = None
-            raise MixError("Unable to open log file %r: %s"%self.fname, e)
+            raise MixError("Unable to open log file %r: %s"%(self.fname, e))
     def close(self):
         "Close the underlying file"
         self.file.close()
@@ -686,16 +846,19 @@ class Log:
                  message or a connection.
               FATAL: nonrecoverable errors that affect the entire system.
 
-       In practise, we instantiate only a single instance of this class,
+       In practice, we instantiate only a single instance of this class,
        accessed as mixminion.Common.LOG."""
     ## Fields:
     # handlers: a list of logHandler objects.
     # severity: a severity below which log messages are ignored.
+    # silenceNoted: true iff we have printed a message about silencing the
+    #     console long.
     def __init__(self, minSeverity):
         """Create a new Log object that ignores all message less severe than
            minSeverity, and sends its output to stderr."""
         self.configure(None)
         self.setMinSeverity(minSeverity)
+        self.silenceNoted = 0
         self.__lock = threading.Lock()
 
     def configure(self, config, keepStderr=0):
@@ -707,28 +870,39 @@ class Log:
            """
         self.handlers = []
         if config == None or not config.has_section("Server"):
+            # We're configuring a client.
             self.setMinSeverity("WARN")
             self.addHandler(_ConsoleLogHandler(sys.stderr))
-        else:
-            self.setMinSeverity(config['Server'].get('LogLevel', "WARN"))
-            logfile = config['Server'].get('LogFile')
-            if logfile is None:
-                homedir = config['Server']['Homedir']
-                if homedir:
-                    logfile = os.path.join(homedir, "log")
-            self.addHandler(_ConsoleLogHandler(sys.stderr))
-            if logfile:
-                try:
-                    self.addHandler(_FileLogHandler(logfile))
-                except MixError, e:
-                    self.error(str(e))
-                if keepStderr:
-                    return
-                if (config['Server'].get('Daemon',0) or
-                    not config['Server'].get('EchoMessages',0)):
+            return
+
+        # We're configuring the log on a server.  Stuff will get complicated
+        # now.  First, we set the severity as specified in the configuration...
+        self.setMinSeverity(config['Server'].get('LogLevel', "WARN"))
+        # ...find out what logfile we're supposed to use ...
+        logfile = config.getLogFile()
+        # ...and add a handler to log any messages to stderr.
+        self.addHandler(_ConsoleLogHandler(sys.stderr))
+        if logfile:
+            # First, try to create a file log handler.
+            try:
+                self.addHandler(_FileLogHandler(logfile))
+            except MixError, e:
+                self.error(str(e))
+                return
+            # If we're successful, and we're absolutely supposed to echo
+            # messages (or keepStderr is set), we don't even consider
+            # dumping the console log handler.
+            if keepStderr or config['Server'].get('EchoMessages',0)==2:
+                return
+            # If we're running in daemon mode, or we're not supposed to echo
+            # messages, we remove the console log handler.
+            if (config['Server'].get('Daemon',0) or
+                not config['Server'].get('EchoMessages',0)):
+                if not self.silenceNoted:
                     print "Silencing the console log; look in %s instead"%(
                         logfile)
-                    del self.handlers[0]
+                    self.silenceNoted = 1
+                del self.handlers[0]
 
     def setMinSeverity(self, minSeverity):
         """Sets the minimum severity of messages to be logged.
@@ -812,7 +986,7 @@ class Log:
         self.log("FATAL", message, *args)
     def log_exc(self, severity, (exclass, ex, tb), message=None, *args):
         """Write an exception and stack trace to the log.  If message and
-           args are provided, use them as an explanitory message; otherwise,
+           args are provided, use them as an explanatory message; otherwise,
            introduce the message as "Unexpected exception".
 
            This should usually be called as
@@ -914,18 +1088,18 @@ def formatTime(when,localtime=0):
         gmt = time.localtime(when)
     else:
         gmt = time.gmtime(when)
-    return "%04d/%02d/%02d %02d:%02d:%02d" % (
+    return "%04d-%02d-%02d %02d:%02d:%02d" % (
         gmt[0],gmt[1],gmt[2],  gmt[3],gmt[4],gmt[5])
 
 def formatDate(when):
     """Given a time in seconds since the epoch, returns a date value in the
        format used by server descriptors (YYYY/MM/DD) in GMT"""
     gmt = time.gmtime(when+1) # Add 1 to make sure we round down.
-    return "%04d/%02d/%02d" % (gmt[0],gmt[1],gmt[2])
+    return "%04d-%02d-%02d" % (gmt[0],gmt[1],gmt[2])
 
 def formatFnameTime(when=None):
     """Given a time in seconds since the epoch, returns a date value suitable
-       for use as part of a fileame.  Defaults to the current time."""
+       for use as part of a filename.  Defaults to the current time."""
     if when is None:
         when = time.time()
     return time.strftime("%Y%m%d%H%M%S", time.localtime(when))
@@ -989,7 +1163,7 @@ class Duration:
         return self
 
 #----------------------------------------------------------------------
-# InteralSet
+# IntervalSet
 
 class IntervalSet:
     """An IntervalSet is a mutable set of numeric intervals, closed below and
@@ -1114,7 +1288,7 @@ class IntervalSet:
         """Returns a list of (start,end) tuples for a the intervals in this
            set."""
         s = []
-        for i in range(0, len(self.edges), 2):
+        for i in xrange(0, len(self.edges), 2):
             s.append((self.edges[i][0], self.edges[i+1][0]))
         return s
 
@@ -1122,7 +1296,7 @@ class IntervalSet:
         """Helper function: raises AssertionError if this set's data is
            corrupted."""
         assert (len(self.edges) % 2) == 0
-        for i in range(0, len(self.edges), 2):
+        for i in xrange(0, len(self.edges), 2):
             assert self.edges[i][0] < self.edges[i+1][0]
             assert self.edges[i][1] == '+'
             assert self.edges[i+1][1] == '-'
@@ -1133,7 +1307,7 @@ class IntervalSet:
         return cmp(self.edges, other.edges)
 
     def start(self):
-        """Return the first point contained in this inverval."""
+        """Return the first point contained in this interval."""
         return self.edges[0][0]
 
     def end(self):
@@ -1143,7 +1317,7 @@ class IntervalSet:
 #----------------------------------------------------------------------
 # SMTP address functionality
 
-# Regular expressions to valide RFC822 addresses.
+# Regular expressions to validate RFC822 addresses.
 # (This is more strict than RFC822, actually.  RFC822 allows tricky stuff to
 #   quote special characters, and I don't trust every MTA or delivery command
 #   to support addresses like <bob@bob."; rm -rf /; echo".com>
@@ -1178,13 +1352,16 @@ def isSMTPMailbox(s):
 
 def waitForChildren(onceOnly=0, blocking=1):
     """Wait until all subprocesses have finished.  Useful for testing."""
+    if sys.platform == 'win32':
+        LOG.trace("Skipping waitForChildren")
+        return
     if blocking:
         options = 0
     else:
         options = os.WNOHANG
     while 1:
         try:
-            # WIN32 This won't work on Windows.  What to do?
+            # WWWW This won't work on Windows.  What to do?
             pid, status = os.waitpid(-1, options)
         except OSError, e:
             return
@@ -1201,7 +1378,6 @@ def _sigChldHandler(signal_num, _):
 
     while 1:
         try:
-            # WIN32 This waitpid call won't work on Windows.  What to do?
             pid, status = os.waitpid(-1, os.WNOHANG)
             if pid == 0:
                 break
@@ -1213,6 +1389,8 @@ def _sigChldHandler(signal_num, _):
 
 def installSIGCHLDHandler():
     '''Register sigchld handler for this process.'''
+    if sys.platform == 'win32':
+        return
     signal.signal(signal.SIGCHLD, _sigChldHandler)
 
 #----------------------------------------------------------------------
@@ -1236,11 +1414,15 @@ def openUnique(fname, mode='w', perms=0600):
     """Helper function. Returns a file open for writing into the file named
        'fname'.  If fname already exists, opens 'fname.1' or 'fname.2' or
        'fname.3' or so on."""
+    if 'b' in mode:
+        bin = O_BINARY
+    else:
+        bin = 0
     base, rest = os.path.split(fname)
     idx = 0
     while 1:
         try:
-            fd = os.open(fname, os.O_WRONLY|os.O_CREAT|os.O_EXCL, perms)
+            fd = os.open(fname, os.O_WRONLY|os.O_CREAT|os.O_EXCL|bin, perms)
             return os.fdopen(fd, mode), fname
         except OSError, e:
             if e.errno != errno.EEXIST:
@@ -1251,6 +1433,11 @@ def openUnique(fname, mode='w', perms=0600):
     raise MixFatalError("unreachable code")
 
 #----------------------------------------------------------------------
+class LockfileLocked(Exception):
+    """Exception raised when trying to get a nonblocking lock on a locked
+       lockfile"""
+    pass
+
 class Lockfile:
     """Class to implement a recursive advisory lock, using flock on a
        'well-known' filename."""
@@ -1269,29 +1456,43 @@ class Lockfile:
         """Return the contents of the lock file."""
         return readFile(self.filename)
 
-    def acquire(self, contents="", blocking=0):
+    def acquire(self, contents="", blocking=0, mode=0600):
         """Acquire this lock.  If we're acquiring the lock for the first time,
-           write 'contents' to the lockfile.  If 'blocking' is true, wait until
-           we can acquire the lock.  If 'blocking' is false, raise IOError if
-           we can't acquire the lock."""
+           write 'contents' to the lockfile.  If 'blocking' is true, wait
+           until we can acquire the lock.  If 'blocking' is false, raise
+           LockfileLocked if we can't acquire the lock."""
+
         if self.count > 0:
             self.count += 1
             return
 
         assert self.fd is None
-        self.fd = os.open(self.filename, os.O_RDWR|os.O_CREAT, 0600)
+        self.fd = os.open(self.filename, os.O_RDWR|os.O_CREAT, mode)
         try:
-            if blocking:
-                fcntl.flock(self.fd, fcntl.LOCK_EX)
-            else:
-                fcntl.flock(self.fd, fcntl.LOCK_EX|fcntl.LOCK_NB)
+            self._lock(self.fd, blocking)
             self.count += 1
             os.write(self.fd, contents)
+            if hasattr(os, "fsync"):
+                os.fsync(self.fd)
         except:
             os.close(self.fd)
             self.fd = None
             raise
 
+    def replaceContents(self, contents):
+        """Replace the current contents of the lockfile with 'contents',
+           without releasing the lock.  Invokers must hold the lock."""
+        # XXXX Actually, on Win32, this actually replaces the first
+        # XXXX len(contents) characters of the lockfile.  This is a bug.
+        assert self.count > 0 and self.fd is not None
+        SEEK_SET = 0
+        os.lseek(self.fd, 0, SEEK_SET)
+        if hasattr(os, 'ftruncate'): # Doesn't exist on windows.
+            os.ftruncate(self.fd, 0)
+        os.write(self.fd, contents)
+        if hasattr(os, 'fsync'):
+            os.fsync(self.fd)
+
     def release(self):
         """Release the lock."""
         assert self.fd is not None
@@ -1303,7 +1504,7 @@ class Lockfile:
         except OSError:
             pass
         try:
-            fcntl.flock(self.fd, fcntl.LOCK_UN)
+            self._unlock(self.fd)
         except OSError:
             pass
         try:
@@ -1312,3 +1513,123 @@ class Lockfile:
             pass
 
         self.fd = None
+
+    def _lock(self, fd, blocking):
+        """Compatibility wrapper to implement file locking for posix and win32
+           systems.  If 'blocking' is false, and the lock cannot be obtained,
+           raises LockfileLocked."""
+        if fcntl:
+            # Posixy systems have a friendly neighborhood flock clone.
+            flags = fcntl.LOCK_EX
+            if not blocking: flags |= fcntl.LOCK_NB
+            try:
+                fcntl.flock(fd, flags)
+            except IOError, e:
+                if e.errno in (errno.EAGAIN, errno.EACCES) and not blocking:
+                    raise LockfileLocked()
+                else:
+                    raise
+        elif msvcrt:
+            # Windows has decided that System V's perennially unstandardized
+            # "locking" is a cool idea.
+            os.lseek(fd, 0, 0)
+            # The msvcrt.locking() function never gives you a blocking lock.
+            # If you ask for one, it just retries once per second for ten
+            # seconds.  This must be some genius's idea of a 'feature'.
+            while 1:
+                try:
+                    msvcrt.locking(fd, msvcrt.LK_NBLCK, 0)
+                    return
+                except IOError, e:
+                    if e.errno not in (errno.EAGAIN, errno.EACCES):
+                        raise
+                    elif not blocking:
+                        raise LockfileLocked()
+                    else:
+                        time.sleep(0.5)
+        else:
+            # There is no locking implementation.
+            _warn_no_locks()
+
+    def _unlock(self, fd):
+        """Compatibility wrapper: unlock a file for unix and windows systems.
+        """
+        if fcntl:
+            fcntl.flock(fd, fcntl.LOCK_UN)
+        elif msvcrt:
+            os.lseek(fd, 0, 0)
+            msvcrt.locking(fd, msvcrt.LK_UNLCK, 0)
+        else:
+            _warn_no_locks()
+
+_warned_no_locks = 0
+def _warn_no_locks():
+    global _warned_no_locks
+    if not _warned_no_locks:
+        _warned_no_locks = 1
+        LOG.warn("Mixminion couldn't find a file locking implementation.")
+        LOG.warn("  (Simultaneous accesses may lead to data corruption.")
+
+#----------------------------------------------------------------------
+# Threading operations
+
+class ClearableQueue(MessageQueue):
+    """Extended version of python's Queue class that supports removing
+       all the items from the queue."""
+    def clear(self):
+        """Remove all the items from this queue."""
+        # If the queue is empty, return.
+        if not self.esema.acquire(0):
+            return
+        self.mutex.acquire()
+        was_full = self._full()
+        self._clear()
+        assert self._empty()
+        # If the queue used to be full, it isn't anymore.
+        if was_full:
+            self.fsema.release()
+        self.mutex.release()
+
+    def _clear(self):
+        """Backend for _clear"""
+        del self.queue[:]
+
+try:
+    q = MessageQueue()
+    q.put(3)
+    q.get(timeout=10)
+    BUILTIN_QUEUE_HAS_TIMEOUT = 1
+except TypeError:
+    BUILTIN_QUEUE_HAS_TIMEOUT = 0
+del q
+
+if BUILTIN_QUEUE_HAS_TIMEOUT:
+    TimeoutQueue = ClearableQueue
+else:
+    class TimeoutQueue(ClearableQueue):
+        """Helper class for Python 2.2. and earlier: extends the 'get'
+           functionality of Queue.Queue to support a 'timeout' argument.
+           If 'block' is true and timeout is provided, wait for no more
+           than 'timeout' seconds before raising QueueEmpty.
+
+           In Python 2.3 and later, this interface is standard.
+        """
+        def get(self, block=1, timeout=None):
+            if timeout is None or not block:
+                return MessageQueue.get(self, block)
+
+            # This logic is adapted from 'Condition' in the Python
+            # threading module.
+            _time = time.time
+            _sleep = time.sleep
+            deadline = timeout+_time()
+            delay = .0005
+            while 1:
+                try:
+                    return MessageQueue.get(self,0)
+                except QueueEmpty:
+                    remaining = deadline-_time()
+                    if remaining <= 0:
+                        raise
+                    delay = min(delay*2,remaining,0.2)
+                    _sleep(delay)
