@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: Config.py,v 1.47 2003/06/05 18:41:40 nickm Exp $
+# $Id: Config.py,v 1.72 2003/12/08 02:22:56 nickm Exp $
 
 """Configuration file parsers for Mixminion client and server
    configuration.
@@ -55,13 +55,20 @@ import calendar
 import binascii
 import os
 import re
-import socket # for inet_aton and error
+import sys
+try:
+    import pwd
+except ImportError:
+    pwd = None
+
 from cStringIO import StringIO
 
 import mixminion.Common
 import mixminion.Crypto
-from mixminion.Common import MixError, LOG, ceilDiv, isPrintingAscii, \
-   stripSpace, stringContains
+import mixminion.NetUtils
+
+from mixminion.Common import MixError, LOG, ceilDiv, englishSequence, \
+     formatBase64, isPrintingAscii, stripSpace, stringContains, UIError
 
 class ConfigError(MixError):
     """Thrown when an error is found in a configuration file."""
@@ -101,7 +108,7 @@ def _parseServerMode(mode):
 
 # re to match strings of the form '9 seconds', '1 month', etc.
 _interval_re = re.compile(r'''^(\d+\.?\d*|\.\d+)\s+
-                     (sec|second|min|minute|hour|day|week|mon|month|year)s?$''',
+                   (sec|second|min|minute|hour|day|week|mon|month|year)s?$''',
                           re.X)
 _seconds_per_unit = {
     'second': 1,
@@ -159,6 +166,29 @@ def _parseIntervalList(s):
             ilist.append(interval)
     return ilist
 
+def _unparseIntervalList(lst):
+    """Helper function: given an interval list, converts it back to the
+       expected format."""
+    if lst == []:
+        return ""
+    r = [ (lst[0], 1) ]
+    for dur in lst[1:]:
+        if dur == r[-1][0]:
+            r[-1] = (dur, r[-1][1]+1)
+        else:
+            r.append((dur,1))
+    result = []
+    for dur, reps in r:
+        d = mixminion.Common.Duration(dur)
+        t = mixminion.Common.Duration(dur*reps)
+        d.reduce()
+        t.reduce()
+        if reps>1:
+            result.append("every %s for %s"%(d,t))
+        else:
+            result.append(str(d))
+    return ", ".join(result)
+
 def _parseInt(integer):
     """Validation function.  Converts a config value to an int.
        Raises ConfigError on failure."""
@@ -168,25 +198,58 @@ def _parseInt(integer):
     except ValueError:
         raise ConfigError("Expected an integer but got %r" % (integer))
 
-# Regular expression to match a dotted quad.
-_ip_re = re.compile(r'^\d+\.\d+\.\d+\.\d+$')
+# regular expression to match a size.
+_size_re = re.compile(r'^(\d+\.?\d*|\.\d+)\s*(k|kb|m|mb|b|byte|octet|)s?')
+_size_name_map = { '' : 1L, 'b' : 1L, 'byte' : 1L, 'octet' : 1L,
+                   'k' : 1L<<10, 'kb' : 1L<<10,
+                   'm' : 1L<<20, 'mb' : 1L<<20,
+                   'g' : 1L<<30, 'gb' : 1L<<30 }
+def _parseSize(size):
+    """Validation function.  Converts a config value to a size in octets.
+       Raises ConfigError on failure."""
+    s = size.strip().lower()
+    m = _size_re.match(s)
+    if not m: raise ConfigError("Invalid size %r"%size)
+    val = m.group(1)
+    unit = _size_name_map[m.group(2)]
+    if '.' in val:
+        return long(float(val)*unit)
+    else:
+        return long(val)*unit
+
+def _unparseSize(size):
+    names = ["b", "KB", "MB", "GB"]
+    idx = 0
+    while 1:
+        if (size & 1023) or names[idx] == "GB":
+            return "%s %s"%(size,names[idx])
+        else:
+            idx += 1
+            size >>= 10
 
 def _parseIP(ip):
     """Validation function.  Converts a config value to an IP address.
        Raises ConfigError on failure."""
-    i = ip.strip()
-
-    # inet_aton is a bit more permissive about spaces and incomplete
-    # IP's than we want to be.  Thus we use a regex to catch the cases
-    # it doesn't.
-    if not _ip_re.match(i):
-        raise ConfigError("Invalid IP %r" % i)
     try:
-        socket.inet_aton(i)
-    except socket.error:
-        raise ConfigError("Invalid IP %r" % i)
+        return mixminion.NetUtils.normalizeIP4(ip)
+    except ValueError, e:
+        raise ConfigError(str(e))
 
-    return i
+def _parseIP6(ip6):
+    """Validation function.  Converts a config value to an IP address.
+       Raises ConfigError on failure."""
+    try:
+        return mixminion.NetUtils.normalizeIP6(ip6)
+    except ValueError, e:
+        raise ConfigError(str(e))
+
+def _parseHost(host):
+    """Validation function.  Checks a config value as a valid hostname.
+       Raises ConfigError on failure."""
+    host = host.strip()
+    if not mixminion.Common.isPlausibleHostname(host):
+        raise ConfigError("%r doesn't look like a valid hostname"%host)
+    return host
 
 # Regular expression to match 'address sets' as used in Allow/Deny
 # configuration lines. General format is "<IP|*> ['/'MASK] [PORT['-'PORT]]"
@@ -203,11 +266,11 @@ def _parseAddressSet_allow(s, allowMode=1):
     s = s.strip()
     m = _address_set_re.match(s)
     if not m:
-        raise ConfigError("Misformatted address rule %r", s)
+        raise ConfigError("Misformatted address rule %r" % s)
     ip, mask, port, porthi = m.groups()
     if ip == '*':
         if mask != None:
-            raise ConfigError("Misformatted address rule %r", s)
+            raise ConfigError("Misformatted address rule %r" % s)
         ip,mask = '0.0.0.0','0.0.0.0'
     else:
         ip = _parseIP(ip)
@@ -287,46 +350,67 @@ def _parsePublicKey(s):
         raise ConfigError("Invalid exponent on public key")
     return key
 
-# Regular expression to match YYYY/MM/DD
-_date_re = re.compile(r"^(\d\d\d\d)/(\d\d)/(\d\d)$")
+# FFFF007/8 stop accepting YYYY/MM/DD
+# Regular expression to match YYYY/MM/DD or YYYY-MM-DD
+_date_re = re.compile(r"^(\d\d\d\d)([/-])(\d\d)([/-])(\d\d)$")
 def _parseDate(s):
-    """Validation function.  Converts from YYYY/MM/DD format to a (long)
-       time value for midnight on that date."""
+    """Validation function.  Converts from YYYY-MM-DD or YYYY/MM/DD
+       format to a (long) time value for midnight on that date."""
     m = _date_re.match(s.strip())
+    if not m or m.group(2) != m.group(4):
+        raise ConfigError("Invalid date %r"%s)
     try:
-        yyyy, MM, dd = map(int, m.groups())
+        yyyy = int(m.group(1))
+        MM = int(m.group(3))
+        dd = int(m.group(5))
     except (ValueError,AttributeError):
         raise ConfigError("Invalid date %r"%s)
     if not ((1 <= dd <= 31) and (1 <= MM <= 12) and
             (1970 <= yyyy)):
-        raise ConfigError("Invalid date %s"%s)
+        raise ConfigError("Invalid date %r"%s)
     return calendar.timegm((yyyy,MM,dd,0,0,0,0,0,0))
 
+# FFFF007/8 stop accepting YYYY/MM/DD
 # Regular expression to match YYYY/MM/DD HH:MM:SS
-_time_re = re.compile(r"^(\d\d\d\d)/(\d\d)/(\d\d) (\d\d):(\d\d):(\d\d)$")
+_time_re = re.compile(r"^(\d\d\d\d)([/-])(\d\d)([/-])(\d\d)\s+"
+                      r"(\d\d):(\d\d):(\d\d)((?:\.\d\d\d)?)$")
 def _parseTime(s):
     """Validation function.  Converts from YYYY/MM/DD HH:MM:SS format
        to a (float) time value for GMT."""
     m = _time_re.match(s.strip())
-    if not m:
+    if not m or m.group(2) != m.group(4):
         raise ConfigError("Invalid time %r" % s)
 
-    yyyy, MM, dd, hh, mm, ss = map(int, m.groups())
+    yyyy = int(m.group(1))
+    MM = int(m.group(3))
+    dd = int(m.group(5))
+    hh = int(m.group(6))
+    mm = int(m.group(7))
+    ss = int(m.group(8))
+    if m.group(9):
+        fsec = float(m.group(9))
+    else:
+        fsec = 0.0
+
     if not ((1 <= dd <= 31) and (1 <= MM <= 12) and
             (1970 <= yyyy)  and (0 <= hh < 24) and
             (0 <= mm < 60)  and (0 <= ss <= 61)):
         raise ConfigError("Invalid time %r" % s)
 
-    return calendar.timegm((yyyy,MM,dd,hh,mm,ss,0,0,0))
+    return calendar.timegm((yyyy,MM,dd,hh,mm,ss,0,0,0))+fsec
 
 _NICKNAME_CHARS = ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"+
                    "abcdefghijklmnopqrstuvwxyz"+
-                   "0123456789_.@-")
+                   "0123456789-")
+_NICKNAME_INITIAL_CHARS = ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"+
+                           "abcdefghijklmnopqrstuvwxyz")
+
 MAX_NICKNAME = 128
 def _parseNickname(s):
-    """Validation function.  Returns true iff s contains a valoid
-       server nickname-- that is, a string of 1..128 characters,
-       containing only the characters [A-Za-z0-9_@], '.' or '-'.
+    """Validation function.  Returns true iff s contains a valid
+       server nickname -- that is, a string of 1..128 characters,
+       containing only the characters [A-Za-z0-9], or '-'.  It must
+       not begin with a digit or a '-'.
        """
     s = s.strip()
     bad = s.translate(mixminion.Common._ALLCHARS, _NICKNAME_CHARS)
@@ -336,6 +420,8 @@ def _parseNickname(s):
         raise ConfigError("Nickname is too long")
     elif len(s) == 0:
         raise ConfigError("Nickname is too short")
+    elif s[0] not in _NICKNAME_INITIAL_CHARS:
+        raise ConfigError("Nickname begins with invalid character %r" %s[0])
     return s
 
 def _parseFilename(s):
@@ -349,6 +435,17 @@ def _parseFilename(s):
 
     return os.path.expanduser(s)
 
+def _parseUser(s):
+    """Validation function.  Matches a username or UID.  Returns a UID."""
+    s = s.strip()
+    try:
+        return pwd.getpwnam(s)[2]
+    except (KeyError,AttributeError):
+        try:
+            return _parseInt(s)
+        except ConfigError:
+            raise ConfigError("Expected a user name or UID, but got %r"%s)
+
 #----------------------------------------------------------------------
 
 # Regular expression to match a section header.
@@ -411,6 +508,8 @@ def _readConfigFile(contents):
     return sections
 
 def _readRestrictedConfigFile(contents):
+    """Same interface as _readConfigFile, but only supports the restrictd
+       file format as used by directories and descriptors."""
     # List of (heading, [(key, val, lineno), ...])
     sections = []
     # [(key, val, lineno)] for the current section.
@@ -450,16 +549,16 @@ def _readRestrictedConfigFile(contents):
 
     return sections
 
-def _formatEntry(key,val,w=79,ind=4):
+def _formatEntry(key,val,w=79,ind=4,strict=0):
     """Helper function.  Given a key/value pair, returns a NL-terminated
        entry for inclusion in a configuration file, such that no line is
        avoidably longer than 'w' characters, and with continuation lines
        indented by 'ind' spaces.
     """
-    ind_s = " "*(ind-1)
-    if len(str(val))+len(key)+2 <= 79:
+    if strict or len(str(val))+len(key)+2 <= 79:
         return "%s: %s\n" % (key,val)
 
+    ind_s = " "*(ind-1)
     lines = [  ]
     linecontents = [ "%s:" % key ]
     linelength = len(linecontents[0])
@@ -475,6 +574,79 @@ def _formatEntry(key,val,w=79,ind=4):
     lines.append("") # so the last line ends with \n
     return "\n".join(lines)
 
+def resolveFeatureName(name, klass):
+    """Given a feature name and a subclass of _ConfigFile, check whether
+       the feature exists, and return a sec/name tuple that, when passed to
+       _ConfigFile.getFeature, gives the value of the appropriate feature.
+       Raises a UIError if the feature name is invalid.
+
+       A feature is either: a special string handled by the class (like
+       'caps' for ServerInfo), a special string handled outside the class
+       (like 'status' for ClientDirectory), a Section:Entry string, or an
+       Entry string.  (If the Entry string is not unique within a section,
+       raises UIError.)  All features are case-insensitive.
+
+       Example features are: 'caps', 'status', 'Incoming/MMTP:Version',
+         'hostname'.
+       """
+    syn = klass._syntax
+    name = name.lower()
+    if klass._features.has_key(name):
+        return "-", name
+    elif ':' in name:
+        idx = name.index(':')
+        sec, ent = name[:idx], name[idx+1:]
+        goodSection = None
+        for section, entries in syn.items():
+            if section.lower() == sec:
+                goodSection = section
+                for entry in entries.keys():
+                    if entry.lower() == ent:
+                        return section, entry
+        if goodSection:
+            raise UIError("Section %s has no entry %r"%(goodSection,ent))
+        else:
+            raise UIError("No such section as %s"%sec)
+    else:
+        result =  []
+        for secname, secitems in syn.items():
+            if secname.lower() == name:
+                raise UIError("No key given for section %s"%secname)
+            for entname in secitems.keys():
+                if entname.lower() == name:
+                    result.append((secname, entname))
+        if len(result) == 0:
+            raise UIError("No key named %r found"%name)
+        elif len(result) > 1:
+            secs = [ "%s:%s"%(secname,entname) for secname,entname
+                     in result ]
+            raise UIError("%r is ambiguous.  Did you mean %s?"%(
+                          name, englishSequence(secs,compound="or")))
+        else:
+            return result[0]
+
+def getFeatureList(klass):
+    """Get a list of all feature names from the _ConfigFile subclass
+       'klass'.  Return a list of tuples, each of which contains all the
+       synonyms for a single feature."""
+    syn = klass._syntax
+    features = []
+    for secname, secitems in syn.items():
+        for entname in secitems.keys():
+            if entname.startswith("__"): continue
+            synonyms = []
+            synonyms.append("%s:%s"%(secname,entname))
+            unique = 1
+            for sn, si in syn.items():
+                if sn != secname and si.has_key(entname):
+                    unique = 0
+                    break
+            if unique:
+                synonyms.append(entname)
+            features.append(tuple(synonyms))
+    features.sort()
+    return features
+
 class _ConfigFile:
     """Base class to parse, validate, and represent configuration files.
     """
@@ -491,12 +663,16 @@ class _ConfigFile:
     # Fields to be set by a subclass:
     #     _syntax is map from sec->{key:
     #                               (ALLOW/REQUIRE/ALLOW*/REQUIRE*,
-    #                                 parseFn,
+    #                                 type,
     #                                 default, ) }
     #     _restrictFormat is 1/0: do we allow full RFC822ness, or do
     #         we insist on a tight data format?
     #     _restrictKeys is 1/0: do we raise a ConfigError when we see an
     #         unrecognized key, or do we simply generate a warning?
+    #     _restrictSections is 1/0: do we raise a ConfigError when we see an
+    #         unrecognized section, or do we simply generate a warning?
+    #     _features is a map from lowercase feature name to 1 for
+    #         features that should be handled by getFeature.
 
     ## Validation rules:
     # A key without a corresponding entry in _syntax gives an error.
@@ -510,9 +686,35 @@ class _ConfigFile:
     #   the entry's value will be set to default.  Otherwise, the value
     #   will be set to None.
 
+    CODING_FNS = {
+        "boolean" :  (_parseBoolean, lambda b: b and "yes" or "no"),
+        "severity" : (_parseSeverity, str),
+        "serverMode"  : (_parseServerMode, str),
+        "interval" : (_parseInterval, str),
+        "intervalList" : (_parseIntervalList, _unparseIntervalList),
+        "int" : (_parseInt, str),
+        "size" : (_parseSize, _unparseSize),
+        "IP" : (_parseIP, str),
+        "IP6" : (_parseIP6, str),
+        "host" : (_parseHost, str),
+        "addressSet_allow" : (_parseAddressSet_allow, str), #XXXX
+        "addressSet_deny" : (_parseAddressSet_deny, str), #XXXX
+        "command" : (_parseCommand, lambda c,o: " ".join([c," ".join(o)])),
+        "base64" : (_parseBase64, mixminion.Common.formatBase64),
+        "hex" : (_parseHex, binascii.b2a_hex),
+        "publicKey" : (_parsePublicKey, lambda r: "<public key>"),
+        "date" : (_parseDate, mixminion.Common.formatDate),
+        "time" : (_parseTime, mixminion.Common.formatTime),
+        "nickname" : (_parseNickname, str),
+        "filename" : (_parseFilename, str),
+        "user" : (_parseUser, str),
+        }
+
     _syntax = None
+    _features = {}
     _restrictFormat = 0
     _restrictKeys = 1
+    _restrictSections = 1
 
     def __init__(self, filename=None, string=None, assumeValid=0):
         """Create a new _ConfigFile.  If <filename> is set, read from
@@ -571,22 +773,35 @@ class _ConfigFile:
             secConfig = self._syntax.get(secName)
 
             if not secConfig:
-                LOG.warn("Skipping unrecognized section %s", secName)
-                continue
+                if self._restrictSections:
+                    raise ConfigError("Skipping unrecognized section %s"
+                                      %secName)
+                else:
+                    LOG.warn("Skipping unrecognized section %s", secName)
+                    continue
 
             # Set entries from the section, searching for bad entries
             # as we go.
             for k,v,line in secEntries:
                 try:
-                    rule, parseFn, default = secConfig[k]
+                    rule, parseType, default = secConfig[k]
                 except KeyError:
+                    msg = "Unrecognized key %s on line %s"%(k,line)
+                    acceptedIn = [ sn for sn,sc in self._syntax.items()
+                                   if sc.has_key(k) ]
+                    acceptedIn.sort()
+                    if acceptedIn:
+                        msg += ". This key belongs in %s, but appears in %s."%(
+                            englishSequence(acceptedIn, compound="or"),
+                            secName)
                     if self._restrictKeys:
-                        raise ConfigError("Unrecognized key %s on line %s" %
-                                          (k, line))
+                        raise ConfigError(msg)
                     else:
-                        LOG.warn("Unrecognized key %s on line %s", k, line)
+                        LOG.warn(msg)
                         continue
 
+                parseFn, _ = self.CODING_FNS.get(parseType,(None,None))
+
                 # Parse and validate the value of this entry.
                 if parseFn is not None:
                     try:
@@ -614,7 +829,7 @@ class _ConfigFile:
 
             # Check for missing entries, setting defaults and detecting
             # missing requirements as we go.
-            for k, (rule, parseFn, default) in secConfig.items():
+            for k, (rule, parseType, default) in secConfig.items():
                 if k == '__SECTION__':
                     continue
                 elif not section.has_key(k):
@@ -622,6 +837,7 @@ class _ConfigFile:
                         raise ConfigError("Missing entry %s from section %s"
                                           % (k, secName))
                     else:
+                        parseFn, _ = self.CODING_FNS.get(parseType,(None,None))
                         if parseFn is None or default is None:
                             if rule == 'ALLOW*':
                                 section[k] = []
@@ -665,6 +881,18 @@ class _ConfigFile:
         """
         return contents
 
+    def getFeature(self,sec,name):
+        """Given a sec/name pair returned by resolveFeatureName, return a
+           string value of that feature for the class."""
+        assert sec not in ("+","-")
+        parseType = self._syntax[sec].get(name)[1]
+        _, unparseFn = self.CODING_FNS.get(parseType, (None,str))
+        try:
+            v = self[sec][name]
+        except KeyError:
+            return "<none>"
+        return unparseFn(v)
+
     def validate(self, entryLines, fileContents):
         """Check additional semantic properties of a set of configuration
            data before overwriting old data.  Subclasses should override."""
@@ -677,6 +905,12 @@ class _ConfigFile:
            section was absent, return an empty map."""
         return self._sections[sec]
 
+    def get(self, sec, val="---"):
+        """Return a section named sec, if any such section exists.  Otherwise
+           return an empty dict, or 'val' if provided."""
+        if val == "---": val = {}
+        return self._sections.get(sec, val)
+
     def has_section(self, sec):
         """Return true if this config object allows a section named 'sec'."""
         return self._sections.has_key(sec)
@@ -693,37 +927,69 @@ class _ConfigFile:
         for s in self._sectionNames:
             lines.append("[%s]\n"%s)
             for k,v in self._sectionEntries[s]:
-                lines.append(_formatEntry(k,v))
-            lines.append("\n")
+                tp = self._syntax[s][k][1]
+                if tp:
+                    v = self.CODING_FNS[tp][1](v)
+                lines.append(_formatEntry(k,v,strict=self._restrictFormat))
+            if not self._restrictFormat:
+                lines.append("\n")
 
         return "".join(lines)
 
+
+if sys.platform == 'win32':
+    # Windows prefers to put configuration in different places, depending
+    # on your version, but it doesn't get the idea of dotfiles.
+    DEFAULT_USER_DIR = "~/mixminion"
+else:
+    # Unix prefers to put configuration in hidden directories in your homedir.
+    DEFAULT_USER_DIR = "~/.mixminion"
+
 class ClientConfig(_ConfigFile):
+    #XXXX Should this go into ClientUtils or something?
     _restrictFormat = 0
+    _restrictKeys = _restrictSections = 1
     _syntax = {
         'Host' : { '__SECTION__' : ('ALLOW', None, None),
-                   'ShredCommand': ('ALLOW', _parseCommand, None),
-                   'EntropySource': ('ALLOW', _parseFilename, "/dev/urandom"),
-                   'TrustedUser': ('ALLOW*', None, None),
+                   'ShredCommand': ('ALLOW', "command", None),
+                   'EntropySource': ('ALLOW', "filename", "/dev/urandom"),
+                   'TrustedUser': ('ALLOW*', "user", None),
+                   'FileParanoia': ('ALLOW', "boolean", "yes"),
                    },
         'DirectoryServers' :
                    { '__SECTION__' : ('REQUIRE', None, None),
                      'ServerURL' : ('ALLOW*', None, None),
-                     'MaxSkew' : ('ALLOW', _parseInterval, "10 minutes") },
-        'User' : { 'UserDir' : ('ALLOW', _parseFilename, "~/.mixminion" ) },
-        'Security' : { 'PathLength' : ('ALLOW', _parseInt, "8"),
+                     'MaxSkew' : ('ALLOW', "interval", "10 minutes"),
+                     'DirectoryTimeout' : ('ALLOW', "interval", "1 minute") },
+        'User' : { 'UserDir' : ('ALLOW', "filename", DEFAULT_USER_DIR) },
+        'Security' : { 'PathLength' : ('ALLOW', "int", "8"),
                        'SURBAddress' : ('ALLOW', None, None),
-                       'SURBPathLength' : ('ALLOW', _parseInt, "4"),
-                       'SURBLifetime' : ('ALLOW', _parseInterval, "7 days"),
-                       'ForwardPath' : ('ALLOW', None, "*"),
-                       'ReplyPath' : ('ALLOW', None, "*"),
-                       'SURBPath' : ('ALLOW', None, "*"),
+                       'SURBPathLength' : ('ALLOW', "int", "4"),
+                       'SURBLifetime' : ('ALLOW', "interval", "7 days"),
+                       'ForwardPath' : ('ALLOW', None, "*6"),
+                       'ReplyPath' : ('ALLOW', None, "*4"),
+                       'SURBPath' : ('ALLOW', None, "*4"),
                        },
-        'Network' : { 'ConnectionTimeout' : ('ALLOW', _parseInterval, None) }
+        'Network' : { 'ConnectionTimeout' : ('ALLOW', "interval", "2 minutes")}
         }
     def __init__(self, fname=None, string=None):
         _ConfigFile.__init__(self, fname, string)
 
+
+    def prevalidate(self, contents):
+        # See if we've been passed a server configuration.
+        foundServer = 0
+        foundUser = 0
+        for s, _ in contents:
+            if s == 'Server':
+                foundServer = 1
+            elif s == 'User':
+                foundUser = 1
+        if foundServer and not foundUser:
+            raise ConfigError("Got a server configuration (mixminiond.conf), but expected a client configuration (.mixminionrc)")
+
+        return contents
+
     def validate(self, lines, contents):
         _validateHostSection(self['Host'])
 
