@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: Crypto.py,v 1.45 2003/06/05 18:41:40 nickm Exp $
+# $Id: Crypto.py,v 1.59 2003/12/04 05:53:13 nickm Exp $
 """mixminion.Crypto
 
    This package contains all the cryptographic primitives required
@@ -11,6 +11,7 @@
 import binascii
 import copy_reg
 import errno
+import math
 import os
 import stat
 import sys
@@ -28,6 +29,7 @@ __all__ = [ 'AESCounterPRNG', 'CryptoError', 'Keyset', 'bear_decrypt',
             'pk_encode_public_key', 'pk_encrypt', 'pk_fingerprint',
             'pk_from_modulus', 'pk_generate', 'pk_get_modulus',
             'pk_same_public_key', 'pk_sign', 'prng', 'sha1', 'strxor', 'trng',
+            'unwhiten', 'whiten',
             'AES_KEY_LEN', 'DIGEST_LEN', 'HEADER_SECRET_MODE', 'PRNG_MODE',
             'RANDOM_JUNK_MODE', 'HEADER_ENCRYPT_MODE', 'APPLICATION_KEY_MODE',
             'PAYLOAD_ENCRYPT_MODE', 'HIDE_HEADER_MODE' ]
@@ -52,7 +54,7 @@ def init_crypto(config=None):
         raise
     except:
         info = sys.exc_info()
-        raise MixFatalError("Error initializing entropy source: %s", info[0])
+        raise MixFatalError("Error initializing entropy source: %s" % info[1])
     openssl_seed(40)
 
 def sha1(s):
@@ -181,6 +183,20 @@ def bear_decrypt(s,(key1,key2)):
     left = _ml.strxor(left, _ml.sha1("".join((key1,right,key1))))
     return left + right
 
+def whiten(s):
+    """Return a whitened version of a string 's', using the whitening
+       algorithm from 'E2E-spec.txt'.
+
+       The functions 'unwhiten' inverts 'whiten', but if any portion of
+       whiten(s) is not known, no part of 's' can be recovered."""
+    keys = Keyset("WHITEN").getLionessKeys("WHITEN")
+    return lioness_encrypt(s, keys)
+
+def unwhiten(s):
+    """Given a whitened string, return the original string."""
+    keys = Keyset("WHITEN").getLionessKeys("WHITEN")
+    return lioness_decrypt(s, keys)
+
 def openssl_seed(count):
     """Seeds the openssl rng with 'count' bytes of real entropy."""
     _ml.openssl_seed(trng(count))
@@ -189,6 +205,8 @@ def trng(count):
     """Returns (count) bytes of true random data from a true source of
        entropy (/dev/urandom).  May read ahead and cache values.
     """
+    if _theTrueRNG is None:
+        configure_trng(None)
     return _theTrueRNG.getBytes(count)
 
 # Specified in the Mixminion spec.  It's a Thomas Paine quotation.
@@ -273,8 +291,8 @@ def pk_fingerprint(key):
 def pk_PEM_save(rsa, filename, password=None):
     """Save a PEM-encoded private key to a file.  If <password> is provided,
        encrypt the key using the password."""
-    fd = os.open(filename, os.O_WRONLY|os.O_CREAT, 0600)
-    f = os.fdopen(fd, 'wb')
+    fd = os.open(filename, os.O_WRONLY|os.O_CREAT,0600)
+    f = os.fdopen(fd, 'w')
     if password:
         rsa.PEM_write_key(f, 0, password)
     else:
@@ -316,7 +334,7 @@ copy_reg.pickle(_ml.RSA, _pickle_rsa, _ml.rsa_decode_key)
 #       ftp://ftp.rsasecurity.com/pub/rsalabs/rsa_algorithm/rsa-oaep_spec.pdf)
 
 def _oaep_mgf(seed, bytes):
-    ''' Mask generation function specified for RAESA-OAEP.  Given a seed
+    ''' Mask generation function specified for RSA-OAEP.  Given a seed
         and a number of bytes, generates a mask for OAEP by computing
         sha1(seed + "\x00\x00\x00\x00")+sha1(seed+"\x00\x00\x00\x01)+...
 
@@ -383,7 +401,7 @@ def _check_oaep_padding(data, p, bytes):
         raise CryptoError("Decoding error")
     return m
 
-# Use the fastest implementaiton of OAEP we have.
+# Use the fastest implementation of OAEP we have.
 if hasattr(_ml, 'check_oaep_padding'):
     check_oaep = _ml.check_oaep_padding
     add_oaep = _ml.add_oaep_padding
@@ -473,6 +491,12 @@ def lioness_keys_from_header(header2):
 # we assert it.
 assert sys.maxint >= 0x7fffffff
 
+# Magic number used for normal distribution
+NV_MAGICCONST = 4 * math.exp(-0.5)/math.sqrt(2.0)
+
+# Flag: is the filesystem case-insensitive?
+FS_IS_CASEI = sys.platform in ('cygwin', 'win32')
+
 class RNG:
     '''Base implementation class for random number generators.  Works
        by requesting a bunch of bytes via self._prng, and doling them
@@ -485,6 +509,7 @@ class RNG:
 
     def getBytes(self, n):
         """Returns a string of 'n' random bytes."""
+        assert n >= 0
 
         if n > len(self.bytes):
             # If we don't have enough bytes, fetch enough so that we'll have
@@ -544,6 +569,7 @@ class RNG:
 
         assert 0 < max < 0x3fffffff
         _ord = ord
+        cutoff = 0x7fffffff - (0x7fffffff % max)
         while 1:
             # Get a random positive int between 0 and 0x7fffffff.
             b = self.getBytes(4)
@@ -553,9 +579,22 @@ class RNG:
                        _ord(b[3]))
             # Retry if we got a value that would fall in an incomplete
             # run of 'max' elements.
-            if 0x7fffffff - max >= o:
+            if o < cutoff:
                 return o % max
 
+    def getNormal(self, m, s):
+        """Return a random value with mean m and standard deviation s.
+        """
+        # Lifted from random.py in standard python dist.
+        while 1:
+            u1 = self.getFloat()
+            u2 = 1.0 - self.getFloat()
+            z = NV_MAGICCONST*(u1-0.5)/u2
+            zz = z*z/4.0
+            if zz <= -math.log(u2):
+                break
+        return m + z*s
+
     def getFloat(self):
         """Return a floating-point number between 0 and 1."""
         b = self.getBytes(4)
@@ -565,7 +604,7 @@ class RNG:
         #return o / float(0x7fffffff)
         return o / 2147483647.0
 
-    def openNewFile(self, dir, prefix="", binary=1):
+    def openNewFile(self, dir, prefix="", binary=1, conflictPrefix=None):
         """Generate a new random filename within a directory with a given
            prefix within a directory, and open a new file within the directory
            with that filename.  Return 2-tuple of a file object and the
@@ -573,6 +612,9 @@ class RNG:
 
            Random portions are generated by choosing 8 random characters
            from the set 'A-Za-z0-9+-'.
+
+           If 'conflictPrefix' is set, do not return any file named
+           prefix+H if a file named conflictPrefix+H already exists.
            """
         flags = os.O_WRONLY|os.O_CREAT|os.O_EXCL
         mode = "w"
@@ -582,7 +624,12 @@ class RNG:
         while 1:
             bytes = self.getBytes(6)
             base = binascii.b2a_base64(bytes).strip().replace("/","-")
+            if FS_IS_CASEI:
+                base = base.lower()
             fname = os.path.join(dir, "%s%s"%(prefix,base))
+            if conflictPrefix and os.path.exists(os.path.join(
+                dir,conflictPrefix+base)):
+                continue
             try:
                 fd = os.open(fname, flags, 0600)
                 return os.fdopen(fd, mode), base
@@ -654,6 +701,18 @@ def configure_trng(config):
     """Initialize the true entropy source from a given Config object.  If
        none is provided, tries some sane defaults."""
     global _TRNG_FILENAME
+    global _theTrueRNG
+
+    if sys.platform == 'win32':
+        # We have two entropy sources on windows: openssl's built-in
+        # entropy generator that takes data from the screen, and
+        # Windows's CryptGenRandom function.  Because the former is
+        # insecure, and the latter is closed-source, we xor them.
+        _ml.win32_openssl_seed()
+        _ml.openssl_seed(_ml.win32_get_random_bytes(32))
+        _theTrueRNG = _XorRNG(_OpensslRNG(), _WinTrueRNG())
+        return
+
     if config is not None:
         requestedFile = config['Host'].get('EntropySource')
     else:
@@ -671,10 +730,10 @@ def configure_trng(config):
         if file is None:
             continue
 
-        verbose = 1#(file == requestedFile)
+        verbose = (file == requestedFile)
         if not os.path.exists(file):
             if verbose:
-                LOG.error("No such file as %s", file)
+                LOG.warn("No such file as %s", file)
         else:
             st = os.stat(file)
             if not (st[stat.ST_MODE] & stat.S_IFCHR):
@@ -686,7 +745,8 @@ def configure_trng(config):
                 break
 
     if randFile is None and _TRNG_FILENAME is None:
-        LOG.fatal("No entropy source available")
+        LOG.fatal("No entropy source available: Tried all of %s",
+                  files)
         raise MixFatalError("No entropy source available")
     elif randFile is None:
         LOG.warn("Falling back to previous entropy source %s",
@@ -694,6 +754,11 @@ def configure_trng(config):
     else:
         LOG.info("Setting entropy source to %r", randFile)
         _TRNG_FILENAME = randFile
+        _theTrueRNG = _TrueRNG(1024)
+
+
+# Global TRN instance, for use by trng().
+_theTrueRNG = None
 
 class _TrueRNG(RNG):
     '''Random number generator that yields pieces of entropy from
@@ -722,10 +787,36 @@ class _TrueRNG(RNG):
         self.__lock.release()
         return b
 
-# Global _TrueRNG instance, for use by trng().
-_theTrueRNG = _TrueRNG(1024)
+if hasattr(_ml, "win32_get_random_bytes"):
+    class _WinTrueRNG(RNG):
+        """A random number generator using the windows crypto API."""
+        def __init__(self):
+            RNG.__init__(self, 1024)
+            self.getBytes(1)
+        def _prng(self,n):
+            return _ml.win32_get_random_bytes(n)
+
+class _OpensslRNG(RNG):
+    """Random number generator that falls back to openssl's implementation."""
+    def __init__(self):
+        RNG.__init__(self, 1024)
+        self.getBytes(1)
+    def _prng(self,n):
+        return _ml.openssl_rand(n)
+
+class _XorRNG(RNG):
+    """A random number generator that takes random data from two sources
+       and XORs the two streams together."""
+    def __init__(self, base1, base2):
+        RNG.__init__(self, 1024)
+        self.base1 = base1
+        self.base2 = base2
+    def _prng(self, n):
+        return strxor(self.base1.getBytes(n), self.base2.getBytes(n))
 
 # Return the shared instance of the true RNG.
 def getTrueRNG():
     """Return the shared instance of the true RNG."""
+    if _theTrueRNG is None:
+        configure_trng(None)
     return _theTrueRNG
