@@ -1,31 +1,29 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: MMTPClient.py,v 1.33 2003/06/03 17:28:11 nickm Exp $
+# $Id: MMTPClient.py,v 1.45 2003/12/08 02:22:56 nickm Exp $
 """mixminion.MMTPClient
 
    This module contains a single, synchronous implementation of the client
    side of the Mixminion Transfer protocol.  You can use this client to
-   upload messages to any conforming Mixminion server.
+   upload packets to any conforming Mixminion server.
 
-   (We don't use this module for tranferring packets between servers;
+   (We don't use this module for transferring packets between servers;
    in fact, MMTPServer makes it redundant.  We only keep this module
    around [A] so that clients have an easy (blocking) interface to
-   introduce messages into the system, and [B] so that we've got an
+   introduce packets into the system, and [B] so that we've got an
    easy-to-verify reference implementation of the protocol.)
    """
 
-__all__ = [ "BlockingClientConnection", "sendMessages" ]
+__all__ = [ "BlockingClientConnection", "sendPackets" ]
 
-import errno
-import signal
 import socket
 import mixminion._minionlib as _ml
+import mixminion.NetUtils
+import mixminion.Packet
+import mixminion.ServerInfo
 from mixminion.Crypto import sha1, getCommonPRNG
 from mixminion.Common import MixProtocolError, MixProtocolReject, \
-     MixProtocolBadAuth, LOG, MixError, formatBase64
-
-class TimeoutError(MixProtocolError):
-    """Exception raised for protocol timeout."""
-    pass
+     MixProtocolBadAuth, LOG, MixError, formatBase64, TimeoutError
+from mixminion.Packet import IPV4Info, MMTPHostInfo
 
 class BlockingClientConnection:
     """A BlockingClientConnection represents a MMTP connection to a single
@@ -37,6 +35,7 @@ class BlockingClientConnection:
     # targetKeyID -- sha1 hash of the ASN1 encoding of the public key we
     #   expect the server to use, or None if we don't care.
     # context: a TLSContext object; used to create connections.
+    # serverName: The name of the server to display in log messages.
     # sock: a TCP socket, open to the server.
     # tls: a TLS socket, wrapping sock.
     # protocol: The MMTP protocol version we're currently using, or None
@@ -44,9 +43,11 @@ class BlockingClientConnection:
     # PROTOCOL_VERSIONS: (static) a list of protocol versions we allow,
     #     in decreasing order of preference.
     PROTOCOL_VERSIONS = ['0.3']
-    def __init__(self, targetIP, targetPort, targetKeyID):
+    def __init__(self, targetFamily, targetAddr, targetPort, targetKeyID,
+                 serverName=None):
         """Open a new connection."""
-        self.targetIP = targetIP
+        self.targetFamily = targetFamily
+        self.targetAddr = targetAddr
         self.targetPort = targetPort
         if targetKeyID != '\x00' *20:
             self.targetKeyID = targetKeyID
@@ -56,6 +57,11 @@ class BlockingClientConnection:
         self.tls = None
         self.sock = None
         self.certCache = PeerCertificateCache()
+        if serverName:
+            self.serverName = serverName
+        else:
+            self.serverName = mixminion.ServerInfo.displayServer(
+                mixminion.Packet.IPV4Info(targetAddr, targetPort, targetKeyID))
 
     def connect(self, connectTimeout=None):
         """Connect to the server, perform the TLS handshake, check the server
@@ -67,70 +73,61 @@ class BlockingClientConnection:
            errors."""
         try:
             self._connect(connectTimeout)
-        except (socket.error, _ml.TLSError, _ml.TLSClosed), e:
+        except (socket.error, _ml.TLSError, _ml.TLSClosed,
+                _ml.TLSWantRead, _ml.TLSWantWrite), e:
             self._raise(e, "connecting")
 
     def _raise(self, err, action):
         """Helper method: given an exception (err) and an action string (e.g.,
            'connecting'), raises an appropriate MixProtocolError.
         """
+        errstr = str(err)
         if isinstance(err, socket.error):
             tp = "Socket"
+            if mixminion.NetUtils.exceptionIsTimeout(err):
+                tp = "Timeout"
         elif isinstance(err, _ml.TLSError):
             tp = "TLS"
+            if errstr == 'wrong version number':
+                errstr = 'wrong version number (or failed handshake)'
         elif isinstance(err, _ml.TLSClosed):
             tp = "TLSClosed"
+        elif isinstance(err, _ml.TLSWantRead):
+            tp = "Unexpected TLSWantRead"
+        elif isinstance(err, _ml.TLSWantWrite):
+            tp = "Unexpected TLSWantWrite"
         else:
             tp = str(type(err))
-        e = MixProtocolError("%s error while %s to %s:%s: %s",
-                             tp, action, self.targetIP, self.targetPort, err)
+        e = MixProtocolError("%s error while %s to %s: %s" %(
+                             tp, action, self.serverName, errstr))
         e.base = err
         raise e
 
     def _connect(self, connectTimeout=None):
         """Helper method; implements _connect."""
-        # FFFF There should be a way to specify timeout for communication.
-        def sigalarmHandler(sig, _):
-            assert sig == signal.SIGALRM
-        if connectTimeout:
-            signal.signal(signal.SIGALRM, sigalarmHandler)
-
         # Connect to the server
-        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        self.sock = socket.socket(self.targetFamily, socket.SOCK_STREAM)
         self.sock.setblocking(1)
-        LOG.debug("Connecting to %s:%s", self.targetIP, self.targetPort)
+        LOG.debug("Connecting to %s", self.serverName)
 
         # Do the TLS handshaking
-        if connectTimeout:
-            signal.alarm(connectTimeout)
-        try:
-            try:
-                self.sock.connect((self.targetIP,self.targetPort))
-            except socket.error, e:
-                if e[0] == errno.EINTR:
-                    raise TimeoutError("Connection timed out")
-                else:
-                    raise MixProtocolError("Error connecting: %s" % e)
-        finally:
-            if connectTimeout:
-                signal.alarm(0)
+        mixminion.NetUtils.connectWithTimeout(
+            self.sock, (self.targetAddr, self.targetPort), connectTimeout)
 
-        LOG.debug("Handshaking with %s:%s",self.targetIP, self.targetPort)
+        LOG.debug("Handshaking with %s:", self.serverName)
         self.tls = self.context.sock(self.sock.fileno())
         self.tls.connect()
         LOG.debug("Connected.")
-
         # Check the public key of the server to prevent man-in-the-middle
         # attacks.
-        self.certCache.check(self.tls, self.targetKeyID,
-                             "%s:%s"%(self.targetIP,self.targetPort))
+        self.certCache.check(self.tls, self.targetKeyID, self.serverName)
 
         ####
         # Protocol negotiation
         # For now, we only support 1.0, but we call it 0.3 so we can
         # change our mind between now and a release candidate, and so we
         # can obsolete betas come release time.
-        LOG.debug("Negotiatiating MMTP protocol")
+        LOG.debug("Negotiating MMTP protocol")
         self.tls.write("MMTP %s\r\n" % ",".join(self.PROTOCOL_VERSIONS))
         # This is ugly, but we have no choice if we want to read up to the
         # first newline.
@@ -147,7 +144,8 @@ class BlockingClientConnection:
                 break
         if not self.protocol:
             raise MixProtocolError("Protocol negotiation failed")
-        LOG.debug("MMTP protocol negotated: version %s", self.protocol)
+        LOG.debug("MMTP protocol negotiated with %s: version %s",
+                  self.serverName, self.protocol)
 
     def renegotiate(self):
         """Re-do the TLS handshake to renegotiate a new connection key."""
@@ -173,7 +171,7 @@ class BlockingClientConnection:
         """Helper method: implements sendPacket and sendJunkPacket.
               packet -- a 32K string to send
               control -- a 6-character string ending with CRLF to
-                  indicate the type of message we're sending.
+                  indicate the type of packet we're sending.
               serverControl -- a 10-character string ending with CRLF that
                   we expect to receive if we've sent correctly.
               hashExtra -- a string to append to the packet when computing
@@ -186,9 +184,8 @@ class BlockingClientConnection:
         try:
             ##
             # We write: "SEND\r\n", 28KB of data, and sha1(packet|"SEND").
-            self.tls.write(control)
-            self.tls.write(packet)
-            self.tls.write(sha1(packet+hashExtra))
+            written = control+packet+sha1(packet+hashExtra)
+            self.tls.write(written)
             LOG.debug("Packet sent; waiting for ACK")
 
             # And we expect, "RECEIVED\r\n", and sha1(packet|"RECEIVED")
@@ -199,27 +196,29 @@ class BlockingClientConnection:
                 LOG.warn("Received bad ACK from server")
                 raise MixProtocolError("Bad ACK received")
             LOG.debug("ACK received; packet successfully delivered")
-        except (socket.error, _ml.TLSError, _ml.TLSClosed), e:
+        except (socket.error, _ml.TLSError, _ml.TLSClosed, _ml.TLSWantRead,
+                _ml.TLSWantWrite, _ml.TLSClosed), e:
             self._raise(e, "sending packet")
 
     def shutdown(self):
         """Close this connection."""
-        LOG.debug("Shutting down connection to %s:%s",
-                       self.targetIP, self.targetPort)
+        LOG.debug("Shutting down connection to %s", self.serverName)
         try:
             if self.tls is not None:
                 self.tls.shutdown()
             if self.sock is not None:
                 self.sock.close()
-        except (socket.error, _ml.TLSError, _ml.TLSClosed), e:
+        except (socket.error, _ml.TLSError, _ml.TLSClosed, _ml.TLSWantRead,
+                _ml.TLSWantWrite, _ml.TLSClosed), e:
             self._raise(e, "closing connection")
         LOG.debug("Connection closed")
 
-def sendMessages(routing, packetList, connectTimeout=None, callback=None):
-    """Sends a list of messages to a server.  Raise MixProtocolError on
+def sendPackets(routing, packetList, connectTimeout=None, callback=None):
+    """Sends a list of packets to a server.  Raise MixProtocolError on
        failure.
 
-       routing -- an instance of mixminion.Packet.IPV4Info.
+       routing -- an instance of mixminion.Packet.IPV4Info or
+                  mixminion.Packet.MMTPHostInfo.
                   If routing.keyinfo == '\000'*20, we ignore the server's
                   keyid.
        packetList -- a list of 32KB packets and control strings.  Control
@@ -238,9 +237,22 @@ def sendMessages(routing, packetList, connectTimeout=None, callback=None):
         elif p == 'RENEGOTIATE':
             packets.append(("RENEGOTIATE", None))
         else:
-            packets.append(("MSG", p))
+            packets.append(("PKT", p))
 
-    con = BlockingClientConnection(routing.ip,routing.port,routing.keyinfo)
+    serverName = mixminion.ServerInfo.displayServer(routing)
+
+    if isinstance(routing, IPV4Info):
+        family, addr = socket.AF_INET, routing.ip
+    else:
+        assert isinstance(routing, MMTPHostInfo)
+        LOG.trace("Looking up %s...",routing.hostname)
+        family, addr, _ = mixminion.NetUtils.getIP(routing.hostname)
+        if family == "NOENT":
+            raise MixProtocolError("Couldn't resolve hostname %s: %s" % (
+                                   routing.hostname, addr))
+
+    con = BlockingClientConnection(family,addr,routing.port,routing.keyinfo,
+                                   serverName=serverName)
     try:
         con.connect(connectTimeout=connectTimeout)
         for idx in xrange(len(packets)):
@@ -250,7 +262,7 @@ def sendMessages(routing, packetList, connectTimeout=None, callback=None):
             elif t == "RENEGOTIATE":
                 con.renegotiate()
             else:
-                con.sendPacket(p)
+                con.sendPacket(str(p))
             if callback is not None:
                 callback(idx)
     finally:
@@ -261,27 +273,29 @@ def pingServer(routing, connectTimeout=5):
 
        May raise MixProtocolBadAuth, or other MixProtocolError if server
        isn't up."""
-    sendMessages(routing, ["JUNK"], connectTimeout=connectTimeout)
+    sendPackets(routing, ["JUNK"], connectTimeout=connectTimeout)
 
 class PeerCertificateCache:
     """A PeerCertificateCache validates certificate chains from MMTP servers,
        and remembers which chains we've already seen and validated."""
-    ## Fieleds
+    ## Fields
     # cache: A map from peer (temporary) KeyID's to a (signing) KeyID.
     def __init__(self):
         self.cache = {}
 
-
-    def check(self, tls, targetKeyID, address):
+    def check(self, tls, targetKeyID, serverName):
         """Check whether the certificate chain on the TLS connection 'tls'
            is valid, current, and matches the keyID 'targetKeyID'.  If so,
-           return.  If not, raise MixProtocolBadAuth.
+           return.  If not, raise MixProtocolBadAuth.  Display all messages
+           using the server 'serverName'.
         """
+
         # First, make sure the certificate is neither premature nor expired.
         try:
             tls.check_cert_alive()
         except _ml.TLSError, e:
-            raise MixProtocolBadAuth("Invalid certificate: %s", str(e))
+            raise MixProtocolBadAuth("Invalid certificate from %s: %s" % (
+                serverName, str(e)))
 
         # If we don't care whom we're talking to, we don't need to check
         # them out.
@@ -290,6 +304,7 @@ class PeerCertificateCache:
 
         # Get the KeyID for the peer (temporary) key.
         hashed_peer_pk = sha1(tls.get_peer_cert_pk().encode_key(public=1))
+
         # Before 0.0.4alpha, a server's keyID was a hash of its current
         # TLS public key.  In 0.0.4alpha, we allowed this for backward
         # compatibility.  As of 0.0.4alpha2, since we've dropped backward
@@ -297,15 +312,13 @@ class PeerCertificateCache:
         # compatibility as well.
         if targetKeyID == hashed_peer_pk:
             raise MixProtocolBadAuth(
-               "Pre-0.0.4 (non-rotatable) certificate from server at %s",
-               address)
+               "Pre-0.0.4 (non-rotatable) certificate from %s" % serverName)
 
         try:
             if targetKeyID == self.cache[hashed_peer_pk]:
                 # We recognize the key, and have already seen it to be
                 # signed by the target identity.
-                LOG.trace("Got a cached certificate from server at %s",
-                          address)
+                LOG.trace("Got a cached certificate from %s", serverName)
                 return # All is well.
             else:
                 # We recognize the key, but some other identity signed it.
@@ -319,13 +332,12 @@ class PeerCertificateCache:
         try:
             identity = tls.verify_cert_and_get_identity_pk()
         except _ml.TLSError, e:
-            raise MixProtocolBadAuth("Invalid KeyID from server at %s: %s"
-                                   %(address, e))
+            raise MixProtocolBadAuth("Invalid KeyID (allegedly) from %s: %s"
+                                   %serverName)
 
         # Okay, remember who has signed this certificate.
         hashed_identity = sha1(identity.encode_key(public=1))
-        LOG.trace("Remembering valid certificate for server at %s",
-                  address)
+        LOG.trace("Remembering valid certificate for %s", serverName)
         self.cache[hashed_peer_pk] = hashed_identity
 
         # Note: we don't need to worry about two identities signing the
@@ -336,4 +348,4 @@ class PeerCertificateCache:
 
         # Was the signer the right person?
         if hashed_identity != targetKeyID:
-            raise MixProtocolBadAuth("Invalid KeyID for server at %s" %address)
+            raise MixProtocolBadAuth("Invalid KeyID for %s" % serverName)
