@@ -1,6 +1,6 @@
 #!/usr/bin/python2
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: Main.py,v 1.49 2003/06/06 07:17:35 nickm Exp $
+# $Id: Main.py,v 1.66 2003/12/14 01:43:38 nickm Exp $
 
 #"""Code to correct the python path, and multiplex between the various
 #   Mixminion CLIs.
@@ -115,11 +115,11 @@ _COMMANDS = {
     "version" :        ( 'mixminion.Main',       'printVersion' ),
     "unittests" :      ( 'mixminion.test',       'testAll' ),
     "benchmarks" :     ( 'mixminion.benchmark',  'timeAll' ),
+    "testvectors" :    ( 'mixminion.testSupport', 'testVectors' ),
     "send" :           ( 'mixminion.ClientMain', 'runClient' ),
-    # XXXX005 Obsolete; use "queue"; remove in 0.0.5
-    "pool" :           ( 'mixminion.ClientMain', 'runClient' ),
     "queue" :          ( 'mixminion.ClientMain', 'runClient' ),
     "import-server" :  ( 'mixminion.ClientMain', 'importServer' ),
+    "list-server" :    ( 'mixminion.ClientMain', 'listServers' ),
     "list-servers" :   ( 'mixminion.ClientMain', 'listServers' ),
     "update-servers" : ( 'mixminion.ClientMain', 'updateServers' ),
     "decode" :         ( 'mixminion.ClientMain', 'clientDecode' ),
@@ -128,12 +128,9 @@ _COMMANDS = {
     "inspect-surb" :   ( 'mixminion.ClientMain', 'inspectSURBs' ),
     "inspect-surbs" :  ( 'mixminion.ClientMain', 'inspectSURBs' ),
     "flush" :          ( 'mixminion.ClientMain', 'flushQueue' ),
-    "inspect-queue" :   ( 'mixminion.ClientMain', 'listQueue' ),
-    # XXXX005 Obsolete; use "inspect-queue"; remove in 0.0.5
-    "inspect-pool" :   ( 'mixminion.ClientMain', 'listQueue' ),
+    "inspect-queue" :  ( 'mixminion.ClientMain', 'listQueue' ),
+    "clean-queue" :    ( 'mixminion.ClientMain', 'cleanQueue' ),
     "ping" :           ( 'mixminion.ClientMain', 'runPing' ),
-    # XXXX005 Obsolete; use "server-start"; remove in 0.0.5
-    "server" :         ( 'mixminion.server.ServerMain', 'runServer' ),
     "server-start" :   ( 'mixminion.server.ServerMain', 'runServer' ),
     "server-stop" :    ( 'mixminion.server.ServerMain', 'signalServer' ),
     "server-reload" :  ( 'mixminion.server.ServerMain', 'signalServer' ),
@@ -142,6 +139,8 @@ _COMMANDS = {
     "server-stats" :   ( 'mixminion.server.ServerMain', 'printServerStats' ),
     "server-DELKEYS" : ( 'mixminion.server.ServerMain', 'runDELKEYS'),
     "dir":             ( 'mixminion.directory.DirMain', 'main'),
+
+    "shell":           ( 'mixminion.Main',       'commandShell' ),
 }
 
 _USAGE = (
@@ -153,6 +152,7 @@ _USAGE = (
   "       queue          [Schedule an anonymous message to be sent later]\n"+
   "       flush          [Send all messages waiting in the queue]\n"+
   "       inspect-queue  [Describe all messages waiting in the queue]\n"+
+  "       clean-queue    [Remove old messages from the queue\n"+
   "       import-server  [Tell the client about a new server]\n"+
   "       list-servers   [Print a list of currently known servers]\n"+
   "       update-servers [Download a fresh server directory]\n"+
@@ -185,13 +185,63 @@ def printVersion(cmd,args):
     print "NOTE: This software is for testing only.  The user set is too small"
     print "      to be anonymous, and the code is too alpha to be reliable."
 
+def rejectCommand(cmd,args):
+    # This function gets called when we have an obsolete command 'cmd'.
+    # First, let's see whether we know an updated equivalent or not.
+    cmd = cmd.split()[-1]
+    # Map from obsolete commands to current versions.
+    cmdDict = { } # Right now, there aren't any obsolete commands still in use.
+    newCmd = cmdDict.get(cmd)
+    if newCmd:
+        print "The command %r is obsolete.  Use %r instead."%(cmd,newCmd)
+    else:
+        print "The command %r is obsolete."%cmd
+
+    sys.exit(1)
+
 def printUsage():
-    import mixminion
-    print "Mixminion version %s" % mixminion.__version__
     print _USAGE
     print "NOTE: This software is for testing only.  The user set is too small"
     print "      to be anonymous, and the code is too alpha to be reliable."
 
+def commandShell(cmd,args):
+    # Used to implement a 'mixminion shell' on systems (like windows) with
+    # somewhat bogus CLI support.
+    import mixminion
+    import shlex
+
+    if "--help" in args:
+        print "Syntax: mixminion shell [options]"
+        sys.exit(0)
+
+    print "Mixminion version %s" % mixminion.__version__
+    print "Type 'help' for information, and 'exit' to quit."
+
+    lexer = shlex.shlex()
+    lexer.whitespace = " \t"
+    lexer.wordchars = "".join(map(chr,range(41,127)))
+    while 1:
+        print
+        print "mixminion>",
+        words = []
+        while 1:
+            word = lexer.get_token()
+            if word not in ['\r','\n']:
+                words.append(word)
+            else:
+                break
+        if not words:
+            continue
+        command = words[0]
+        args = words[1:]
+        if command == 'exit' or command == 'quit':
+            sys.exit(0)
+        try:
+            #print "calling main with",[sys.argv[0]]+words
+            main([sys.argv[0]]+words)
+        except SystemExit:
+            pass
+
 def main(args):
     "Use <args> to fix path, pick a command and pass it arguments."
     # Specifically, args[0] is used to fix sys.path so we can import
@@ -200,19 +250,30 @@ def main(args):
 
     correctPath(args[0])
 
+##     if len(args) > 2 and args[1] == 'mixminiond':
+##         if _COMMANDS.has_key("server-"+args[2]):
+##             args[1:3] = "server-"+args[2]
+##         else:
+##             printUsage()
+##             sys.exit(1)
+
     # Check whether we have a recognized command.
-    if len(args) == 1 or not _COMMANDS.has_key(args[1]):
+    if len(args) == 1  or not _COMMANDS.has_key(args[1]):
         printUsage()
         sys.exit(1)
 
-    if args[1] not in ('unittests', 'benchmarks', 'version'):
+    if args[1] not in ('unittests', 'benchmarks', 'version') and \
+       '--quiet' not in args and '-Q' not in args:
+        import mixminion
+        print "Mixminion version %s" % mixminion.__version__
         print "This software is for testing purposes only."\
               "  Anonymity is not guaranteed."
 
     # Read the 'common' module to get the UIError class.  To simplify
     # command implementation code, we catch all UIError exceptions here.
     commonModule = __import__('mixminion.Common', {}, {}, ['UIError'])
-    uiErrorClass = getattr(commonModule, 'UIError')
+    uiErrorClass = commonModule.UIError
+    filePermissionErrorClass = commonModule.MixFilePermissionError
 
     # Read the module and function.
     command_module, command_fn = _COMMANDS[args[1]]
@@ -230,8 +291,23 @@ def main(args):
         func(commandStr, ["--help"])
     except uiErrorClass, e:
         e.dumpAndExit()
-    except KeyboardInterrupt:
+    except filePermissionErrorClass, e:
+        print str(e)
+        print "(You can disable file permission checking by setting",
+        print "the MM_NO_FILE_PARANOIA"
+        print "environment variable.)"
+        sys.exit(1)
+    except KeyboardInterrupt, e:
+        if cmdName in ('unittests', 'benchmarks'):
+            raise
         print "Interrupted."
+    except ImportError, e:
+        if str(e).find("_socket") >= 0:
+            print "Error: Your python installation is missing the standard "
+            print "socket module.  Halting."
+            sys.exit(1)
+        else:
+            raise
 
 if __name__ == '__main__':
     main(sys.argv)
