@@ -1,9 +1,9 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: Packet.py,v 1.46 2003/06/06 06:04:57 nickm Exp $
+# $Id: Packet.py,v 1.69 2003/12/14 01:29:25 nickm Exp $
 """mixminion.Packet
 
    Functions, classes, and constants to parse and unparse Mixminion
-   messages and related structures.
+   messages, packets, and related structures.
 
    For functions that handle client-side generation and decoding of
    packets, see BuildMessage.py.  For functions that handle
@@ -11,27 +11,36 @@
 
 __all__ = [ 'compressData', 'CompressedDataTooLong', 'DROP_TYPE',
             'ENC_FWD_OVERHEAD', 'ENC_SUBHEADER_LEN',
-            'FRAGMENT_PAYLOAD_OVERHEAD', 'FWD_TYPE', 'FragmentPayload',
+            'encodeMailHeaders', 'encodeMessageHeaders',
+            'FRAGMENT_PAYLOAD_OVERHEAD', 'FWD_HOST_TYPE', 'FWD_IPV4_TYPE',
+            'FragmentPayload',
+            'FRAGMENT_MESSAGEID_LEN', 'FRAGMENT_TYPE',
             'HEADER_LEN', 'IPV4Info', 'MAJOR_NO', 'MBOXInfo',
             'MBOX_TYPE', 'MINOR_NO', 'MIN_EXIT_TYPE',
-            'MIN_SUBHEADER_LEN', 'Message',
+            'MIN_SUBHEADER_LEN', 'MMTPHostInfo', 'Packet',
             'OAEP_OVERHEAD', 'PAYLOAD_LEN', 'ParseError', 'ReplyBlock',
             'ReplyBlock', 'SECRET_LEN', 'SINGLETON_PAYLOAD_OVERHEAD',
-            'SMTPInfo', 'SMTP_TYPE', 'SWAP_FWD_TYPE', 'SingletonPayload',
+            'SMTPInfo', 'SMTP_TYPE', 'SWAP_FWD_IPV4_TYPE',
+            'SWAP_FWD_HOST_TYPE', 'SingletonPayload',
             'Subheader', 'TAG_LEN', 'TextEncodedMessage',
-            'parseHeader', 'parseIPV4Info',
-            'parseMBOXInfo', 'parseMessage', 'parsePayload', 'parseReplyBlock',
+            'parseHeader', 'parseIPV4Info', 'parseMMTPHostInfo',
+            'parseMBOXInfo', 'parsePacket', 'parseMessageAndHeaders',
+            'parsePayload', 'parseRelayInfoByType', 'parseReplyBlock',
             'parseReplyBlocks', 'parseSMTPInfo', 'parseSubheader',
-            'parseTextEncodedMessages', 'parseTextReplyBlocks', 'uncompressData'            ]
+            'parseTextEncodedMessages', 'parseTextReplyBlocks',
+            'uncompressData'
+            ]
 
 import binascii
 import re
 import struct
 import sys
+import types
 import zlib
 from socket import inet_ntoa, inet_aton
 from mixminion.Common import MixError, MixFatalError, encodeBase64, \
-     floorDiv, formatTime, isSMTPMailbox, LOG, armorText, unarmorText
+     floorDiv, formatTime, isSMTPMailbox, LOG, armorText, unarmorText, \
+     isPlausibleHostname
 from mixminion.Crypto import sha1
 
 if sys.version_info[:3] < (2,2,0):
@@ -40,12 +49,12 @@ if sys.version_info[:3] < (2,2,0):
 # Major and minor number for the understood packet format.
 MAJOR_NO, MINOR_NO = 0,3
 
-# Length of a Mixminion message
-MESSAGE_LEN = 1 << 15
+# Length of a Mixminion packet
+PACKET_LEN = 1 << 15
 # Length of a header section
 HEADER_LEN  = 128 * 16
 # Length of a single payload
-PAYLOAD_LEN = MESSAGE_LEN - HEADER_LEN*2
+PAYLOAD_LEN = PACKET_LEN - HEADER_LEN*2
 
 # Bytes taken up by OAEP padding in RSA-encrypted data
 OAEP_OVERHEAD = 42
@@ -70,16 +79,29 @@ TAG_LEN = 20
 #----------------------------------------------------------------------
 # Values for the 'Routing type' subheader field
 # Mixminion types
-DROP_TYPE      = 0x0000  # Drop the current message
-FWD_TYPE       = 0x0001  # Forward the msg to an IPV4 addr via MMTP
-SWAP_FWD_TYPE  = 0x0002  # SWAP, then forward the msg to an IPV4 addr via MMTP
+DROP_TYPE          = 0x0000 # Drop the packet
+FWD_IPV4_TYPE      = 0x0001 # Forward the packet to an IPV4 addr via MMTP
+SWAP_FWD_IPV4_TYPE = 0x0002 # SWAP, then FWD_IPV4
+FWD_HOST_TYPE      = 0x0003 # Forward the pkt to a hostname, via MMTP.
+SWAP_FWD_HOST_TYPE = 0x0004 # SWAP, then FWD_HOST
 
 # Exit types
 MIN_EXIT_TYPE  = 0x0100  # The numerically first exit type.
 SMTP_TYPE      = 0x0100  # Mail the message
 MBOX_TYPE      = 0x0101  # Send the message to one of a fixed list of addresses
+NEWS_TYPE      = 0x0102  # Post the message to some ngs, and maybe mail it too
+FRAGMENT_TYPE  = 0x0103  # Find the actual delivery info in the message payload
 MAX_EXIT_TYPE  = 0xFFFF
 
+# Set of exit types that don't get tag fields.
+# XXXX007 This interface is really brittle; it needs to change.  I added it
+# XXXX007 in order to allow 'fragment' to be an exit type without adding a
+# XXXX007 needless tag field to every fragment routing info.
+_TYPES_WITHOUT_TAGS = { FRAGMENT_TYPE : 1 }
+
+def typeIsSwap(tp):
+    return tp in (SWAP_FWD_IPV4_TYPE,SWAP_FWD_HOST_TYPE)
+
 class ParseError(MixError):
     """Thrown when a message or portion thereof is incorrectly formatted."""
     pass
@@ -87,28 +109,28 @@ class ParseError(MixError):
 #----------------------------------------------------------------------
 # PACKET-LEVEL STRUCTURES
 
-def parseMessage(s):
-    """Given a 32K string, returns a Message object that breaks it into
+def parsePacket(s):
+    """Given a 32K string, returns a Packet object that breaks it into
        two headers and a payload."""
-    if len(s) != MESSAGE_LEN:
-        raise ParseError("Bad message length")
+    if len(s) != PACKET_LEN:
+        raise ParseError("Bad packet length")
 
-    return Message(s[:HEADER_LEN],
+    return Packet(s[:HEADER_LEN],
                    s[HEADER_LEN:HEADER_LEN*2],
                    s[HEADER_LEN*2:])
 
-class Message:
+class Packet:
     """Represents a complete Mixminion packet
 
        Fields: header1, header2, payload"""
     def __init__(self, header1, header2, payload):
-        """Create a new Message object from three strings."""
+        """Create a new Packet object from three strings."""
         self.header1 = header1
         self.header2 = header2
         self.payload = payload
 
     def pack(self):
-        """Return the 32K string value of this message."""
+        """Return the 32K string value of this packet."""
         return "".join([self.header1,self.header2,self.payload])
 
 def parseHeader(s):
@@ -139,7 +161,7 @@ def parseSubheader(s):
     underflow = ""
     if rlen < len(ri):
         ri, underflow = ri[:rlen], ri[rlen:]
-    if rt >= MIN_EXIT_TYPE and rlen < 20:
+    if rt >= MIN_EXIT_TYPE and not _TYPES_WITHOUT_TAGS.get(rt) and rlen < TAG_LEN:
         raise ParseError("Subheader missing tag")
     return Subheader(major,minor,secret,digest,rt,ri,rlen,underflow)
 
@@ -183,15 +205,23 @@ class Subheader:
         """Return the part of the routingInfo that contains the delivery
            address.  (Requires that routingType is an exit type.)"""
         assert self.routingtype >= MIN_EXIT_TYPE
-        assert len(self.routinginfo) >= TAG_LEN
-        return self.routinginfo[TAG_LEN:]
+        #XXXX007 This interface is completely insane.  Change it.
+        if _TYPES_WITHOUT_TAGS.get(self.routingtype):
+            return self.routinginfo
+        else:
+            assert len(self.routinginfo) >= TAG_LEN
+            return self.routinginfo[TAG_LEN:]
 
     def getTag(self):
         """Return the part of the routingInfo that contains the decoding
            tag. (Requires that routingType is an exit type.)"""
         assert self.routingtype >= MIN_EXIT_TYPE
-        assert len(self.routinginfo) >= TAG_LEN
-        return self.routinginfo[:TAG_LEN]
+        #XXXX007 This interface is completely insane.  Change it.
+        if _TYPES_WITHOUT_TAGS.get(self.routingtype):
+            return ""
+        else:
+            assert len(self.routinginfo) >= TAG_LEN
+            return self.routinginfo[:TAG_LEN]
 
     def setRoutingInfo(self, info):
         """Change the routinginfo, and the routinglength to correspond."""
@@ -241,12 +271,12 @@ class Subheader:
 # Length of the 'MessageID' field in a fragment payload
 FRAGMENT_MESSAGEID_LEN = 20
 # Maximum number of fragments associated with a given message
-MAX_N_FRAGMENTS = 0x7ffff
+MAX_N_FRAGMENTS = 0x7ffffff
 
 # Number of bytes taken up by header fields in a singleton payload.
 SINGLETON_PAYLOAD_OVERHEAD = 2 + DIGEST_LEN
 # Number of bytes taken up by header fields in a fragment payload.
-FRAGMENT_PAYLOAD_OVERHEAD = 2 + DIGEST_LEN + FRAGMENT_MESSAGEID_LEN + 4
+FRAGMENT_PAYLOAD_OVERHEAD = 3 + DIGEST_LEN + FRAGMENT_MESSAGEID_LEN + 4
 # Number of bytes taken up from OAEP padding in an encrypted forward
 # payload, minus bytes saved by spilling the RSA-encrypted block into the
 # tag, minus the bytes taken by the session key.
@@ -261,9 +291,10 @@ def parsePayload(payload):
     bit0 = ord(payload[0]) & 0x80
     if bit0:
         # We have a fragment
-        idx, hash, msgID, msgLen = struct.unpack(FRAGMENT_UNPACK_PATTERN,
-                                         payload[:FRAGMENT_PAYLOAD_OVERHEAD])
-        idx &= 0x7f
+        idxhi, idxlo, hash, msgID, msgLen = \
+               struct.unpack(FRAGMENT_UNPACK_PATTERN,
+                             payload[:FRAGMENT_PAYLOAD_OVERHEAD])
+        idx = ((idxhi & 0x7f) << 16) + idxlo
         contents = payload[FRAGMENT_PAYLOAD_OVERHEAD:]
         if msgLen <= len(contents):
             raise ParseError("Payload has an invalid size field")
@@ -280,9 +311,9 @@ def parsePayload(payload):
 # A singleton payload starts with a 0 bit, 15 bits of size, and a 20-byte hash
 SINGLETON_UNPACK_PATTERN = "!H%ds" % (DIGEST_LEN)
 
-# A fragment payload starts with a 1 bit, a 15-bit paket index, a 20-byte hash,
-# a 20-byte message ID, and 4 bytes of message size.
-FRAGMENT_UNPACK_PATTERN = "!H%ds%dsL" % (DIGEST_LEN, FRAGMENT_MESSAGEID_LEN)
+# A fragment payload starts with a 1 bit, a 23-bit packet index, a
+# 20-byte hash, a 20-byte message ID, and 4 bytes of message size.
+FRAGMENT_UNPACK_PATTERN = "!BH%ds%dsL" % (DIGEST_LEN, FRAGMENT_MESSAGEID_LEN)
 
 class _Payload:
     pass
@@ -295,6 +326,11 @@ class SingletonPayload(_Payload):
         self.hash = hash
         self.data = data
 
+    def computeHash(self):
+        """Update the hash field of this payload to correspond to the hash
+           of the data."""
+        self.hash = sha1(self.data)
+
     def isSingleton(self):
         """Returns true; this is a singleton payload."""
         return 1
@@ -303,6 +339,17 @@ class SingletonPayload(_Payload):
         """Returns the non-padding portion of this payload's data"""
         return self.data[:self.size]
 
+    def getUncompressedContents(self, force=None):
+        """Return the original message from this payload's data, removing
+           compression.  Raise CompressedDataTooLong if the data is too
+           long, and force is not true."""
+        d = self.data[:self.size]
+        if force:
+            return uncompressData(d)
+        else:
+            maxLen = max(20*1024, 20*len(d))
+            return uncompressData(d, maxLen)
+
     def pack(self):
         """Check for reasonable values of fields, and return a packed payload.
         """
@@ -316,9 +363,6 @@ class SingletonPayload(_Payload):
 
 class FragmentPayload(_Payload):
     """Represents the fields of a decoded fragment payload.
-
-       FFFF Fragments are not yet fully supported; there's no code to generate
-            or decode them.
     """
     def __init__(self, index, hash, msgID, msgLen, data):
         self.index = index
@@ -327,6 +371,13 @@ class FragmentPayload(_Payload):
         self.msgLen = msgLen
         self.data = data
 
+    def computeHash(self):
+        """Update the hash field of this payload to correspond to the hash
+           of the data."""
+        self.hash = "X"*DIGEST_LEN
+        p = self.pack()
+        self.hash = sha1(p[23:])
+
     def isSingleton(self):
         """Return false; not a singleton"""
         return 0
@@ -339,11 +390,48 @@ class FragmentPayload(_Payload):
         assert len(self.data) < self.msgLen < 0x100000000L
         assert (PAYLOAD_LEN - FRAGMENT_PAYLOAD_OVERHEAD - len(self.data)) in \
                (0, ENC_FWD_OVERHEAD)
-        idx = self.index | 0x8000
-        header = struct.pack(FRAGMENT_UNPACK_PATTERN, idx, self.hash,
-                             self.msgID, self.msgLen)
+        idxhi = ((self.index & 0xff0000) >> 16) | 0x80
+        idxlo = self.index & 0x00fffff
+        header = struct.pack(FRAGMENT_UNPACK_PATTERN, idxhi, idxlo,
+                             self.hash, self.msgID, self.msgLen)
         return "%s%s" % (header, self.data)
 
+    def getOverhead(self):
+        return PAYLOAD_LEN - FRAGMENT_PAYLOAD_OVERHEAD - len(self.data)
+
+#----------------------------------------------------------------------
+# Encoding of messages fragmented for reassembly by the exit server.
+#
+# Such messages are encoded by adding routing-type, routing-len, and
+# routing-info fields at the start of the payload before fragmentation,
+# so that the server doesn't recover the delivery address before it's time
+# to deliver the message.
+
+SSF_UNPACK_PATTERN = "!HH"
+SSF_PREFIX_LEN = 4
+
+def parseServerSideFragmentedMessage(s):
+    if len(s) < SSF_PREFIX_LEN:
+        raise ParseError("Server-side fragmented message too short")
+
+    rt, rl = struct.unpack(SSF_UNPACK_PATTERN, s[:SSF_PREFIX_LEN])
+    if len(s) < SSF_PREFIX_LEN + rl:
+        raise ParseError("Server-side fragmented message too short")
+    ri = s[SSF_PREFIX_LEN:SSF_PREFIX_LEN+rl]
+    comp = s[SSF_PREFIX_LEN+rl:]
+    return ServerSideFragmentedMessage(rt, ri, comp)
+
+class ServerSideFragmentedMessage:
+    def __init__(self, routingtype, routinginfo, compressedContents):
+        self.routingtype = routingtype
+        self.routinginfo = routinginfo
+        self.compressedContents = compressedContents
+    def pack(self):
+        return "%s%s%s" % (struct.pack(SSF_UNPACK_PATTERN, self.routingtype,
+                                       len(self.routinginfo)),
+                           self.routinginfo,
+                           self.compressedContents)
+
 #----------------------------------------------------------------------
 # REPLY BLOCKS
 
@@ -418,8 +506,8 @@ def _parseReplyBlock(s):
         raise ParseError("Misformatted reply block")
 
     if major != 0x00 or minor != 0x01:
-        raise ParseError("Unrecognized version on reply block %s.%s",
-                         major,minor)
+        raise ParseError("Unrecognized version on reply block %s.%s"
+                         %(major,minor))
 
     ri = s[MIN_RB_LEN:]
     length = rlen + MIN_RB_LEN
@@ -441,15 +529,18 @@ class ReplyBlock:
         self.encryptionKey = key
 
     def format(self):
+        from mixminion.ServerInfo import displayServer
         hash = binascii.b2a_hex(sha1(self.pack()))
         expiry = formatTime(self.timestamp)
-        if self.routingType == SWAP_FWD_TYPE:
-            server = parseIPV4Info(self.routingInfo).format()
+        if self.routingType == SWAP_FWD_IPV4_TYPE:
+            routing = parseIPV4Info(self.routingInfo)
+        elif self.routingType == SWAP_FWD_HOST_TYPE:
+            routing = parseMMTPHostInfo(self.routingInfo)
         else:
-            server = "????"
+            routing = None
         return """Reply block hash: %s
 Expires at: %s GMT
-First server is: %s""" % (hash, expiry, server)
+First server is: %s""" % (hash, expiry, displayServer(routing))
 
     def pack(self):
         """Returns the external representation of this reply block"""
@@ -466,13 +557,30 @@ First server is: %s""" % (hash, expiry, server)
 #----------------------------------------------------------------------
 # Routing info
 
+def parseRelayInfoByType(routingType,routingInfo):
+    """Parse the routingInfo contained in the string 'routinginfo',
+       according to the type in 'routingType'.  Only relay types are
+       supported."""
+    if routingType in (FWD_IPV4_TYPE, SWAP_FWD_IPV4_TYPE):
+        parseFn = parseIPV4Info
+        parsedType = IPV4Info
+    elif routingType in (FWD_HOST_TYPE, SWAP_FWD_HOST_TYPE):
+        parseFn = parseMMTPHostInfo
+        parsedType = MMTPHostInfo
+    else:
+        raise MixFatalError("Unrecognized relay type 0x%04X"%routingType)
+    if type(routingInfo) == types.StringType:
+        routingInfo = parseFn(routingInfo)
+    assert isinstance(routingInfo, parsedType)
+    return routingInfo
+
 # An IPV4 address (Used by SWAP_FWD and FWD) is packed as: four bytes
 # of IP address, a short for the portnum, and DIGEST_LEN bytes of keyid.
 IPV4_PAT = "!4sH%ds" % DIGEST_LEN
 
 def parseIPV4Info(s):
     """Converts routing info for an IPV4 address into an IPV4Info object,
-       suitable for use by FWD or SWAP_FWD modules."""
+       suitable for use by FWD_IPV4 or SWAP_FWD_IPV4 modules."""
     if len(s) != 4+2+DIGEST_LEN:
         raise ParseError("IPV4 information with wrong length (%d)" % len(s))
     try:
@@ -483,11 +591,13 @@ def parseIPV4Info(s):
     return IPV4Info(ip, port, keyinfo)
 
 class IPV4Info:
-    """An IPV4Info object represents the routinginfo for a FWD or
-       SWAP_FWD hop.
+    """An IPV4Info object represents the routinginfo for a FWD_IPV4 or
+       SWAP_FWD_IPV4 hop.  This kind of routing is only used with older
+       servers that don't support hostname-based routing.
 
        Fields: ip (a dotted quad string), port (an int from 0..65535),
        and keyinfo (a digest)."""
+    #XXXX007/8 phase this out.
     def __init__(self, ip, port, keyinfo):
         """Construct a new IPV4Info"""
         assert 0 <= port <= 65535
@@ -521,6 +631,59 @@ class IPV4Info:
         if r: return r
         return cmp(self.keyinfo, other.keyinfo)
 
+MMTP_HOST_PAT = "!H%ds" % DIGEST_LEN
+
+def parseMMTPHostInfo(s):
+    """Converts routing info for a hostname address into an MMTPHostInfo
+    object, suitable for use by FWD_HOST or SWAP_FWD_HOST modules."""
+    if len(s) < 2+DIGEST_LEN+1:
+        raise ParseError("Routing information is too short.")
+    try:
+        port, keyinfo = struct.unpack(MMTP_HOST_PAT, s[:2+DIGEST_LEN])
+    except struct.error:
+        raise ParseError("Misformatted routing info")
+    host = s[2+DIGEST_LEN:]
+    if not isPlausibleHostname(host):
+        raise ParseError("Nonsensical hostname")
+    return MMTPHostInfo(s[2+DIGEST_LEN:], port, keyinfo)
+
+class MMTPHostInfo:
+    """An MMTPHostInfo object represents the routinginfo for a FWD_HOST or
+       SWAP_FWD_HOST hop.
+
+       Fields: hostname, port (an int from 0..65535), and keyinfo (a
+       digest)."""
+    def __init__(self, hostname, port, keyinfo):
+        assert 0 <= port <= 65535
+        self.hostname = hostname.lower()
+        self.port = port
+        self.keyinfo = keyinfo
+
+    def format(self):
+        return "%s:%s (keyid=%s)"%(self.hostname, self.port,
+                                   binascii.b2a_hex(self.keyinfo))
+
+    def pack(self):
+        """Return the routing info for this address"""
+        assert len(self.keyinfo) == DIGEST_LEN
+        return struct.pack(MMTP_HOST_PAT,self.port,self.keyinfo)+self.hostname
+
+    def __repr__(self):
+        return "MMTPHostInfo(%r, %r, %r)"%(
+            self.hostname,self.port,self.keyinfo)
+
+    def __hash__(self):
+        return hash(self.pack())
+
+    def __cmp__(self, other):
+        r = cmp(type(self), type(other))
+        if r: return r
+        r = cmp(self.hostname, other.hostname)
+        if r: return r
+        r = cmp(self.port, other.port)
+        if r: return r
+        return cmp(self.keyinfo, other.keyinfo)
+
 def parseSMTPInfo(s):
     """Convert the encoding of an SMTP exitinfo into an SMTPInfo object."""
     if not isSMTPMailbox(s):
@@ -566,22 +729,21 @@ class MBOXInfo:
 # the body is base64-encoded.
 
 MESSAGE_ARMOR_NAME = "TYPE III ANONYMOUS MESSAGE"
- 
-def parseTextEncodedMessages(msg,force=0):
 
+def parseTextEncodedMessages(msg,force=0):
     """Given a text-encoded Type III packet, return a list of
        TextEncodedMessage objects or raise ParseError.
-       
+
           force -- uncompress the message even if it's overcompressed.
     """
-    
+
     def isBase64(t,f):
         for k,v in f:
             if k == "Message-type":
                 if v != 'plaintext':
                     return 1
         return 0
-    
+
     unarmored = unarmorText(msg, (MESSAGE_ARMOR_NAME,), base64fn=isBase64)
     res = []
     for tp,fields,val in unarmored:
@@ -596,6 +758,8 @@ def parseTextEncodedMessages(msg,force=0):
             msgType = "BIN"
         elif d['Message-type'] == 'encrypted':
             msgType = "ENC"
+        elif d['Message-type'] == 'fragment':
+            msgType = "FRAG"
         else:
             raise ParseError("Unknown message type: %r"%d["Message-type"])
 
@@ -605,8 +769,8 @@ def parseTextEncodedMessages(msg,force=0):
 
         if msgType == 'LONG' and force:
             msg = uncompressData(msg)
-            
-        if msgType in ('TXT','BIN','LONG'):
+
+        if msgType in ('TXT','BIN','LONG','FRAG'):
             res.append(TextEncodedMessage(val, msgType))
         else:
             assert msgType == 'ENC'
@@ -625,10 +789,11 @@ class TextEncodedMessage:
        over a text-based medium."""
     def __init__(self, contents, messageType, tag=None):
         """Create a new TextEncodedMessage given a set of contents, a
-           messageType ('TXT', 'ENC', 'LONG', or 'BIN'), and optionally
-           a tag."""
-        assert messageType in ('TXT', 'ENC', 'LONG', 'BIN')
-        assert tag is None or (messageType == 'ENC' and len(tag) == 20)
+           messageType ('TXT', 'ENC', 'LONG', 'FRAG' or 'BIN'), and optionally
+           a tag.
+           """
+        assert messageType in ('TXT', 'ENC', 'LONG', 'BIN', 'FRAG')
+        assert tag is None or (messageType == 'ENC' and len(tag) == TAG_LEN)
         self.contents = contents
         self.messageType = messageType
         self.tag = tag
@@ -644,6 +809,9 @@ class TextEncodedMessage:
     def isOvercompressed(self):
         """Return true iff this is an overcompressed plaintext packet."""
         return self.messageType == 'LONG'
+    def isFragment(self):
+        """Return true iff this is a fragment packet."""
+        return self.messageType == 'FRAG'
     def getContents(self):
         """Return the (unencoded) contents of this packet."""
         return self.contents
@@ -657,7 +825,8 @@ class TextEncodedMessage:
                    { 'TXT' : "plaintext",
                      'LONG' : "overcompressed",
                      'BIN' : "binary",
-                     'ENC' : "encrypted" }[self.messageType]),
+                     'ENC' : "encrypted",
+                     'FRAG' : 'fragment' }[self.messageType]),
                   ]
         if self.messageType == 'ENC':
             fields.append(("Decoding-handle",
@@ -666,6 +835,74 @@ class TextEncodedMessage:
         return armorText(c, MESSAGE_ARMOR_NAME, headers=fields,
                          base64=(self.messageType!='TXT'))
 
+#----------------------------------------------------------------------
+# Header encoding
+
+# Longest allowed length of a single header.
+MAX_HEADER_LEN = 900
+
+def encodeMailHeaders(subject=None, fromAddr=None, inReplyTo=None,
+                      references=None):
+    """Given (optionally) any of the headers permissible for email
+       messages, return a string to be prepended to a message before
+       encoding.  Raise MixError on failure."""
+    headers = {}
+    if subject:
+        headers['SUBJECT'] = subject
+    if fromAddr:
+        for badchar in ('"', '[', ']', ':'):
+            if badchar in fromAddr:
+                raise MixError("Forbidden character %r in from address"%
+                               badchar)
+        headers['FROM'] = fromAddr
+    if inReplyTo:
+        headers['IN-REPLY-TO'] = inReplyTo
+    if references:
+        headers['REFERENCES'] = references
+    return encodeMessageHeaders(headers)
+
+def encodeMessageHeaders(headers):
+    """Given a dictionary of (header,header-value) entries, encode the
+       entries and return a string to be prepended to a message before
+       encoding.  Requires that headers are in acceptable format.
+       Raises MixError on failure.
+    """
+    items = []
+    hitems = headers.items()
+    hitems.sort()
+    for k,v in hitems:
+        item = "%s:%s\n"%(k,v)
+        if not HEADER_RE.match(item) or "\n" in k or "\n" in v:
+            raise ParseError("Invalid value for %s header"%k)
+        if len(v) > 900:
+            raise ParseError("The %s header is too long"%k.lower())
+        items.append(item)
+    items.append("\n")
+    return "".join(items)
+
+HEADER_RE = re.compile(r'^([!-9;-~]+):([ -~]*)\n')
+
+def parseMessageAndHeaders(message):
+    """Given a message with encoded headers, return a 2-tuple containing
+       the message, and a dictionary mapping header names to header values.
+       Skips improperly formatted headers."""
+    headers = {}
+    msg = message
+    while 1:
+        if msg[0] == '\n':
+            return msg[1:], headers
+        m = HEADER_RE.match(msg)
+        if m:
+            k,v = m.groups()
+            if len(v) > MAX_HEADER_LEN:
+                LOG.warn("Rejecting overlong exit header %r:%r...",k,v[:30])
+            else:
+                headers[k] = v
+            msg = msg[m.end():]
+        else:
+            LOG.warn("Could not parse headers on message; not using them.")
+            return message, headers
+
 #----------------------------------------------------------------------
 # COMPRESSION FOR PAYLOADS
 
@@ -757,12 +994,14 @@ def _validateZlib():
        would make senders partitionable by payload compression.)
     """
     global _ZLIB_LIBRARY_OK
-    ver = getattr(zlib, "ZLIB_VERSION")
+    ver = getattr(zlib, "ZLIB_VERSION", None)
     if ver and ver < "1.1.2":
         raise MixFatalError("Zlib version %s is not supported"%ver)
 
     _ZLIB_LIBRARY_OK = 0.5
-    if ver in ("1.1.2", "1.1.3", "1.1.4"):
+    if ver in ("1.1.2", "1.1.3", "1.1.4", "1.2.0", "1.2.0.1", "1.2.0.2",
+               "1.2.0.3", "1.2.0.4", "1.2.0.5", "1.2.0.6", "1.2.0.7", 
+               "1.2.0.8", "1.2.1"):
         _ZLIB_LIBRARY_OK = 1
         return
 
