@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: ServerInfo.py,v 1.49 2003/06/09 00:03:37 nickm Exp $
+# $Id: ServerInfo.py,v 1.69 2003/12/12 22:56:16 nickm Exp $
 
 """mixminion.ServerInfo
 
@@ -8,18 +8,21 @@
    descriptors.
    """
 
-__all__ = [ 'ServerInfo', 'ServerDirectory' ]
+__all__ = [ 'ServerInfo', 'ServerDirectory', 'displayServer' ]
 
 import re
 import time
+import types
 
 import mixminion.Config
 import mixminion.Crypto
+import mixminion.MMTPClient
+import mixminion.Packet
 
 from mixminion.Common import IntervalSet, LOG, MixError, createPrivateDir, \
     formatBase64, formatDate, formatTime, readPossiblyGzippedFile
 from mixminion.Config import ConfigError
-from mixminion.Packet import IPV4Info
+from mixminion.Packet import IPV4Info, MMTPHostInfo
 from mixminion.Crypto import CryptoError, DIGEST_LEN, pk_check_signature
 
 # Longest allowed Contact email
@@ -37,6 +40,59 @@ PACKET_KEY_BYTES = 2048 >> 3
 # Length of MMTP key
 MMTP_KEY_BYTES = 1024 >> 3
 
+# ----------------------------------------------------------------------
+def displayServer(s):
+    """Return the best possible human-readable name for a server 's'.
+       's' must be one of: None, IPV4Info, MMTPHostInfo, ServerInfo,
+       string.
+    """
+    if isinstance(s, types.StringType):
+        return s
+    elif isinstance(s, ServerInfo):
+        if s.getHostname():
+            addr = "%s:%s" % (s.getHostname(), s.getPort())
+        else:
+            addr = "%s:%s" % (s.getIP(), s.getPort())
+        nickname = "'%s'" % s.getNickname()
+    elif (isinstance(s, mixminion.Packet.IPV4Info) or
+          isinstance(s, mixminion.Packet.MMTPHostInfo)):
+        nickname = getNicknameByKeyID(s.keyinfo)
+        if nickname:
+            nickname = "'%s'" % nickname
+        else:
+            nickname = 'server'
+        if isinstance(s, mixminion.Packet.IPV4Info):
+            addr = "%s:%s" % (s.ip, s.port)
+        else:
+            addr = "%s:%s" % (s.hostname, s.port)
+    elif s is None:
+        return "unknown server"
+    else:
+        assert 0
+
+    return "%s at %s" % (nickname, addr)
+
+def getNicknameByKeyID(keyid):
+    """Given a 20-byte keyid, look up the nickname of the corresponding
+       server.  Return the nickname on success and None if we don't recognize
+       the server.
+
+       FFFF Right now, this is not supported for servers, since they don't
+       FFFF download directories.
+    """
+    #FFFF Be cleverer with all-zero keyids.
+    if _keyIDToNicknameFn:
+        return _keyIDToNicknameFn(keyid)
+    else:
+        return None
+
+# This variable should hold None, or a function that maps from keyids to
+# nicknames, and returns None on failure.  Currently set by
+# ClientDirectory.ClientDirectory._installAsKeyIDResolver().
+_keyIDToNicknameFn = None
+
+# ----------------------------------------------------------------------
+
 # tmp alias to make this easier to spell.
 C = mixminion.Config
 class ServerInfo(mixminion.Config._ConfigFile):
@@ -47,48 +103,78 @@ class ServerInfo(mixminion.Config._ConfigFile):
 
     """A ServerInfo object holds a parsed server descriptor."""
     _restrictFormat = 1
-    _restrictKeys = 0
+    _restrictKeys = _restrictSections = 0
     _syntax = {
         "Server" : { "__SECTION__": ("REQUIRE", None, None),
                      "Descriptor-Version": ("REQUIRE", None, None),
-                     "Nickname": ("REQUIRE", C._parseNickname, None),
-                     "Identity": ("REQUIRE", C._parsePublicKey, None),
-                     "Digest": ("REQUIRE", C._parseBase64, None),
-                     "Signature": ("REQUIRE", C._parseBase64, None),
-                     "Published": ("REQUIRE", C._parseTime, None),
-                     "Valid-After": ("REQUIRE", C._parseDate, None),
-                     "Valid-Until": ("REQUIRE", C._parseDate, None),
+                     "Nickname": ("REQUIRE", "nickname", None),
+                     "Identity": ("REQUIRE", "publicKey", None),
+                     "Digest": ("REQUIRE", "base64", None),
+                     "Signature": ("REQUIRE", "base64", None),
+                     "Published": ("REQUIRE", "time", None),
+                     "Valid-After": ("REQUIRE", "date", None),
+                     "Valid-Until": ("REQUIRE", "date", None),
                      "Contact": ("ALLOW", None, None),
                      "Comments": ("ALLOW", None, None),
-                     "Packet-Key": ("REQUIRE", C._parsePublicKey, None),
+                     "Packet-Key": ("REQUIRE", "publicKey", None),
                      "Contact-Fingerprint": ("ALLOW", None, None),
+                     "Packet-Formats": ("ALLOW", None, None),#XXXX007 remove
                      # XXXX010 change these next few to "REQUIRE".
-                     "Packet-Formats": ("ALLOW", None, None),
+                     "Packet-Versions": ("ALLOW", None, None),
                      "Software": ("ALLOW", None, None),
-                     "Secure-Configuration": ("ALLOW", C._parseBoolean, None),
+                     "Secure-Configuration": ("ALLOW", "boolean", None),
+                     "Why-Insecure": ("ALLOW", None, None),
                      },
         "Incoming/MMTP" : {
                      "Version": ("REQUIRE", None, None),
-                     "IP": ("REQUIRE", C._parseIP, None),
-                     "Port": ("REQUIRE", C._parseInt, None),
-                     "Key-Digest": ("REQUIRE", C._parseBase64, None),
+                     "IP": ("ALLOW", "IP", None),#XXXX007/8 remove
+                     "Hostname": ("ALLOW", "host", None),#XXXX008 require
+                     "Port": ("REQUIRE", "int", None),
+                     "Key-Digest": ("ALLOW", "base64", None),#XXXX007/8 rmv
                      "Protocols": ("REQUIRE", None, None),
-                     "Allow": ("ALLOW*", C._parseAddressSet_allow, None),
-                     "Deny": ("ALLOW*", C._parseAddressSet_deny, None),
+                     "Allow": ("ALLOW*", "addressSet_allow", None),
+                     "Deny": ("ALLOW*", "addressSet_deny", None),
                      },
         "Outgoing/MMTP" : {
                      "Version": ("REQUIRE", None, None),
                      "Protocols": ("REQUIRE", None, None),
-                     "Allow": ("ALLOW*", C._parseAddressSet_allow, None),
-                     "Deny": ("ALLOW*", C._parseAddressSet_deny, None),
+                     "Allow": ("ALLOW*", "addressSet_allow", None),
+                     "Deny": ("ALLOW*", "addressSet_deny", None),
                      },
         "Delivery/MBOX" : {
                      "Version": ("REQUIRE", None, None),
+                     # XXXX007 change to 'REQUIRE'
+                     "Maximum-Size": ("ALLOW", "int", "32"),
+                     # XXXX007 change to 'REQUIRE'
+                     "Allow-From": ("ALLOW", "boolean", "yes"),
                      },
         "Delivery/SMTP" : {
                      "Version": ("REQUIRE", None, None),
-                     }
+                     # XXXX007 change to 'REQUIRE'
+                     "Maximum-Size": ("ALLOW", "int", "32"),
+                     "Allow-From": ("ALLOW", "boolean", "yes"),
+                     },
+        "Delivery/Fragmented" : {
+                     "Version": ("REQUIRE", None, None),
+                     "Maximum-Fragments": ("REQUIRE", "int", None),
+                     },
+        # We never read these values, except to see whether we should
+        # regenerate them.  Depending on these options would violate
+        # the spec.
+        "Testing" : {
+                     "Platform": ("ALLOW", None, None),
+                     "Configuration": ("ALLOW", None, None),
+                     },
         }
+    _features = { "capabilities" : 1, "caps" : 1 }
+    expected_versions = {
+         "Server" : ( "Descriptor-Version", "0.2"),
+         "Incoming/MMTP" : ("Version", "0.1"),
+         "Outgoing/MMTP" : ("Version", "0.1"),
+         "Delivery/Fragmented" : ("Version", "0.1"),
+         "Delivery/MBOX" : ("Version", "0.1"),
+         "Delivery/SMTP" : ("Version", "0.1"),
+         }
 
     def __init__(self, fname=None, string=None, assumeValid=0,
                  validatedDigests=None):
@@ -113,17 +199,32 @@ class ServerInfo(mixminion.Config._ConfigFile):
                     if k == 'Descriptor-Version' and v.strip() != '0.2':
                         raise ConfigError("Unrecognized descriptor version: %s"
                                           % v.strip())
-            # FFFF005 Remove sections with unrecognized versions.
 
-        return contents
 
+        # Remove any sections with unrecognized versions.
+        revisedContents = []
+        for name, ents in contents:
+            v = self.expected_versions.get(name)
+            if not v:
+                revisedContents.append((name, ents))
+                continue
+            versionkey, versionval = v
+            for k,v,_ in ents:
+                if k == versionkey and v.strip() != versionval:
+                    LOG.warn("Skipping %s section with unrecognized version %s"
+                             , name, v.strip())
+                    break
+            else:
+                revisedContents.append((name, ents))
+
+        return revisedContents
 
     def validate(self, lines, contents):
         ####
         # Check 'Server' section.
         server = self['Server']
         if server['Descriptor-Version'] != '0.2':
-            raise ConfigError("Unrecognized descriptor version %r",
+            raise ConfigError("Unrecognized descriptor version %r" %
                               server['Descriptor-Version'])
 
         ####
@@ -178,6 +279,8 @@ class ServerInfo(mixminion.Config._ConfigFile):
             if len(inMMTP['Key-Digest']) != DIGEST_LEN:
                 raise ConfigError("Invalid key digest %s"%
                                   formatBase64(inMMTP['Key-Digest']))
+            if not inMMTP['IP'] and not inMMTP['Hostname']:
+                raise ConfigError("Incoming/MMTP section has neither IP nor hostname")
 
         ## Outgoing/MMTP section
         outMMTP = self['Outgoing/MMTP']
@@ -200,9 +303,19 @@ class ServerInfo(mixminion.Config._ConfigFile):
            descriptor."""
         return self['Server']['Digest']
 
-    def getAddr(self):
-        """Returns this server's IP address"""
-        return self['Incoming/MMTP']['IP']
+    def getIP(self):
+        """Returns this server's IP address.  (Returns None for servers
+           running version 0.0.7 or later.)"""
+        return self['Incoming/MMTP'].get('IP')
+
+    def getHostname(self):
+        """Return this server's Hostname. (Returns None for servers running
+           version 0.0.5 or earlier.)"""
+        #XXXX006 remove this.  0.0.6alpha1 could crash when it got hostnames.
+        #XXXX006 Sadly, some people installed it anyway.
+        if self['Server'].get("Software","").endswith("0.0.6alpha1"):
+            return None
+        return self['Incoming/MMTP'].get("Hostname")
 
     def getPort(self):
         """Returns this server's IP port"""
@@ -212,20 +325,98 @@ class ServerInfo(mixminion.Config._ConfigFile):
         """Returns the RSA key this server uses to decrypt messages"""
         return self['Server']['Packet-Key']
 
-    def getKeyID(self):
+    def getKeyDigest(self):
         """Returns a hash of this server's MMTP key"""
-        return self['Incoming/MMTP']['Key-Digest']
+        return mixminion.Crypto.sha1(
+            mixminion.Crypto.pk_encode_public_key(self['Server']['Identity']))
+
+    def getIPV4Info(self):
+        """Returns a mixminion.Packet.IPV4Info object for routing messages
+           to this server.  (Returns None for servers running version 0.0.5
+           or earlier.)"""
+        ip = self.getIP()
+        if ip is None: return None
+        return IPV4Info(ip, self.getPort(), self.getKeyDigest())
+
+    def getMMTPHostInfo(self):
+        """Returns a mixminion.Packet.MMTPHostInfo object for routing messages
+           to this server.  (Returns None for servers running version 0.0.7
+           or later.)"""
+        host = self.getHostname()
+        if host is None: return None
+        return MMTPHostInfo(host, self.getPort(), self.getKeyDigest())
 
     def getRoutingInfo(self):
-        """Returns a mixminion.Packet.IPV4Info object for routing messages
-           to this server."""
-        return IPV4Info(self.getAddr(), self.getPort(), self.getKeyID())
+        """Return whichever of MMTPHostInfo or IPV4 info is best for
+           delivering to this server (assuming that the sending host
+           supports both."""
+        if self.getHostname():
+            return self.getMMTPHostInfo()
+        else:
+            return self.getIPV4Info()
 
     def getIdentity(self):
+        """Return this server's public identity key."""
         return self['Server']['Identity']
 
+    def getIncomingMMTPProtocols(self):
+        """Return a list of the MMTP versions supported by this this server
+           for incoming packets."""
+        inc = self['Incoming/MMTP']
+        if not inc.get("Version"):
+            return []
+        return [ s.strip() for s in inc["Protocols"].split(",") ]
+
+    def getOutgoingMMTPProtocols(self):
+        """Return a list of the MMTP versions supported by this this server
+           for outgoing packets."""
+        inc = self['Outgoing/MMTP']
+        if not inc.get("Version"):
+            return []
+        return [ s.strip() for s in inc["Protocols"].split(",") ]
+
+    def canRelayTo(self, otherDesc):
+        """Return true iff this server can relay packets to the server
+           described by otherDesc."""
+        if self.hasSameNicknameAs(otherDesc):
+            return 1
+        myOutProtocols = self.getOutgoingMMTPProtocols()
+        otherInProtocols = otherDesc.getIncomingMMTPProtocols()
+        for out in myOutProtocols:
+            if out in otherInProtocols:
+                return 1
+        return 0
+
+    def canStartAt(self):
+        """Return true iff this server is one we (that is, this
+           version of Mixminion) can send packets to directly."""
+        myInProtocols = self.getIncomingMMTPProtocols()
+        for out in mixminion.MMTPClient.BlockingClientConnection.PROTOCOL_VERSIONS:
+            if out in myInProtocols:
+                return 1
+        return 0
+
+    def getRoutingFor(self, otherDesc, swap=0):
+        """Return a 2-tuple of (routingType, routingInfo) for relaying
+           a packet from this server to the server described by
+           otherDesc.  If swap is true, the relay is at a crossover
+           point."""
+        assert self.canRelayTo(otherDesc)
+        assert 0 <= swap <= 1
+        if self.getHostname() and otherDesc.getHostname():
+            ri = otherDesc.getMMTPHostInfo().pack()
+            rt = [mixminion.Packet.FWD_HOST_TYPE,
+                  mixminion.Packet.SWAP_FWD_HOST_TYPE][swap]
+        else:
+            ri = otherDesc.getIPV4Info().pack()
+            rt = [mixminion.Packet.FWD_IPV4_TYPE,
+                  mixminion.Packet.SWAP_FWD_IPV4_TYPE][swap]
+
+        return rt, ri
+
     def getCaps(self):
-        # FFFF refactor this once we have client addresses.
+        """Return a list of strings to describe this servers abilities in
+           a concise human-readable format."""
         caps = []
         if not self['Incoming/MMTP'].get('Version'):
             return caps
@@ -236,8 +427,19 @@ class ServerInfo(mixminion.Config._ConfigFile):
         # XXXX This next check is highly bogus.
         if self['Outgoing/MMTP'].get('Version'):
             caps.append('relay')
+        if self['Delivery/Fragmented'].get('Version'):
+            caps.append('frag')
         return caps
 
+    def isSameDescriptorAs(self, other):
+        """Return true iff this is the same server descriptor as other."""
+        return self.getDigest() == other.getDigest()
+
+    def hasSameNicknameAs(self, other):
+        """Return true iff this server descriptor has the same nickname as
+           other."""
+        return self.getNickname().lower() == other.getNickname().lower()
+
     def isValidated(self):
         """Return true iff this ServerInfo has been validated"""
         return self._isValidated
@@ -260,14 +462,14 @@ class ServerInfo(mixminion.Config._ConfigFile):
     def isValidFrom(self, startAt, endAt):
         """Return true iff this ServerInfo is valid at all time from 'startAt'
            to 'endAt'."""
-        assert startAt < endAt
+        assert startAt <= endAt
         return (self['Server']['Valid-After'] <= startAt and
                 endAt <= self['Server']['Valid-Until'])
 
     def isValidAtPartOf(self, startAt, endAt):
         """Return true iff this ServerInfo is valid at some time between
            'startAt' and 'endAt'."""
-        assert startAt < endAt
+        assert startAt <= endAt
         va = self['Server']['Valid-After']
         vu = self['Server']['Valid-Until']
         return ((startAt <= va and va <= endAt) or
@@ -299,6 +501,15 @@ class ServerInfo(mixminion.Config._ConfigFile):
                 valid -= o.getIntervalSet()
         return valid.isEmpty()
 
+    def getFeature(self,sec,name):
+        """Overrides getFeature from _ConfigFile."""
+        if sec == '-':
+            if name in ("caps", "capabilities"):
+                return " ".join(self.getCaps())
+            assert 0
+        else:
+            return mixminion.Config._ConfigFile.getFeature(self,sec,name)
+
 #----------------------------------------------------------------------
 # Server Directories
 
@@ -379,19 +590,19 @@ class _DirectoryHeader(mixminion.Config._ConfigFile):
     # expectedDigest: the 20-byte digest we expect to find in this
     #    directory's header.
     _restrictFormat = 1
-    _restrictKeys = 0
+    _restrictKeys = _restrictSections = 0
     _syntax = {
         'Directory': { "__SECTION__": ("REQUIRE", None, None),
                        "Version": ("REQUIRE", None, None),
-                       "Published": ("REQUIRE", C._parseTime, None),
-                       "Valid-After": ("REQUIRE", C._parseDate, None),
-                       "Valid-Until": ("REQUIRE", C._parseDate, None),
+                       "Published": ("REQUIRE", "time", None),
+                       "Valid-After": ("REQUIRE", "date", None),
+                       "Valid-Until": ("REQUIRE", "date", None),
                        "Recommended-Servers": ("REQUIRE", None, None),
                        },
         'Signature': {"__SECTION__": ("REQUIRE", None, None),
-                 "DirectoryIdentity": ("REQUIRE", C._parsePublicKey, None),
-                 "DirectoryDigest": ("REQUIRE", C._parseBase64, None),
-                 "DirectorySignature": ("REQUIRE", C._parseBase64, None),
+                 "DirectoryIdentity": ("REQUIRE", "publicKey", None),
+                 "DirectoryDigest": ("REQUIRE", "base64", None),
+                 "DirectorySignature": ("REQUIRE", "base64", None),
                       },
         'Recommended-Software': {"__SECTION__": ("ALLOW", None, None),
                 "MixminionClient": ("ALLOW", None, None),
@@ -455,7 +666,7 @@ _leading_whitespace_re = re.compile(r'^[ \t]+', re.M)
 _trailing_whitespace_re = re.compile(r'[ \t]+$', re.M)
 _abnormal_line_ending_re = re.compile(r'\r\n?')
 def _cleanForDigest(s):
-    """Helper function: clean line endigs and whitespace so we can calculate
+    """Helper function: clean line endings and whitespace so we can calculate
        our digests with uniform results."""
     # should be shared with config, serverinfo.
     s = _abnormal_line_ending_re.sub("\n", s)
@@ -473,7 +684,7 @@ def _getDigestImpl(info, regex, digestField=None, sigField=None, rsa=None):
 
        info -- the string to digest or sign.
        regex -- a compiled regex that matches the line containing the digest
-          and the line containting the signature.
+          and the line containing the signature.
        digestField -- If not signing, None.  Otherwise, the name of the digest
           field.
        sigField -- If not signing, None.  Otherwise, the name of the signature
