@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: benchmark.py,v 1.42 2003/06/06 07:54:46 nickm Exp $
+# $Id: benchmark.py,v 1.50 2003/12/12 22:37:13 nickm Exp $
 
 """mixminion.benchmark
 
@@ -23,8 +23,8 @@ from time import time
 import mixminion._minionlib as _ml
 import mixminion.server.ServerQueue
 
-from mixminion.BuildMessage import _buildHeader, buildForwardMessage, \
-     compressData, uncompressData, _encodePayload, decodePayload
+from mixminion.BuildMessage import _buildHeader, buildForwardPacket, \
+     compressData, uncompressData, encodeMessage, decodePayload
 from mixminion.Common import secureDelete, installSIGCHLDHandler, \
      waitForChildren, formatBase64, Lockfile
 from mixminion.Crypto import *
@@ -41,7 +41,7 @@ from mixminion.testSupport import mix_mktemp
 # If PRECISION_FACTOR is >1, we time everything for PRECISION_FACTOR times
 # more iterations than ususal.
 #
-# If PRESISION_FACTOR is 0, we only try stuff once.  Good for testing this
+# If PRECISION_FACTOR is 0, we only try stuff once.  Good for testing this
 # file in a hurry.
 PRECISION_FACTOR = 1
 
@@ -287,7 +287,7 @@ def rsaTiming():
 
     print "RSA generate (1024 bit,e=100073471)", timeit(
         lambda: pk_generate(1024, 100073471), 10)
-                             
+
     rsa = pk_generate(1024, 100073471)
     print "Pad+RSA public encrypt",
     print timeit((lambda rsa=rsa: pk_encrypt(s70b, rsa)),1000)
@@ -300,7 +300,7 @@ def rsaTiming():
         print "[generating key...]"
         rsa2 = pk_generate(bits)
         t = time()-t
-        print "RSA genrate (%d bit)"%bits, timestr(t)
+        print "RSA generate (%d bit)"%bits, timestr(t)
         enc = pk_encrypt(s70b, rsa2)
         print "Pad+RSA public encrypt (%d bit)"%bits,
         print timeit((lambda rsa2=rsa2: pk_encrypt("zzz", rsa2)),it)
@@ -431,14 +431,14 @@ def buildMessageTiming():
     print "Uncompress (1K, no max)", \
           timeit(lambda c=compressed: uncompressData(c), 1000)
     compressed = compressData("Hello!!!"*(128*28))
-    print "Unompress (28K, no max)", \
+    print "Uncompress (28K, no max)", \
           timeit(lambda c=compressed: uncompressData(c), 1000)
 
     compressed = compressData("Hello!!!"*128)
     print "Uncompress (1K, 1K max)", \
           timeit(lambda c=compressed: uncompressData(c, 1024), 1000)
     compressed = compressData("Hello!!!"*(128*28))
-    print "Unompress (28K, 28K max)", \
+    print "Uncompress (28K, 28K max)", \
           timeit(lambda c=compressed: uncompressData(c, 28<<10), 1000)
 
     payload = ("Junky qoph flags vext crwd zimb."*1024)[:22*1024]
@@ -460,10 +460,12 @@ def buildMessageTiming():
     bh(8,20)
     bh(16,10)
 
+    payload = encodeMessage(payload, 0)[0]
+
     def bm(np1,np2,it,serverinfo=serverinfo,payload=payload):
         tm = timeit_( \
               lambda np1=np1,np2=np2,it=it,serverinfo=serverinfo,
-                      payload=payload: buildForwardMessage(payload,
+                      payload=payload: buildForwardPacket(payload,
                                                500,
                                                "Hello",
                                                serverinfo[:np1],
@@ -479,7 +481,8 @@ def buildMessageTiming():
 #----------------------------------------------------------------------
 def serverQueueTiming():
     print "#================= SERVER QUEUES ====================="
-    Queue = mixminion.server.ServerQueue.Queue
+    import mixminion.Filestore
+    Queue = mixminion.Filestore.MixedStore
     DeliveryQueue = mixminion.server.ServerQueue.DeliveryQueue
     d1 = mix_mktemp()
     q1 = Queue(d1, create=1)
@@ -487,7 +490,7 @@ def serverQueueTiming():
     d2 = mix_mktemp()
     os.mkdir(d2,0700)
     getCommonPRNG().getBytes(1)
-    
+
     #for ln,it in (32*1024,100),(128,400),(1024,400), (32*1024,100):
     for ln,it in ():
         msg = "z"*ln
@@ -552,39 +555,44 @@ def serverProcessTiming():
     server = FakeServerInfo("127.0.0.1", 1, pk, "X"*20)
     sp = PacketHandler([pk], [DummyLog()])
 
-    m_noswap = buildForwardMessage("Hello world", SMTP_TYPE, "f@invalid",
-                                   [server, server], [server, server])
+    payload = encodeMessage("Hello world",0)[0]
+    m_noswap = buildForwardPacket(payload, SMTP_TYPE, "f@invalid",
+                                  [server, server], [server, server])
 
     print "Server process (no swap, no log)", timeit(
-        lambda sp=sp, m_noswap=m_noswap: sp.processMessage(m_noswap), 100)
+        lambda sp=sp, m_noswap=m_noswap: sp.processPacket(m_noswap), 100)
 
-    m_swap = buildForwardMessage("Hello world", SMTP_TYPE, "f@invalid",
+    m_swap = buildForwardPacket(payload, SMTP_TYPE, "f@invalid",
                                  [server], [server, server])
 
     print "Server process (swap, no log)", timeit(
-        lambda sp=sp, m_swap=m_swap: sp.processMessage(m_swap), 100)
+        lambda sp=sp, m_swap=m_swap: sp.processPacket(m_swap), 100)
 
 def encodingTiming():
     print "#=============== END-TO-END ENCODING =================="
     shortP = "hello world"
     prng = AESCounterPRNG()
-    p = _encodePayload(shortP, 0, prng)
+    p = encodeMessage(shortP, 0, prng)[0]
     t = prng.getBytes(20)
-    print "Decode short payload", timeit(
+    print "Check short payload", timeit(
+        lambda p=p: mixminion.BuildMessage._checkPayload(p), 1000)
+    print "Decode without uncompress", timeit(
         lambda p=p,t=t: decodePayload(p, t), 1000)
+    print "Decode short payload", timeit(
+        lambda p=p,t=t: decodePayload(p, t).getUncompressedContents(), 1000)
 
     k20 = prng.getBytes(20*1024)
-    p = _encodePayload(k20, 0, prng)
+    p = encodeMessage(k20, 0, prng)[0]
     t = prng.getBytes(20)
     print "Decode 20K payload", timeit(
-        lambda p=p,t=t: decodePayload(p, t), 1000)
+        lambda p=p,t=t: decodePayload(p, t).getUncompressedContents(), 1000)
 
     comp = "x"*(20*1024)
-    p = _encodePayload(comp, 0, prng)
+    p = encodeMessage(comp, 0, prng)[0]
     t = prng.getBytes(20)
     def decode(p=p,t=t):
         try:
-            decodePayload(p,t)
+            decodePayload(p,t).getUncompressedContents()
         except CompressedDataTooLong:
             pass
     print "Decode overcompressed payload", timeit(decode, 1000)
@@ -595,7 +603,7 @@ def timeEfficiency():
     # Here we compare the time spent in an operation with the time we think
     # is required for its underlying operations, in order to try to measure
     # its efficiency.  If function X is pretty efficient, there's not much
-    # reason to try to optimise its implementation; instead, we need to attack
+    # reason to try to optimize its implementation; instead, we need to attack
     # the functions it uses.
 
     ##### LIONESS
@@ -660,11 +668,12 @@ def timeEfficiency():
     server = FakeServerInfo("127.0.0.1", 1, pk, "X"*20)
     sp = PacketHandler([pk], [DummyLog()])
 
-    m_noswap = buildForwardMessage("Hello world", SMTP_TYPE, "f@invalid",
-                                   [server, server], [server, server])
+    payload = encodeMessage("Hello world",0)[0]
+    m_noswap = buildForwardPacket(payload, SMTP_TYPE, "f@invalid",
+                                  [server, server], [server, server])
 
     sp_ns = timeit_(
-        lambda sp=sp, m_noswap=m_noswap: sp.processMessage(m_noswap), 100)
+        lambda sp=sp, m_noswap=m_noswap: sp.processPacket(m_noswap), 100)
 
     expected = rsa_128b+sha1_hdr+sha1_key*5+aes_2k+lioness_28k+prng_128b
     expected += lioness_2k
@@ -724,7 +733,30 @@ def fileOpsTiming():
     t = time()-t1
     print "          (sync)", timestr(t/100)
 
+#----------------------------------------------------------------------
+def fecTiming():
+    print "#================= FEC =========================="
 
+    r = getCommonPRNG()
+    for k,n,it in [(5,10,1000),
+                   (20,25,300),
+                   (30,40,100),
+                   (40,50,100)
+                   ]:
+        print "FEC (%s/%s)"%(k,n)
+        msg = [ r.getBytes(28*1024) for i in xrange(k) ]
+        fec = _ml.FEC_generate(k,n)
+        tm = timeit_(lambda f=fec, m=msg,k=k: f.encode(k+1,m), it)
+        print "Encode a single 28KB check block:", timestr(tm)
+        print "                (time/(k*28KB)) =", timestr(tm/(k*28)), "/ KB"
+        missing_1 = [ (i, fec.encode(i,msg)) for i in xrange(1,k+1) ]
+        missing_max = [ (i, fec.encode(i,msg)) for i in xrange(n-k,n) ]
+        tm = timeit_(lambda f=fec, m=missing_1: f.decode(m), it)
+        print "              Decode (1 missing):", timestr(tm)
+        print "              (time/(k*28KB*1)) =", timestr(tm/(k*28)), "/ KB"
+        tm = timeit_(lambda f=fec, m=missing_max: f.decode(m), it)
+        print "            Decode (k-n missing):", timestr(tm)
+        print "          (time/(k*28KB*(n-k))) =", timestr(tm/(k*28*(n-k))), "/ KB"
 #----------------------------------------------------------------------
 def testLeaks1():
     print "Trying to leak (sha1,aes,xor,seed,oaep)"
@@ -858,7 +890,6 @@ def testLeaks5_send2():
     import mixminion.MMTPClient
 
     #msg = "X" * 32 * 1024
-    n = 0
     server, listener, messagesIn, keyid = _getMMTPServer(1,port=(TEST_PORT+1))
     #t = threading.Thread(None, testLeaks5_send,
     #                     args=(keyid,))
@@ -874,7 +905,7 @@ def testLeaks5_send2():
 
     import socket
 
-    context = _getTLSContext(0)
+    #context = _getTLSContext(0)
 
     i = 0
     while 0:
@@ -888,10 +919,10 @@ def testLeaks5_send2():
         #tls.get_peer_cert_pk()
         certcache.check(tls, keyid, ("127.0.0.1", TEST_PORT))
         #print certcache.cache
-        
+
         tls.shutdown()
         sock.close()
-            
+
 
 
     while 1:
@@ -968,12 +999,20 @@ def testLeaks6_2():
         tls.shutdown()
         sock.close()
 
+def testLeaks_FEC():
+    inp = [ "aaaaa"*1024, "bbbbb"*1024, "ccccc"*1024 ]
+    while 1:
+        fec = _ml.FEC_generate(3,5)
+        chunks = [ fec.encode(i, inp) for i in xrange(5) ]
+        fec.decode([(i, chunks[i]) for i in xrange(2,5) ])
+
 #----------------------------------------------------------------------
 def timeAll(name, args):
     if 0:
-        testLeaks5_send()
+        timeEfficiency()
         return
-    
+
+    fecTiming()
     cryptoTiming()
     rsaTiming()
     buildMessageTiming()
