@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: ServerList.py,v 1.33 2003/06/12 04:08:17 nickm Exp $
+# $Id: ServerList.py,v 1.47 2003/12/15 05:04:55 nickm Exp $
 
 """mixminion.directory.ServerList
 
@@ -35,8 +35,8 @@ from mixminion.Config import ConfigError
 from mixminion.ServerInfo import ServerDirectory, ServerInfo, \
      _getDirectoryDigestImpl
 
-ACCEPTABLE_CLIENT_VERSIONS = "0.0.4rc3 0.0.4rc4 0.0.4"
-ACCEPTABLE_SERVER_VERSIONS = "0.0.4rc4 0.0.4"
+ACCEPTABLE_CLIENT_VERSIONS = "0.0.6rc1 0.0.6rc2 0.0.6"
+ACCEPTABLE_SERVER_VERSIONS = "0.0.6rc1 0.0.6rc2 0.0.6"
 
 # Make sure that the VERSIONS variables above are parseable.
 for vl in (ACCEPTABLE_CLIENT_VERSIONS.split(),
@@ -89,7 +89,7 @@ class ServerList:
     #          dir-dategenerated.N ...
     #     identity
     #     .lock
-    
+
     def __init__(self, baseDir, idCache=None):
         """Initialize a ServerList to store servers under baseDir/servers,
            creating directories as needed.
@@ -131,6 +131,17 @@ class ServerList:
         finally:
             self._unlock()
 
+    def rebuildIDCache(self):
+        for fn in os.listdir(self.serverIDDir):
+            fname = os.path.join(self.serverIDDir, fn)
+            tp,val = readPickled(fname)
+            if tp != "V0":
+                LOG.warn("Weird file version %s on %s",tp,fname)
+                continue
+            nickname, ident = val
+            ID = mixminion.Crypto.sha1(ident)
+            self.idCache.insertID(nickname, ID)
+
     def learnServerID(self, server):
         """Mark the ID for a server descriptor as the canonical
            identity key associated with that server's nickname."""
@@ -142,7 +153,7 @@ class ServerList:
                 if self.idCache.containsServer(server):
                     LOG.warn("Server %s already known", nickname)
             except mixminion.directory.MismatchedID:
-                raise MixFatalError("Mismatched ID for server %s", nickname)
+                raise MixFatalError("Mismatched ID for server %s" % nickname)
 
             LOG.info("Learning identity for new server %s", nickname)
             self.idCache.insertServer(server)
@@ -220,8 +231,7 @@ class ServerList:
             servers = self.serversByNickname[lcnickname]
             for fn in servers:
                 LOG.info("  Removing %s", fn)
-                os.rename(os.path.join(self.serverDir, fn),
-                          os.path.join(self.archiveDir, fn))
+                _moveServer(self.serverDir, self.archiveDir, fn)
                 del self.servers[fn]
             del self.serversByNickname[lcnickname]
             LOG.info("  (%s servers removed)", len(servers))
@@ -232,11 +242,17 @@ class ServerList:
                           startAt, endAt, extraTime,
                           identityKey,
                           publicationTime=None,
-                          badServers=()):
+                          badServers=(),
+                          excludeServers=()):
         """Generate and sign a new directory, to be effective from <startAt>
            through <endAt>.  It includes all servers that are valid at
            any time between <startAt> and <endAt>+<extraTime>.  The directory
-           is signed with <identityKey> """
+           is signed with <identityKey>.
+
+           Any servers whose nicknames appear in 'badServers' are marked as
+           not recommended; any servers whose nicknames appear in
+           'excludeServers' are left off the directory entirely.
+        """
         try:
             self._lock()
             self.clean()
@@ -245,14 +261,17 @@ class ServerList:
             if previousMidnight(startAt) >= previousMidnight(endAt):
                 raise MixError("Validity range does not contain a full day.")
 
+            excludeServers = [ nickname.lower() for nickname in excludeServers]
+
             # First, sort all servers by nickname.
             includedByNickname =  {}
             for fn, s in self.servers.items():
                 nickname = s.getNickname().lower()
+                if nickname in excludeServers: continue
                 includedByNickname.setdefault(nickname, []).append((s, fn))
 
             # Second, find all servers that are valid for part of the period,
-            # and that aren't superceded for the whole period.
+            # and that aren't superseded for the whole period.
             timeRange = IntervalSet([(previousMidnight(startAt),
                                       endAt+extraTime)])
 
@@ -352,7 +371,7 @@ class ServerList:
         return os.path.join(self.baseDir, "directory")
 
     def clean(self, now=None):
-        """Remove all expired or superceded servers from the active directory.
+        """Remove all expired or superseded servers from the active directory.
         """
         # This algorithm is inefficient: O(N_descs * N_descs_per_nickname).
         # We're just going to ignore that.
@@ -371,7 +390,7 @@ class ServerList:
                 servers = [ s for _, _, s  in servers ]
                 for idx in range(len(servers)):
                     if servers[idx].isSupersededBy(servers[idx+1:]):
-                        removed[fns[idx]] = "superceded"
+                        removed[fns[idx]] = "superseded"
 
             # Find all expired servers.
             for fn, s in self.servers.items():
@@ -384,9 +403,7 @@ class ServerList:
             # Now, do the actual removing.
             for fn, why in removed.items():
                 LOG.info("Removing %s descriptor %s", why, fn)
-                os.rename(os.path.join(self.serverDir, fn),
-                          os.path.join(self.archiveDir, fn))
-
+                _moveServer(self.serverDir, self.archiveDir, fn)
                 del self.servers[fn]
 
             self.__buildNicknameMap()
@@ -407,7 +424,7 @@ class ServerList:
                     LOG.warn("Somehow, a bad server named %s got in our store",
                              filename)
                     LOG.warn(" (Error was: %s)", str(e))
-                    os.rename(path, os.path.join(self.rejectDir, filename))
+                    _moveServer(self.serverDir, self.rejectDir, filename)
 
             # Next, rebuild self.serverIDs:
             self.serverIDs = {}
@@ -434,11 +451,11 @@ class ServerList:
                 try:
                     ident = self.serverIDs[lcn]
                 except KeyError:
-                    raise UIError("No stored key for server %s",
+                    raise UIError("No stored key for server %s" %
                                   s.getNickname())
 
                 if not pk_same_public_key(ident, s.getIdentity()):
-                    raise UIError("Inconsistent stored key for server %s",
+                    raise UIError("Inconsistent stored key for server %s" %
                                   s.getNickname())
 
             # Then, rebuild self.serversByNickname
@@ -464,6 +481,18 @@ class ServerList:
         self.lockfile.release()
         self.rlock.release()
 
+def _moveServer(directory1, directory2, fname):
+    """Move a server contained in directory1/fname into directory2, chosing
+       a new name for it as needed, and avoiding duplicates."""
+    fname1 = os.path.join(directory1, fname)
+    if "-" in fname:
+        nn = fname[:fname.find("-")]
+    else:
+        nn = "UNKNOWN"
+    s = readPossiblyGzippedFile(fname1)
+    _writeServer(directory2, s, nn)
+    os.unlink(fname1)
+
 def _writeServer(directory, contents, nickname, mode=0600):
     """Write a server descriptor in 'contents' into a new file in
        'directory'.  The file will have permissions 'mode', and a name
