@@ -1,13 +1,13 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: PacketHandler.py,v 1.17 2003/06/06 06:04:58 nickm Exp $
+# $Id: PacketHandler.py,v 1.34 2003/12/08 02:22:56 nickm Exp $
 
-"""mixminion.PacketHandler: Code to process mixminion packets on a server"""
+"""mixminion.server.PacketHandler: Code to process mixminion packets"""
 
 import binascii
 import threading
 import types
 
-from mixminion.Common import encodeBase64, formatBase64
+from mixminion.Common import encodeBase64, formatBase64, LOG
 import mixminion.Crypto as Crypto
 import mixminion.Packet as Packet
 import mixminion.Common as Common
@@ -26,7 +26,7 @@ class PacketHandler:
     """Class to handle processing packets.  Given an incoming packet,
        it removes one layer of encryption, does all necessary integrity
        checks, swaps headers if necessary, re-pads, and decides whether
-       to drop the message, relay the message, or send the message to
+       to drop the packet, relay the packet, or send the packet to
        an exit handler."""
     ## Fields:
     # privatekeys: a list of 2-tuples of
@@ -90,11 +90,11 @@ class PacketHandler:
         finally:
             self.lock.release()
 
-    def processMessage(self, msg):
-        """Given a 32K mixminion message, processes it completely.
+    def processPacket(self, msg):
+        """Given a 32K mixminion packet, processes it completely.
 
            Return one of:
-                    None [if the mesesage should be dropped.]
+                    None [if the packet should be dropped.]
                     a DeliveryPacket object
                     a RelayedPacket object
 
@@ -103,16 +103,19 @@ class PacketHandler:
            unhandleable.
 
            WARNING: This implementation does nothing to prevent timing
-           attacks: dropped messages, messages with bad digests, replayed
-           messages, and exit messages are all processed faster than
-           forwarded messages.  You must prevent timing attacks elsewhere."""
+           attacks: dropped packets, packets with bad digests, replayed
+           packets, and exit packets are all processed faster than
+           forwarded packets.  You must prevent timing attacks elsewhere."""
 
         # Break into headers and payload
-        msg = Packet.parseMessage(msg)
-        header1 = Packet.parseHeader(msg.header1)
+        pkt = Packet.parsePacket(msg)
+        header1 = Packet.parseHeader(pkt.header1)
         encSubh = header1[:Packet.ENC_SUBHEADER_LEN]
         header1 = header1[Packet.ENC_SUBHEADER_LEN:]
 
+        assert len(header1) == Packet.HEADER_LEN - Packet.ENC_SUBHEADER_LEN
+        assert len(header1) == (128*16) - 256 == 1792
+
         # Try to decrypt the first subheader.  Try each private key in
         # order.  Only fail if all private keys fail.
         subh = None
@@ -132,6 +135,9 @@ class PacketHandler:
             # most-recently-received error.
             raise e
 
+        if len(subh) != Packet.MAX_SUBHEADER_LEN:
+            raise ContentError("Bad length in RSA-encrypted part of subheader")
+
         subh = Packet.parseSubheader(subh) #may raise ParseError
 
         # Check the version: can we read it?
@@ -142,13 +148,13 @@ class PacketHandler:
         if subh.digest != Crypto.sha1(header1):
             raise ContentError("Invalid digest")
 
-        # Get ready to generate message keys.
+        # Get ready to generate packet keys.
         keys = Crypto.Keyset(subh.secret)
 
         # Replay prevention
-        replayhash = keys.get(Crypto.REPLAY_PREVENTION_MODE, 20)
+        replayhash = keys.get(Crypto.REPLAY_PREVENTION_MODE, Crypto.DIGEST_LEN)
         if hashlog.seenHash(replayhash):
-            raise ContentError("Duplicate message detected.")
+            raise ContentError("Duplicate packet detected.")
         else:
             hashlog.logHash(replayhash)
 
@@ -169,6 +175,12 @@ class PacketHandler:
                                Packet.OAEP_OVERHEAD + Packet.MIN_SUBHEADER_LEN
                                + subh.routinglen)
 
+        assert len(header1) == (Packet.HEADER_LEN - Packet.ENC_SUBHEADER_LEN
+                             + Packet.OAEP_OVERHEAD+Packet.MIN_SUBHEADER_LEN
+                                + subh.routinglen)
+        assert len(header1) == 1792 + 42 + 42 + subh.routinglen == \
+               1876 + subh.routinglen
+
         # Decrypt the rest of header 1, encrypting the padding.
         header1 = Crypto.ctr_crypt(header1, header_sec_key)
 
@@ -179,10 +191,16 @@ class PacketHandler:
             subh.appendOverflow(header1[:overflowLength])
             header1 = header1[overflowLength:]
 
+        assert len(header1) == (
+            1876 + subh.routinglen
+            - max(0,subh.routinglen-Packet.MAX_ROUTING_INFO_LEN))
+
         header1 = subh.underflow + header1
 
+        assert len(header1) == Packet.HEADER_LEN
+
         # Decrypt the payload.
-        payload = Crypto.lioness_decrypt(msg.payload,
+        payload = Crypto.lioness_decrypt(pkt.payload,
                               keys.getLionessKeys(Crypto.PAYLOAD_ENCRYPT_MODE))
 
         # If we're an exit node, there's no need to process the headers
@@ -194,17 +212,18 @@ class PacketHandler:
 
         # If we're not an exit node, make sure that what we recognize our
         # routing type.
-        if rt not in (Packet.SWAP_FWD_TYPE, Packet.FWD_TYPE):
+        if rt not in (Packet.SWAP_FWD_IPV4_TYPE, Packet.FWD_IPV4_TYPE,
+                      Packet.SWAP_FWD_HOST_TYPE, Packet.FWD_HOST_TYPE):
             raise ContentError("Unrecognized Mixminion routing type")
 
         # Decrypt header 2.
-        header2 = Crypto.lioness_decrypt(msg.header2,
+        header2 = Crypto.lioness_decrypt(pkt.header2,
                            keys.getLionessKeys(Crypto.HEADER_ENCRYPT_MODE))
 
         # If we're the swap node, (1) decrypt the payload with a hash of
         # header2... (2) decrypt header2 with a hash of the payload...
         # (3) and swap the headers.
-        if rt == Packet.SWAP_FWD_TYPE:
+        if Packet.typeIsSwap(rt):
             hkey = Crypto.lioness_keys_from_header(header2)
             payload = Crypto.lioness_decrypt(payload, hkey)
 
@@ -214,23 +233,23 @@ class PacketHandler:
             header1, header2 = header2, header1
 
         # Build the address object for the next hop
-        address = Packet.parseIPV4Info(subh.routinginfo)
+        address = Packet.parseRelayInfoByType(rt, subh.routinginfo)
 
-        # Construct the message for the next hop.
-        msg = Packet.Message(header1, header2, payload).pack()
+        # Construct the packet for the next hop.
+        pkt = Packet.Packet(header1, header2, payload).pack()
 
-        return RelayedPacket(address, msg)
+        return RelayedPacket(address, pkt)
 
 class RelayedPacket:
     """A packet that is to be relayed to another server; returned by
-       returned by PacketHandler.processMessage."""
+       returned by PacketHandler.processPacket."""
     ## Fields:
     # address -- an instance of IPV4Info
     # msg -- a 32K packet.
     def __init__(self, address, msg):
-        """Create a new packet, given an instance of IPV4Info and a 32K
-           packet."""
-        assert isinstance(address, Packet.IPV4Info)
+        """Create a new packet, given an instance of IPV4Info or
+           MMTPHostInfo and a 32K packet."""
+        assert isinstance(address, Packet.IPV4Info) or isinstance(address, Packet.MMTPHostInfo)
         assert len(msg) == 1<<15
         self.address = address
         self.msg = msg
@@ -240,8 +259,8 @@ class RelayedPacket:
         return 0
 
     def getAddress(self):
-        """Return an instance of IPV4Info indicating the address where this
-           packet is to be delivered."""
+        """Return an instance of IPV4Info or MMTPHostInfo indicating
+           the address where this packet is to be delivered."""
         return self.address
 
     def getPacket(self):
@@ -250,7 +269,7 @@ class RelayedPacket:
 
 class DeliveryPacket:
     """A packet that is to be delivered via some exit module; returned by
-       PacketHandler.processMessage"""
+       PacketHandler.processPacket"""
     ##Fields:
     # exitType -- a 2-byte integer indicating which exit module to use.
     # address -- a string encoding the address to deliver to.
@@ -262,12 +281,21 @@ class DeliveryPacket:
     # type -- until decode is called, None.  After decode is called,
     #     one of 'plain' (plaintext message), 'long' (overcompressed message),
     #     'enc' (encrypted message), or 'err' (malformed message).
+    # headers -- a map from key to value for the delivery headers in
+    #     this message's payload.  In the case of a fragment, or a
+    #     non-plaintext message, the map is empty.
+    # isfrag -- Is this packet a fragment of a complete message?  If so, the
+    #     type must be 'plain'.
+    # dPayload -- An instance of mixminion.Packet.Payload for this object.
+    # error -- None, or a string containing an error encountered while trying
+    #     to decode the payload.
     def __init__(self, routingType, routingInfo, applicationKey,
                  tag, payload):
         """Construct a new DeliveryPacket."""
         assert 0 <= routingType <= 0xFFFF
         assert len(applicationKey) == 16
-        assert len(tag) == 20
+        #assert len(tag) == 20 #XXXX007 make tag system sane.
+        assert len(tag) == 20 or routingType == Packet.FRAGMENT_TYPE
         assert len(payload) == 28*1024
         self.exitType = routingType
         self.address = routingInfo
@@ -276,6 +304,31 @@ class DeliveryPacket:
         self.payload = payload
         self.contents = None
         self.type = None
+        self.headers = None
+        self.isfrag = 0
+        self.dPayload = None
+        self.error = None
+
+    def __getstate__(self):
+        return "V0", self.__dict__
+
+    def __setstate__(self, state):
+        if type(state) == types.DictType:
+            #XXXX007 remove this case.  (Not used since 0.0.5alpha)
+            LOG.warn("Found ancient packet format.")
+            self.__dict__.update(state)
+            if not hasattr(self, 'isfrag'):
+                self.isfrag = 0
+            if not hasattr(self, 'dPayload'):
+                self.dPayload = None
+            if not hasattr(self, 'error'):
+                self.error = None
+            if not hasattr(self, 'headers'):
+                self.headers = {}
+        elif state[0] == 'V0':
+            self.__dict__.update(state[1])
+        else:
+            raise MixError("Unrecognized state version %s" % state[0])
 
     def isDelivery(self):
         """Return true iff this packet is a delivery (non-relay) packet."""
@@ -292,6 +345,11 @@ class DeliveryPacket:
         if self.type is None: self.decode()
         return self.contents
 
+    def getDecodedPayload(self):
+        """Return an instance of mixminion.Packet.Payload for this packet."""
+        if self.type is None: self.decode()
+        return self.dPayload
+
     def isPlaintext(self):
         """Return true iff this packet is a plaintext, forward packet."""
         if self.type is None: self.decode()
@@ -303,6 +361,11 @@ class DeliveryPacket:
         if self.type is None: self.decode()
         return self.type == 'long'
 
+    def isFragment(self):
+        """Return true iff this packet is part of a fragmented message."""
+        if self.type is None: self.decode()
+        return self.isfrag
+
     def isEncrypted(self):
         """Return true iff this packet may be an encrypted forward or
            reply packet."""
@@ -327,23 +390,34 @@ class DeliveryPacket:
         message = self.payload
         self.contents = None
         try:
-            self.contents = mixminion.BuildMessage.decodePayload(message,
+            self.dPayload = mixminion.BuildMessage.decodePayload(message,
                                                                  self.tag)
-            if self.contents is None:
+            if self.dPayload is None:
                 # encrypted message
                 self.type = 'enc'
                 self.contents = message
-            else:
-                # forward message
+                self.headers = {}
+            elif self.dPayload.isSingleton():
+                # forward message, singleton.
                 self.type = 'plain'
-                # self.contents is right
+                body = self.dPayload.getUncompressedContents()
+                self.contents, self.headers = \
+                               Packet.parseMessageAndHeaders(body)
+            else:
+                # forward message, fragment.
+                self.isfrag = 1
+                self.type = 'plain'
+                self.contents = message
+                self.headers = {}
         except Packet.CompressedDataTooLong, _:
-            self.contents = (mixminion.Packet.parsePayload(message)
-                                             .getContents())
+            self.contents = Packet.parsePayload(message).getContents()
             self.type = 'long'
-        except MixError:
+            self.headers = {}
+        except MixError, e:
             self.contents = message
+            self.error = str(e)
             self.type = 'err'
+            self.headers = {}
 
         self.payload = None
 
@@ -358,6 +432,15 @@ class DeliveryPacket:
         else:
             return encodeBase64(self.contents)
 
+    def getHeaders(self):
+        """Return a dict containing the headers for this message."""
+        if self.type is None:
+            self.decode()
+        if self.headers is None:
+            LOG.warn("getHeaders found no decoded headers")
+            return {}
+        return self.headers
+
     def getAsciiTag(self):
         """Return a base64-representation of this message's decoding handle."""
         return formatBase64(self.tag)
@@ -373,6 +456,9 @@ class DeliveryPacket:
         elif self.isPrintingAscii():
             assert self.isPlaintext()
             tp = 'TXT'
+        elif self.isFragment():
+            assert self.isPlaintext()
+            tp = 'FRAG'
         else:
             assert self.isPlaintext()
             tp = 'BIN'
