@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: ServerConfig.py,v 1.33 2003/06/05 18:41:40 nickm Exp $
+# $Id: ServerConfig.py,v 1.45 2003/12/08 02:26:39 nickm Exp $
 
 """Configuration format for server configuration files.
 
@@ -8,6 +8,7 @@
 __all__ = [ "ServerConfig" ]
 
 import operator
+import os
 
 import mixminion.Config
 import mixminion.server.Modules
@@ -20,10 +21,12 @@ class ServerConfig(mixminion.Config._ConfigFile):
     #   moduleManager
     #
     _restrictFormat = 0
+
     def __init__(self, fname=None, string=None, moduleManager=None):
         # We use a copy of SERVER_SYNTAX, because the ModuleManager will
         # mess it up.
         self._syntax = SERVER_SYNTAX.copy()
+        self.CODING_FNS = CODING_FNS
 
         if moduleManager is None:
             self.moduleManager = mixminion.server.Modules.ModuleManager()
@@ -38,7 +41,7 @@ class ServerConfig(mixminion.Config._ConfigFile):
             entries = self._sectionEntries
             return len([e for e in entries[section] if e[0] == ent]) != 0
 
-        # Pre-emptively configure the log before validation, so we don't
+        # Preemptively configure the log before validation, so we don't
         # write to the terminal if we've been asked not to.
         if not self['Server'].get("EchoMessages", 0):
             LOG.handlers = []
@@ -57,8 +60,10 @@ class ServerConfig(mixminion.Config._ConfigFile):
             LOG.warn("Encrypted private keys not yet implemented")
         if server['PublicKeyLifetime'].getSeconds() < 24*60*60:
             raise ConfigError("PublicKeyLifetime must be at least 1 day.")
-        if server['PublicKeyOverlap'].getSeconds() > 6*60*60:
-            raise ConfigError("PublicKeyOverlap must be <= 6 hours")
+        if server['PublicKeyOverlap'].getSeconds() < 6*60*60:
+            raise ConfigError("PublicKeyOverlap must be >= 6 hours")
+        if server['PublicKeyOverlap'].getSeconds() > 72*60*60:
+            raise ConfigError("PublicKeyOverlap must be <= 72 hours")
 
         if _haveEntry(self, 'Server', 'Mode'):
             LOG.warn("Mode specification is not yet supported.")
@@ -120,10 +125,10 @@ class ServerConfig(mixminion.Config._ConfigFile):
         if server['LogLevel'] in ('TRACE', 'DEBUG'):
             reasons.append("Log is too verbose")
         if server['LogStats'] and server['StatsInterval'].getSeconds() \
-               < 24*60*60:
+               < 2*60*60:
             reasons.append("StatsInterval is too short")
-        if not server["EncryptIdentityKey"]:
-            reasons.append("Identity key is not encrypted")
+        #if not server["EncryptIdentityKey"]:
+        #    reasons.append("Identity key is not encrypted")
         # ???? Pkey lifetime, sloppiness?
         if server["MixAlgorithm"] not in _SECURE_MIX_RULES:
             reasons.append("Mix algorithm is not secure")
@@ -142,6 +147,32 @@ class ServerConfig(mixminion.Config._ConfigFile):
 
         return reasons
 
+    def getConfigurationSummary(self):
+        """Return a human-readable description of this server's configuration,
+           for inclusion in the testing section of the server descriptor."""
+        res = []
+        for section,entries in [
+            ("Server", ['LogLevel', 'LogStats', 'StatsInterval',
+                        'PublicKeyOverlap', 'Mode', 'MixAlgorithm',
+                        'MixInterval', 'MixPoolRate', 'MixPoolMinSize',
+                        'Timeout',]),
+            ("Outgoing/MMTP", ['Retry']),
+            ("Delivery/SMTP",
+             ['Enabled', 'Retry', 'SMTPServer', 'ReturnAddress', 'FromTag',
+              'SubjectLine', 'MaximumSize']),
+            ("Delivery/SMTP-Via-Mixmaster",
+             ['Enabled', 'Retry', 'Server', 'ReturnAddress', 'FromTag',
+              'SubjectLine', 'MaximumSize']),
+            ("Delivery/Fragmented",
+             ['Enabled', 'MaximumSize','MaximumInterval']),
+            ]:
+            sec = self[section]
+            for k in entries:
+                v = sec.get(k, None)
+                if v:
+                    res.append("%s/%s=%r"%(section,k,v))
+        return "; ".join(res)
+
     def validateRetrySchedule(self, sectionName, entryName='Retry'):
         """Check whether the retry schedule in self[sectionName][entryName]
            is reasonable.  Warn or raise ConfigError if it isn't.  Ignore
@@ -153,6 +184,52 @@ class ServerConfig(mixminion.Config._ConfigFile):
         mixInterval = self['Server']['MixInterval'].getSeconds()
         _validateRetrySchedule(mixInterval, entry, sectionName)
 
+    def _get_fname(self, sec, ent, defaultRel):
+        """Helper function.  Returns the filename in self[sec][ent],
+           but defaults to ${BASEDIR}/defaultRel.  All relative paths
+           are also interpreted relative to ${BASEDIR}.
+        """
+        raw = self[sec].get(ent)
+        homedir = self.getBaseDir()
+        if not raw:
+            return os.path.join(homedir, defaultRel)
+        if os.path.isabs(raw):
+            return raw
+        else:
+            return os.path.join(homedir, raw)
+
+    def getBaseDir(self):
+        """Return the base directory for this configuration."""
+        v = self["Server"]["BaseDir"]
+        if v is None:
+            v = self["Server"]["Homedir"]
+        if v is None:
+            LOG.warn("Defaulting base directory to /var/spool/minion; this will change.")
+            v = "/var/spool/minion"
+        return v
+
+    def getLogFile(self):
+        """Return the configured logfile location."""
+        return self._get_fname("Server", "LogFile", "log")
+    def getStatsFile(self):
+        """Return the configured stats file location."""
+        return self._get_fname("Server", "StatsFile", "stats")
+    def getKeyDir(self):
+        """Return the configured key directory"""
+        return self._get_fname("Server", "KeyDir", "keys")
+    def getWorkDir(self):
+        """Return the configured work directory"""
+        return self._get_fname("Server", "WorkDir", "work")
+    def getPidFile(self):
+        """Return the configured PID file location"""
+        return self._get_fname("Server", "PidFile", "pid")
+    def getQueueDir(self):
+        """Return the configured queue directory."""
+        if self["Server"]["QueueDir"] is None:
+            return os.path.join(self.getWorkDir(), 'queues')
+        else:
+            return self._get_fname("Server", "QueueDir", "work/queues")
+
 def _validateRetrySchedule(mixInterval, schedule, sectionName):
     """Backend for ServerConfig.validateRetrySchedule -- separated for testing.
 
@@ -226,61 +303,72 @@ def _parseFraction(frac):
     return ratio
 
 # alias to make the syntax more terse.
-C = mixminion.Config
 
 SERVER_SYNTAX =  {
-        'Host' : C.ClientConfig._syntax['Host'],
+        'Host' : mixminion.Config.ClientConfig._syntax['Host'],
         'Server' : { '__SECTION__' : ('REQUIRE', None, None),
-                     'Homedir' :
-                         ('ALLOW', C._parseFilename, "/var/spool/minion"),
-                     'LogFile' : ('ALLOW', None, None),
-                     'LogLevel' : ('ALLOW', C._parseSeverity, "WARN"),
-                     'EchoMessages' : ('ALLOW', C._parseBoolean, "no"),
-                     'Daemon' : ('ALLOW', C._parseBoolean, "no"),
-                     'LogStats' : ('ALLOW', C._parseBoolean, 'yes'),
-                     'StatsInterval' : ('ALLOW', C._parseInterval,
+                     'BaseDir' : ("ALLOW", "filename", None),
+                     'Homedir' : ('ALLOW', "filename", None),
+                     'LogFile' : ('ALLOW', "filename", None),
+                     'StatsFile' : ('ALLOW', "filename", None),
+                     'KeyDir' : ('ALLOW', "filename", None),
+                     'WorkDir' : ('ALLOW', "filename", None),
+                     'QueueDir' : ('ALLOW', "filename", None),
+                     'PidFile' : ('ALLOW', "filename", None),
+
+                     'LogLevel' : ('ALLOW', "severity", "WARN"),
+                     'EchoMessages' : ('ALLOW', "boolean", "no"),
+                     'Daemon' : ('ALLOW', "boolean", "no"),
+                     'LogStats' : ('ALLOW', "boolean", 'yes'),
+                     'StatsInterval' : ('ALLOW', "interval",
                                         "1 day"),
-                     'StatsFile' : ('ALLOW', None, None),
-                     'EncryptIdentityKey' :('ALLOW', C._parseBoolean, "no"),
-                     'IdentityKeyBits': ('ALLOW', C._parseInt, "2048"),
-                     'PublicKeyLifetime' : ('ALLOW', C._parseInterval,
+                     'EncryptIdentityKey' :('ALLOW', "boolean", "no"),
+                     'IdentityKeyBits': ('ALLOW', "int", "2048"),
+                     'PublicKeyLifetime' : ('ALLOW', "interval",
                                             "30 days"),
-                     'PublicKeyOverlap': ('ALLOW', C._parseInterval,
-                                          "5 minutes"),
-                     'EncryptPrivateKey' : ('ALLOW', C._parseBoolean, "no"),
-                     'Mode' : ('REQUIRE', C._parseServerMode, "local"),
-                     'Nickname': ('REQUIRE', C._parseNickname, None),
+                     'PublicKeyOverlap': ('ALLOW', "interval",
+                                          "24 hours"),
+                     'EncryptPrivateKey' : ('ALLOW', "boolean", "no"),
+                     'Mode' : ('REQUIRE', "serverMode", "local"),
+                     'Nickname': ('REQUIRE', "nickname", None),
                      'Contact-Email': ('ALLOW', None, None),
                      'Comments': ('ALLOW', None, None),
                      'ModulePath': ('ALLOW', None, None),
                      'Module': ('ALLOW*', None, None),
-                     'MixAlgorithm' : ('ALLOW', _parseMixRule, "Timed"),
-                     'MixInterval' : ('ALLOW', C._parseInterval, "30 min"),
-                     'MixPoolRate' : ('ALLOW', _parseFraction, "60%"),
-                     'MixPoolMinSize' : ('ALLOW', C._parseInt, "5"),
-		     'Timeout' : ('ALLOW', C._parseInterval, "5 min"),
-                     '__DEBUG_GC' : ('ALLOW', C._parseBoolean, "no"),
+                     'MixAlgorithm' : ('ALLOW', "mixRule", "Timed"),
+                     'MixInterval' : ('ALLOW', "interval", "30 min"),
+                     'MixPoolRate' : ('ALLOW', "fraction", "60%"),
+                     'MixPoolMinSize' : ('ALLOW', "int", "5"),
+		     'Timeout' : ('ALLOW', "interval", "5 min"),
                      },
         'DirectoryServers' : { # '__SECTION__' : ('REQUIRE', None, None),
                                'ServerURL' : ('ALLOW*', None, None),
                                'PublishURL' : ('ALLOW*', None, None),
-                               'Publish' : ('ALLOW', C._parseBoolean, "no"),
-                               'MaxSkew' : ('ALLOW', C._parseInterval,
+                               'Publish' : ('ALLOW', "boolean", "no"),
+                               'MaxSkew' : ('ALLOW', "interval",
                                             "10 minutes",) },
         # FFFF Generic multi-port listen/publish options.
-        'Incoming/MMTP' : { 'Enabled' : ('REQUIRE', C._parseBoolean, "no"),
-                            'IP' : ('ALLOW', C._parseIP, "0.0.0.0"),
-                          'Port' : ('ALLOW', C._parseInt, "48099"),
-                          'ListenIP' : ('ALLOW', C._parseIP, None),
-                          'ListenPort' : ('ALLOW', C._parseInt, None),
-  		          'Allow' : ('ALLOW*', C._parseAddressSet_allow, None),
-                          'Deny' : ('ALLOW*', C._parseAddressSet_deny, None)
+        'Incoming/MMTP' : { 'Enabled' : ('REQUIRE', "boolean", "no"),
+                            #XXXX007/8 deprecate or remove IP.
+                            'IP' : ('ALLOW', "IP", "0.0.0.0"),
+                          'Hostname' : ('ALLOW', "host", None),
+                          'Port' : ('ALLOW', "int", "48099"),
+                          'ListenIP' : ('ALLOW', "IP", None),
+                          'ListenPort' : ('ALLOW', "int", None),
+                          'ListenIP6' : ('ALLOW', "IP6", None),
+  		          'Allow' : ('ALLOW*', "addressSet_allow", None),
+                          'Deny' : ('ALLOW*', "addressSet_deny", None)
 			 },
-        'Outgoing/MMTP' : { 'Enabled' : ('REQUIRE', C._parseBoolean, "no"),
-                            'Retry' : ('ALLOW', C._parseIntervalList,
+        'Outgoing/MMTP' : { 'Enabled' : ('REQUIRE', "boolean", "no"),
+                            'Retry' : ('ALLOW', "intervalList",
                                  "every 1 hour for 1 day, 7 hours for 5 days"),
-                          'Allow' : ('ALLOW*', C._parseAddressSet_allow, None),
-                          'Deny' : ('ALLOW*', C._parseAddressSet_deny, None) },
+                          'Allow' : ('ALLOW*', "addressSet_allow", None),
+                          'Deny' : ('ALLOW*', "addressSet_deny", None) },
         # FFFF Missing: Queue-Size / Queue config options
         # FFFF         listen timeout??
         }
+
+CODING_FNS = mixminion.Config._ConfigFile.CODING_FNS.copy()
+CODING_FNS.update({'mixRule':(_parseMixRule,str),
+                   'fraction':(_parseFraction,
+                               lambda r: "%.2f%%"%(100.*r))})
