@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: ServerKeys.py,v 1.44 2003/06/10 10:40:21 nickm Exp $
+# $Id: ServerKeys.py,v 1.61 2003/12/03 23:18:53 nickm Exp $
 
 """mixminion.ServerKeys
 
@@ -22,6 +22,7 @@ import urllib2
 
 import mixminion._minionlib
 import mixminion.Crypto
+import mixminion.NetUtils
 import mixminion.Packet
 import mixminion.server.HashLog
 import mixminion.server.MMTPServer
@@ -30,9 +31,10 @@ import mixminion.server.ServerMain
 from mixminion.ServerInfo import ServerInfo, PACKET_KEY_BYTES, MMTP_KEY_BYTES,\
      signServerInfo
 from mixminion.Common import AtomicFile, LOG, MixError, MixFatalError, \
-     ceilDiv, createPrivateDir, checkPrivateFile, formatBase64, formatDate, \
-     formatTime, previousMidnight, readFile, secureDelete, tryUnlink, \
-     writeFile
+     ceilDiv, createPrivateDir, checkPrivateFile, englishSequence, \
+     formatBase64, formatDate, formatTime, previousMidnight, readFile, \
+     secureDelete, tryUnlink, UIError, writeFile
+from mixminion.Config import ConfigError
 
 #----------------------------------------------------------------------
 
@@ -69,6 +71,7 @@ class ServerKeyring:
     #      have keys on disk.
     # currentKeys: None, if we haven't checked for currently live keys, or
     #      a list of currently live ServerKeyset objects.
+    # dhFile: pathname to file holding diffie-helman parameters.
     # _lock: A lock to prevent concurrent key generation or rotation.
 
     def __init__(self, config):
@@ -79,9 +82,10 @@ class ServerKeyring:
     def configure(self, config):
         "Set up a ServerKeyring from a config object"
         self.config = config
-        self.homeDir = config['Server']['Homedir']
-        self.keyDir = os.path.join(self.homeDir, 'keys')
-        self.hashDir = os.path.join(self.homeDir, 'work', 'hashlogs')
+        self.homeDir = config.getBaseDir()
+        self.keyDir = config.getKeyDir()
+        self.hashDir = os.path.join(config.getWorkDir(), 'hashlogs')
+        self.dhFile = os.path.join(config.getWorkDir(), 'tls', 'dhparam')
         self.keyOverlap = config['Server']['PublicKeyOverlap'].getSeconds()
         self.nextUpdate = None
         self.currentKeys = None
@@ -95,6 +99,7 @@ class ServerKeyring:
            are invalid.
            """
         self.keySets = []
+        badKeySets = []
         firstKey = sys.maxint
         lastKey = 0
 
@@ -103,7 +108,7 @@ class ServerKeyring:
         if not os.path.exists(self.keyDir):
             LOG.info("Creating server keystore at %s", self.keyDir)
             createPrivateDir(self.keyDir)
-        
+
         # Iterate over the entires in HOME/keys
         for dirname in os.listdir(self.keyDir):
             # Skip any that aren't directories named "key_INT"
@@ -126,13 +131,38 @@ class ServerKeyring:
 
             # Find the server descriptor...
             keyset = ServerKeyset(self.keyDir, keysetname, self.hashDir)
-            t1, t2 = keyset.getLiveness()
-            self.keySets.append( (t1, t2, keyset) )
-                
-            LOG.trace("Found key %s (valid from %s to %s)",
-                      dirname, formatDate(t1), formatDate(t2))
+            ok = 1
+            try:
+                keyset.checkKeys()
+            except MixError, e:
+                LOG.warn("Error checking private keys in keyset %s: %s",
+                         keysetname, str(e))
+                ok = 0
 
-        LOG.debug("Found %s keys.", len(self.keySets))
+            try:
+                if ok:
+                    keyset.getServerDescriptor()
+            except (ConfigError, IOError), e:
+                LOG.warn("Key set %s has invalid/missing descriptor: %s",
+                         keysetname, str(e))
+                ok = 0
+
+            if ok:
+                t1, t2 = keyset.getLiveness()
+                self.keySets.append( (t1, t2, keyset) )
+
+                LOG.trace("Found key %s (valid from %s to %s)",
+                          dirname, formatDate(t1), formatDate(t2))
+            else:
+                badKeySets.append(keyset)
+
+        LOG.debug("Found %s keysets: %s were incomplete or invalid.",
+                  len(self.keySets), len(badKeySets))
+
+        if badKeySets:
+            LOG.warn("Removing %s invalid keysets", len(badKeySets))
+        for b in badKeySets:
+            b.delete()
 
         # Now, sort the key intervals by starting time.
         self.keySets.sort()
@@ -152,8 +182,8 @@ class ServerKeyring:
 
     def checkDescriptorConsistency(self, regen=1):
         """Check whether the server descriptors in this keyring are
-           consistent with the server's configuration.  If 'regen' are true,
-           inconsistent descriptors are regenerated."""
+           consistent with the server's configuration.  If 'regen' is
+           true, inconsistent descriptors are regenerated."""
         identity = None
         state = []
         for _,_,ks in self.keySets:
@@ -164,10 +194,10 @@ class ServerKeyring:
 
         if not state:
             return
-        
+
         LOG.warn("Some generated keysets do not match "
                   "current configuration...")
-        
+
         for ok, ks in state:
             va,vu = ks.getLiveness()
             LOG.warn("Keyset %s (%s--%s):",ks.keyname,formatTime(va,1),
@@ -241,10 +271,9 @@ class ServerKeyring:
             LOG.warn("Removing identity key")
             secureDelete([fn], blocking=1)
 
-        dhfile = os.path.join(self.homeDir, 'work', 'tls', 'dhparam')
-        if os.path.exists('dhfile'):
+        if os.path.exists(self.dhFile):
             LOG.info("Removing diffie-helman parameters file")
-            secureDelete([dhfile], blocking=1)
+            secureDelete([self.dhFile], blocking=1)
 
     def createKeysAsNeeded(self,now=None):
         """Generate new keys and descriptors as needed, so that the next
@@ -257,12 +286,14 @@ class ServerKeyring:
 
         if self.keySets:
             lastExpiry = self.keySets[-1][1]
+            if lastExpiry < now:
+                lastExpiry = now
         else:
             lastExpiry = now
 
         needToCoverUntil = now+PUBLICATION_LATENCY+PREPUBLICATION_INTERVAL
         timeToCover = needToCoverUntil-lastExpiry
-        
+
         lifetime = self.config['Server']['PublicKeyLifetime'].getSeconds()
         nKeys = int(ceilDiv(timeToCover, lifetime))
 
@@ -280,6 +311,8 @@ class ServerKeyring:
         if startAt is None:
             if self.keySets:
                 startAt = self.keySets[-1][1]+60
+                if startAt < time.time():
+                    startAt = time.time()+60
             else:
                 startAt = time.time()+60
 
@@ -352,8 +385,9 @@ class ServerKeyring:
         self.checkKeys()
 
     def getDeadKeys(self,now=None):
-        """DOCDOC
-           doesn't checkKeys.
+        """Helper function: return a list of (informative-message, keyset
+           object) for each expired keyset in the keystore.  Does not rescan
+           the keystore or remove dead keys.
         """
         if now is None:
             now = time.time()
@@ -407,20 +441,19 @@ class ServerKeyring:
     def _getDHFile(self):
         """Return the filename for the diffie-helman parameters for the
            server.  Creates the file if it doesn't yet exist."""
-        dhdir = os.path.join(self.homeDir, 'work', 'tls')
+        dhdir = os.path.split(self.dhFile)[0]
         createPrivateDir(dhdir)
-        dhfile = os.path.join(dhdir, 'dhparam')
-        if not os.path.exists(dhfile):
+        if not os.path.exists(self.dhFile):
             # ???? This is only using 512-bit Diffie-Hellman!  That isn't
             # ???? remotely enough.
             LOG.info("Generating Diffie-Helman parameters for TLS...")
-            mixminion._minionlib.generate_dh_parameters(dhfile, verbose=0)
+            mixminion._minionlib.generate_dh_parameters(self.dhFile, verbose=0)
             LOG.info("...done")
         else:
             LOG.debug("Using existing Diffie-Helman parameter from %s",
-                           dhfile)
+                           self.dhFile)
 
-        return dhfile
+        return self.dhFile
 
     def _getTLSContext(self, keys=None):
         """Create and return a TLS context from the currently live key."""
@@ -447,7 +480,7 @@ class ServerKeyring:
                  len(deadKeys), " ".join(deadKeyNames))
         if mmtpServer is not None:
             context = self._getTLSContext(keys[-1])
-            mmtpServer.setContext(context)
+            mmtpServer.setServerContext(context)
         if packetHandler is not None:
             packetKeys = []
             hashLogs = []
@@ -518,7 +551,7 @@ class ServerKeyring:
         desc = keys.getServerDescriptor()
         return (desc['Incoming/MMTP']['IP'],
                 desc['Incoming/MMTP']['Port'],
-                desc['Incoming/MMTP']['Key-Digest'])
+                desc.getKeyDigest())
 
     def lock(self, blocking=1):
         return self._lock.acquire(blocking)
@@ -539,7 +572,7 @@ class ServerKeyset:
        to store the keys' lifetimes.
 
        When we create a new ServerKeyset object, the associated keys are not
-       read from disk unil the object's load method is called."""
+       read from disk until the object's load method is called."""
     ## Fields:
     # keydir: Directory to store this keyset's data.
     # hashlogFile: filename of this keyset's hashlog.
@@ -580,6 +613,7 @@ class ServerKeyset:
         files = [self.packetKeyFile,
                  self.mmtpKeyFile,
                  self.certFile,
+                 self.certFile+"_tmp",
                  self.descFile,
                  self.publishedFile,
                  self.hashlogFile ]
@@ -588,11 +622,15 @@ class ServerKeyset:
         mixminion.server.HashLog.deleteHashLog(self.hashlogFile)
         os.rmdir(self.keydir)
 
+    def checkKeys(self):
+        """Check whether all the required keys exist and are private."""
+        checkPrivateFile(self.packetKeyFile)
+        checkPrivateFile(self.mmtpKeyFile)
+
     def load(self, password=None):
         """Read the short-term keys from disk.  Must be called before
            getPacketKey or getMMTPKey."""
-        checkPrivateFile(self.packetKeyFile)
-        checkPrivateFile(self.mmtpKeyFile)
+        self.checkKeys()
         self.packetKey = mixminion.Crypto.pk_PEM_load(self.packetKeyFile,
                                                       password)
         self.mmtpKey = mixminion.Crypto.pk_PEM_load(self.mmtpKeyFile,
@@ -603,6 +641,11 @@ class ServerKeyset:
                                      password)
         mixminion.Crypto.pk_PEM_save(self.mmtpKey, self.mmtpKeyFile,
                                      password)
+
+    def clear(self):
+        """Stop holding the keys in memory."""
+        self.packetKey = self.mmtpKey = None
+
     def getCertFileName(self): return self.certFile
     def getHashLogFileName(self): return self.hashlogFile
     def getDescriptorFileName(self): return self.descFile
@@ -672,17 +715,22 @@ class ServerKeyset:
         fname = self.getDescriptorFileName()
         descriptor = readFile(fname)
         fields = urllib.urlencode({"desc" : descriptor})
+        f = None
         try:
             try:
                 f = urllib2.urlopen(url, fields)
                 info = f.info()
                 reply = f.read()
+            except IOError, e:
+                LOG.error("Error while publishing server descriptor: %s",e)
+                return 'error'
             except:
                 LOG.error_exc(sys.exc_info(),
                               "Error publishing server descriptor")
                 return 'error'
         finally:
-            f.close()
+            if f is not None:
+                f.close()
 
         if info.get('Content-Type') != 'text/plain':
             LOG.error("Bad content type %s from directory"%info.get(
@@ -735,6 +783,7 @@ def checkDescriptorConsistency(info, config, log=1, isPublished=1):
 
        If 'log' is true, warn as well.  Does not check keys.
     """
+    #XXXX This needs unit tests.  For now, though, it seems to work.
     warn = _WarnWrapper(silence = not log, isPublished=isPublished)
 
     config_s = config['Server']
@@ -769,13 +818,25 @@ def checkDescriptorConsistency(info, config, log=1, isPublished=1):
         warn("(Future keys will be generated with the correct lifetime")
         warn.errors -= 2 # We can't do anything about this!
 
+    insecurities = config.getInsecurities()
+    if insecurities:
+        if (info_s['Secure-Configuration'] or
+            info_s.get('Why-Insecure',None)!=", ".join(insecurities)):
+            warn("Mismatched Secure-Configuration: %r %r %r",
+                 info_s['Secure-Configuration'],
+                 info_s.get("Why-Insecure",None),
+                 ", ".join(insecurities))
+    else:
+        if not info_s['Secure-Configuration'] or info_s.get('Why-Insecure'):
+            warn("Mismatched Secure-Configuration")
+
     info_im = info['Incoming/MMTP']
     config_im = config['Incoming/MMTP']
     if info_im['Port'] != config_im['Port']:
         warn("Mismatched ports: %s configured; %s published.",
              config_im['Port'], info_im['Port'])
 
-    info_ip = info['Incoming/MMTP']['IP']
+    info_ip = info_im.get('IP',None)
     if config_im['IP'] == '0.0.0.0':
         guessed = _guessLocalIP()
         if guessed != info_ip:
@@ -785,6 +846,17 @@ def checkDescriptorConsistency(info, config, log=1, isPublished=1):
         warn("Mismatched IPs: %s configured; %s published.",
              config_im['IP'], info_ip)
 
+    info_host = info_im.get('Hostname',None)
+    config_host = config_im['Hostname']
+    if config_host is None:
+        guessed = socket.getfqdn()
+        if guessed != info_host:
+            warn("Mismatched hostnames: %s guessed; %s published",
+                 guessed, info_host)
+    elif config_host != info_host:
+        warn("Mismatched hostnames: %s configured, %s published",
+             config_host, info_host)
+
     if config_im['Enabled'] and not info_im.get('Version'):
         warn("Incoming MMTP enabled but not published.")
     elif not config_im['Enabled'] and info_im.get('Version'):
@@ -800,6 +872,12 @@ def checkDescriptorConsistency(info, config, log=1, isPublished=1):
         if config_out and not info_out:
             warn("%s enabled, but not published.", section)
 
+    info_testing = info.get("Testing",{})
+    if info_testing.get("Platform", "") != getPlatformSummary():
+        warn("Mismatched platform summary")
+    if not warn.errors and info_testing.get("Configuration", "") != config.getConfigurationSummary():
+        warn("Mismatched configuration summary")
+
     if warn.errors:
         return "bad"
     elif warn.called:
@@ -861,13 +939,14 @@ def generateServerDescriptorAndKeys(config, identityKey, keydir, keyname,
     if not validAt:
         validAt = now
 
-    if config.getInsecurities():
+    insecurities = config.getInsecurities()
+    if insecurities:
         secure = "no"
     else:
         secure = "yes"
 
     # Calculate descriptor and X509 certificate lifetimes.
-    # (Round validAt to previous mignight.)
+    # (Round validAt to previous midnight.)
     validAt = mixminion.Common.previousMidnight(validAt+30)
     if not validUntil:
         keyLifetime = config['Server']['PublicKeyLifetime'].getSeconds()
@@ -889,12 +968,14 @@ def generateServerDescriptorAndKeys(config, identityKey, keydir, keyname,
     mmtpProtocolsIn = ",".join(mmtpProtocolsIn)
     mmtpProtocolsOut = ",".join(mmtpProtocolsOut)
 
+    #XXXX007 remove
     identityKeyID = formatBase64(
                       mixminion.Crypto.sha1(
                           mixminion.Crypto.pk_encode_public_key(identityKey)))
 
     fields = {
         "IP": config['Incoming/MMTP'].get('IP', "0.0.0.0"),
+        "Hostname": config['Incoming/MMTP'].get('Hostname', None),
         "Port": config['Incoming/MMTP'].get('Port', 0),
         "Nickname": nickname,
         "Identity":
@@ -914,13 +995,28 @@ def generateServerDescriptorAndKeys(config, identityKey, keydir, keyname,
         }
 
     # If we don't know our IP address, try to guess
-    if fields['IP'] == '0.0.0.0':
+    if fields['IP'] == '0.0.0.0': #XXXX007 remove
         try:
             fields['IP'] = _guessLocalIP()
             LOG.warn("No IP configured; guessing %s",fields['IP'])
         except IPGuessError, e:
             LOG.error("Can't guess IP: %s", str(e))
-            raise MixError("Can't guess IP: %s" % str(e))
+            raise UIError("Can't guess IP: %s" % str(e))
+    # If we don't know our Hostname, try to guess
+    if fields['Hostname'] is None:
+        fields['Hostname'] = socket.getfqdn()
+        LOG.warn("No Hostname configured; guessing %s",fields['Hostname'])
+    _checkHostnameIsLocal(fields['Hostname'])
+    try:
+        dnsResults = mixminion.NetUtils.getIPs(fields['Hostname'])
+    except socket.error, e:
+        LOG.warn("Can't resolve configured hostname %r: %s",
+                 fields['Hostname'],str(e))
+    else:
+        found = [ ip for _,ip,_ in dnsResults ]
+        if fields['IP'] not in found:
+            LOG.warn("Configured hostname %r resolves to %s, but we're publishing the IP %s",
+                     fields['Hostname'], englishSequence(found), fields['IP'])
 
     # Fill in a stock server descriptor.  Note the empty Digest: and
     # Signature: lines.
@@ -935,10 +1031,12 @@ def generateServerDescriptorAndKeys(config, identityKey, keydir, keyname,
         Valid-After: %(ValidAfter)s
         Valid-Until: %(ValidUntil)s
         Packet-Key: %(PacketKey)s
-        Packet-Formats: %(PacketFormat)s
+        Packet-Versions: %(PacketFormat)s
         Software: Mixminion %(mm_version)s
         Secure-Configuration: %(Secure)s
         """ % fields
+    if insecurities:
+        info += "Why-Insecure: %s\n"%(", ".join(insecurities))
     if contact:
         info += "Contact: %s\n"%contact
     if comments:
@@ -950,6 +1048,7 @@ def generateServerDescriptorAndKeys(config, identityKey, keydir, keyname,
             [Incoming/MMTP]
             Version: 0.1
             IP: %(IP)s
+            Hostname: %(Hostname)s
             Port: %(Port)s
             Key-Digest: %(KeyID)s
             Protocols: %(MMTPProtocolsIn)s
@@ -977,6 +1076,13 @@ def generateServerDescriptorAndKeys(config, identityKey, keydir, keyname,
     # Ask our modules for their configuration information.
     info += "".join(config.moduleManager.getServerInfoBlocks())
 
+    info += """\
+          [Testing]
+          Platform: %s
+          Configuration: %s
+          """ %(getPlatformSummary(),
+                config.getConfigurationSummary())
+
     # Remove extra (leading or trailing) whitespace from the lines.
     lines = [ line.strip() for line in info.split("\n") ]
     # Remove empty lines
@@ -994,6 +1100,11 @@ def generateServerDescriptorAndKeys(config, identityKey, keydir, keyname,
     # FFFF Remove this once we're more confident.
     inf = ServerInfo(string=info)
     ok = checkDescriptorConsistency(inf, config, log=0, isPublished=0)
+    if ok not in ('good', 'so-so'):
+        print "========"
+        print info
+        print "======"
+        checkDescriptorConsistency(inf, config, log=1, isPublished=0)
     assert ok in ('good', 'so-so')
 
     return info
@@ -1079,6 +1190,28 @@ def _guessLocalIP():
 
     return IP
 
+_KNOWN_LOCAL_HOSTNAMES = {}
+
+def _checkHostnameIsLocal(name):
+    if _KNOWN_LOCAL_HOSTNAMES.has_key(name):
+        return
+    try:
+        r = mixminion.NetUtils.getIPs(name)
+        for family, addr, _ in r:
+            if family == mixminion.NetUtils.AF_INET:
+                if addr.startswith("127.") or addr.startswith("0."):
+                    LOG.warn("Hostname %r resolves to reserved address %s",
+                             name, addr)
+            else:
+                if addr in ("::", "::1"):
+                    LOG.warn("Hostname %r resolves to reserved address %s",
+                             name,addr)
+    except socket.error, e:
+        # ???? Turn this into a warning, if people have a real reason to
+        # ???? use a hostname that they themselves cannot resolve.
+        raise UIError("Cannot resolve hostname %r: %s"%(name,e))
+    _KNOWN_LOCAL_HOSTNAMES[name] = 1
+
 def generateCertChain(filename, mmtpKey, identityKey, nickname,
                       certStarts, certEnds):
     """Create a two-certificate chain for use in MMTP.
@@ -1106,3 +1239,13 @@ def generateCertChain(filename, mmtpKey, identityKey, nickname,
     identityCertText = readFile(fname)
     os.unlink(fname)
     writeFile(filename, certText+identityCertText, 0600)
+
+def getPlatformSummary():
+    """Return a string describing the current software and platform."""
+    if hasattr(os, "uname"):
+        uname = " ".join(os.uname())
+    else:
+        uname = sys.platform
+
+    return "Mixminion %s; Python %r on %r" % (
+        mixminion.__version__, sys.version, uname)
