@@ -1,5 +1,5 @@
-/* Copyright (c) 2002 Nick Mathewson.  See LICENSE for licensing information */
-/* $Id: _minionlib.h,v 1.14 2003/02/12 01:23:24 nickm Exp $ */
+/* Copyright (c) 2002,2003 Nick Mathewson.  See LICENSE for licensing information */
+/* $Id: _minionlib.h,v 1.20 2003/10/09 03:56:30 nickm Exp $ */
 #ifndef _MINIONLIB_H
 #define _MINIONLIB_H
 
@@ -19,6 +19,12 @@ the README file for further instructions."
 #include <rsa.h>
 #endif
 
+#ifdef _MSC_VER
+#define INLINE __inline
+#else
+#define INLINE inline
+#endif
+
 /* We provide our own implementation of counter mode; see aes_ctr.c
  */
 void mm_aes_counter128(const char *in, char *out, unsigned int len,
@@ -38,6 +44,7 @@ typedef struct mm_RSA {
 
 extern PyTypeObject mm_TLSContext_Type;
 extern PyTypeObject mm_TLSSock_Type;
+extern PyTypeObject mm_FEC_Type;
 
 /**
  * Macros to declare function prototypes with the proper signatures for Python.
@@ -53,14 +60,19 @@ extern PyTypeObject mm_TLSSock_Type;
                         METH_VARARGS|METH_KEYWORDS,       \
                         (char*)obj##_##name##__doc__ }
 
-
 /* Functions from crypt.c */
 FUNC_DOC(mm_sha1);
 FUNC_DOC(mm_sha1);
 FUNC_DOC(mm_aes_key);
 FUNC_DOC(mm_aes_ctr128_crypt);
+FUNC_DOC(mm_aes128_block_crypt);
 FUNC_DOC(mm_strxor);
 FUNC_DOC(mm_openssl_seed);
+#ifdef MS_WINDOWS
+FUNC_DOC(mm_win32_openssl_seed);
+FUNC_DOC(mm_win32_get_random_bytes);
+#endif
+FUNC_DOC(mm_openssl_rand);
 FUNC_DOC(mm_add_oaep_padding);
 FUNC_DOC(mm_check_oaep_padding);
 FUNC_DOC(mm_rsa_generate);
@@ -74,6 +86,11 @@ FUNC_DOC(mm_generate_cert);
 extern PyObject *mm_CryptoError;
 extern char mm_CryptoError__doc__[];
 
+/* From fec.c */
+FUNC_DOC(mm_FEC_generate);
+extern PyTypeObject mm_FEC_Type;
+extern PyObject *mm_FECError;
+extern char mm_FECError__doc__[];
 
 /* From tls.c */
 extern PyTypeObject mm_TLSSock_Type;
