@@ -1,7 +1,17 @@
-/* Copyright (c) 2002 Nick Mathewson.  See LICENSE for licensing information */
-/* $Id: crypt.c,v 1.24 2003/06/05 18:41:41 nickm Exp $ */
+/* Copyright (c) 2002-2003 Nick Mathewson.  See LICENSE for licensing information */
+/* $Id: crypt.c,v 1.35 2003/11/28 04:14:05 nickm Exp $ */
 #include <Python.h>
 
+#ifdef MS_WINDOWS
+#define WIN32_WINNT 0x0400
+#define _WIN32_WINNT 0x0400
+#include <windows.h>
+#include <wincrypt.h>
+#ifndef ALG_CLASS_ANY
+#error no good
+#endif
+#endif
+
 #include <time.h>
 
 #ifndef TRUNCATED_OPENSSL_INCLUDES
@@ -23,11 +33,15 @@
 #include <pem.h>
 #include <evp.h>
 #endif
-#include <_minionlib.h>
+
+#include "_minionlib.h"
 #include <assert.h>
 
 #define TYPE_ERR(s) PyErr_SetString(PyExc_TypeError, s)
 #define KEY_IS_PRIVATE(rsa) ((rsa)->p)
+#ifdef MS_WINDOWS
+#define WIN_ERR(s) PyErr_SetString(PyExc_WindowsError, s)
+#endif
 
 char mm_CryptoError__doc__[] =
   "mixminion._minionlib.SSLError\n\n"
@@ -214,6 +228,45 @@ mm_aes_ctr128_crypt(PyObject *self, PyObject *args, PyObject *kwdict)
         return output;
 }
 
+const char mm_aes128_block_crypt__doc__[] =
+"aes128_block_crypt(key, block, encrypt=0) -> result\n\n"
+"For testing only.  Encrypt or decrypt a single RSA block.\n";
+
+PyObject*
+mm_aes128_block_crypt(PyObject *self, PyObject *args, PyObject *kwdict)
+{
+        static char *kwlist[] = { "key", "block", "encrypt", NULL };
+        char *input;
+        long inputlen;
+        int encrypt=0;
+        PyObject *result;
+        AES_KEY *aes_key = NULL;
+
+        if (!PyArg_ParseTupleAndKeywords(args, kwdict,
+                                         "O&s#|i:aes128_block_crypt", kwlist,
+                                         aes_arg_convert, &aes_key,
+                                         &input, &inputlen,
+                                         &encrypt))
+                return NULL;
+
+        if (inputlen != 16) {
+                TYPE_ERR("aes128_block_crypt expected a single block.");
+                return NULL;
+        }
+
+        if (!(result = PyString_FromStringAndSize(NULL, 16))) {
+                PyErr_NoMemory();
+                return NULL;
+        }
+        if (encrypt) {
+                AES_encrypt(input, PyString_AS_STRING(result), aes_key);
+        } else {
+                AES_decrypt(input, PyString_AS_STRING(result), aes_key);
+        }
+
+        return result;
+}
+
 const char mm_strxor__doc__[]=
   "strxor(str1, str2) -> str\n\n"
   "Computes the bitwise xor of two equally-long strings.  Throws TypeError\n"
@@ -278,6 +331,150 @@ mm_openssl_seed(PyObject *self, PyObject *args, PyObject *kwdict)
         return Py_None;
 }
 
+#ifdef MS_WINDOWS
+const char mm_win32_openssl_seed__doc__[]=
+  "openssl_seed_win32()\n\n"
+  "Windows-only function: seed OpenSSL's random number generator baswed on\n"
+  "the current contents of the screen.\n";
+
+PyObject *
+mm_win32_openssl_seed(PyObject *self, PyObject *args, PyObject *kwdict)
+{
+        static char *kwlist[] = { NULL };
+
+        if (!PyArg_ParseTupleAndKeywords(args, kwdict, ":win32_openssl_seed",
+                                         kwlist))
+                return NULL;
+
+        Py_BEGIN_ALLOW_THREADS
+        RAND_screen();
+        Py_END_ALLOW_THREADS
+
+        Py_INCREF(Py_None);
+        return Py_None;
+}
+
+/* Flag: set to true if the variable 'provider' has been set. */
+static int provider_set = 0;
+/* A global handle to a crypto context. */
+static HCRYPTPROV provider;
+
+/* Helper method: return a handle for a Windows Crypto API crypto provider,
+ * initializing it if necessary.  Raise an exception and return 0 on failure.
+ */
+static HCRYPTPROV getProvider()
+{
+        if (provider_set)
+                return provider;
+
+        if (!CryptAcquireContext(&provider,
+                                 NULL,
+                                 NULL,
+                                 PROV_RSA_FULL,
+                                 0)) {
+                if (GetLastError() != NTE_BAD_KEYSET) {
+                        WIN_ERR("Can't get CryptoAPI provider [1]");
+                        return 0;
+                }
+                if (!CryptAcquireContext(&provider, NULL, NULL, PROV_RSA_FULL,
+                                        CRYPT_NEWKEYSET)) {
+                        WIN_ERR("Can't get CryptoAPI provider [2]");
+                        return 0;
+                }
+        }
+
+        provider_set =  1;
+
+        return provider;
+}
+
+const char mm_win32_get_random_bytes__doc__[]=
+  "win32_get_random_bytes(n)\n\n"
+  "Return n random bytes, using Windows's internal supposedly secure random\n"
+  "number generator.  \n";
+
+PyObject *
+mm_win32_get_random_bytes(PyObject *self, PyObject *args, PyObject *kwdict)
+{
+
+        static char *kwlist[] = { "n", NULL };
+        PyObject *result;
+        int n,r;
+        HCRYPTPROV p;
+
+        if (!PyArg_ParseTupleAndKeywords(args, kwdict,
+                                         "i:win32_get_random_bytes",
+                                         kwlist, &n))
+                return NULL;
+
+        if (n<0) {
+                TYPE_ERR("n must be >= 0");
+                return NULL;
+        }
+
+        p = getProvider();
+        if (!provider_set) {
+                /* Exception is already set. */
+                return NULL;
+        }
+
+
+        if (!(result = PyString_FromStringAndSize(NULL, n))) {
+                PyErr_NoMemory(); return NULL;
+        }
+
+        Py_BEGIN_ALLOW_THREADS
+        r = CryptGenRandom(getProvider(), n, PyString_AS_STRING(result));
+        Py_END_ALLOW_THREADS
+
+        if (!r) {
+                Py_DECREF(result);
+                WIN_ERR("Error generating random bytes");
+                return NULL;
+        }
+
+        return result;
+}
+#endif
+
+const char mm_openssl_rand__doc__[]=
+  "openssl_rand(n)\n\n"
+  "Return n random bytes from OpenSSL's random number generator.\n";
+
+PyObject *
+mm_openssl_rand(PyObject *self, PyObject *args, PyObject *kwdict)
+{
+        static char *kwlist[] = { "n", NULL };
+        int bytes;
+        PyObject *result;
+        int r;
+
+        if (!PyArg_ParseTupleAndKeywords(args, kwdict, "i:openssl_rand",
+                                         kwlist, &bytes))
+                return NULL;
+
+        if (bytes < 0) {
+                TYPE_ERR("n must be >= 0");
+        }
+
+        if (!(result = PyString_FromStringAndSize(NULL,bytes))) {
+                PyErr_NoMemory(); return NULL;
+        }
+
+        Py_BEGIN_ALLOW_THREADS
+        r = RAND_bytes(PyString_AsString(result), bytes);
+        Py_END_ALLOW_THREADS
+
+        if (!r) {
+                Py_DECREF(result);
+                mm_SSL_ERR(1);
+                return NULL;
+        }
+
+        return result;
+}
+
+
 
 static void
 mm_RSA_dealloc(mm_RSA *self)
@@ -693,6 +890,58 @@ mm_RSA_get_public_key(PyObject *self, PyObject *args, PyObject *kwdict)
         return output;
 }
 
+const char mm_RSA_get_private_key__doc__[]=
+"rsa.get_private_key() -> (n,e,d,p,q)\n";
+
+PyObject *
+mm_RSA_get_private_key(PyObject *self, PyObject *args, PyObject *kwdict)
+{
+        /* ???? should be threadified? */
+        static char *kwlist[] = {  NULL };
+
+        RSA *rsa;
+        PyObject *n, *e, *d, *p, *q;
+        PyObject *output;
+
+        n = e = d = p = q = NULL;
+
+        assert(mm_RSA_Check(self));
+        if (!PyArg_ParseTupleAndKeywords(args, kwdict,
+                                         ":get_private_key", kwlist))
+                return NULL;
+
+        rsa = ((mm_RSA*)self)->rsa;
+        if (!rsa->n) { TYPE_ERR("Key has no modulus"); return NULL;}
+        if (!rsa->e) { TYPE_ERR("Key has no e"); return NULL; }
+        if (!rsa->d) { TYPE_ERR("Key has no d"); return NULL; }
+        if (!rsa->p) { TYPE_ERR("Key has no p"); return NULL; }
+        if (!rsa->q) { TYPE_ERR("Key has no q"); return NULL; }
+        output = NULL;
+        if (!(n = bn2pylong(rsa->n))) {
+                PyErr_NoMemory(); goto done;
+        }
+        if (!(e = bn2pylong(rsa->e))) {
+                PyErr_NoMemory(); goto done;
+        }
+        if (!(d = bn2pylong(rsa->d))) {
+                PyErr_NoMemory(); goto done;
+        }
+        if (!(p = bn2pylong(rsa->p))) {
+                PyErr_NoMemory(); goto done;
+        }
+        if (!(q = bn2pylong(rsa->q))) {
+                PyErr_NoMemory(); goto done;
+        }
+        output = Py_BuildValue("OOOOO", n, e, d, p, q);
+ done:
+        if (n) { Py_DECREF(n); }
+        if (e) { Py_DECREF(e); }
+        if (d) { Py_DECREF(d); }
+        if (p) { Py_DECREF(p); }
+        if (q) { Py_DECREF(q); }
+        return output;
+}
+
 const char mm_RSA_get_exponent__doc__[]=
    "rsa.get_exponent() -> e\n";
 
@@ -775,6 +1024,7 @@ static PyMethodDef mm_RSA_methods[] = {
         METHOD(mm_RSA, get_public_key),
         METHOD(mm_RSA, get_exponent),
         METHOD(mm_RSA, PEM_write_key),
+        METHOD(mm_RSA, get_private_key),
         { NULL, NULL }
 };
 
@@ -995,7 +1245,7 @@ mm_generate_cert(PyObject *self, PyObject *args, PyObject *kwargs)
         X509_NAME *name_issuer = NULL;
         int nid;
         PyObject *retval;
-        time_t time;
+        time_t _time;
 
         if (!PyArg_ParseTupleAndKeywords(args, kwargs,
                                          "sO!O!ssdd:generate_cert",
@@ -1027,7 +1277,10 @@ mm_generate_cert(PyObject *self, PyObject *args, PyObject *kwargs)
                 goto error;
         if (!(X509_set_version(x509, 2)))
                 goto error;
-        if (!(ASN1_INTEGER_set(X509_get_serialNumber(x509),0L)))
+
+        /* Seconds since jan 1 2003. */
+        _time = time(NULL) - 1041397200;
+        if (!(ASN1_INTEGER_set(X509_get_serialNumber(x509),(long)_time)))
                 goto error;
 
 #define SET_PART(n, part, val)                                   \
@@ -1050,11 +1303,11 @@ mm_generate_cert(PyObject *self, PyObject *args, PyObject *kwargs)
         if (!(X509_set_subject_name(x509, name)))
                 goto error;
 
-        time = (time_t) start_time;
-        if (!X509_time_adj(X509_get_notBefore(x509),0,&time))
+        _time = (time_t) start_time;
+        if (!X509_time_adj(X509_get_notBefore(x509),0,&_time))
                 goto error;
-        time = (time_t) end_time;
-        if (!X509_time_adj(X509_get_notAfter(x509),0,&time))
+        _time = (time_t) end_time;
+        if (!X509_time_adj(X509_get_notAfter(x509),0,&_time))
                 goto error;
         if (!(X509_set_pubkey(x509, pkey)))
                 goto error;
