@@ -1,5 +1,5 @@
 /* Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information */
-/* $Id: main.c,v 1.14 2003/05/17 00:08:45 nickm Exp $ */
+/* $Id: main.c,v 1.20 2003/10/02 21:46:23 nickm Exp $ */
 
 /*
   If you're not familiar with writing Python extensions, you should
@@ -7,7 +7,7 @@
   "http://www.python.org/doc/current/ext/ext.html".
 */
 
-#include <_minionlib.h>
+#include "_minionlib.h"
 
 #ifndef TRUNCATED_OPENSSL_INCLUDES
 #include <openssl/ssl.h>
@@ -29,8 +29,14 @@ static struct PyMethodDef _mixcryptlib_functions[] = {
         ENTRY(sha1),
         ENTRY(aes_key),
         ENTRY(aes_ctr128_crypt),
+        ENTRY(aes128_block_crypt),
         ENTRY(strxor),
         ENTRY(openssl_seed),
+        ENTRY(openssl_rand),
+#ifdef MS_WINDOWS
+        ENTRY(win32_openssl_seed),
+        ENTRY(win32_get_random_bytes),
+#endif
         ENTRY(add_oaep_padding),
         ENTRY(check_oaep_padding),
         ENTRY(rsa_generate),
@@ -41,6 +47,8 @@ static struct PyMethodDef _mixcryptlib_functions[] = {
         ENTRY(generate_cert),
 
         ENTRY(TLSContext_new),
+
+        ENTRY(FEC_generate),
         { NULL, NULL }
 };
 
@@ -51,7 +59,7 @@ static struct PyMethodDef _mixcryptlib_functions[] = {
    exception: Set to point to a pointer to the newly allocated exception.
    longName: The fully qualified name of this exception.
    itemString: The name of this exception within the module.
-   doc: The doctring for this exception.
+   doc: The docstring for this exception.
 
    returns 1 on failure; 0 on success */
 static int
@@ -116,10 +124,13 @@ init_minionlib(void)
         if (exc(d, &mm_TLSClosed, "mixminion._minionlib.TLSClosed",
                 "TLSClosed", mm_TLSClosed__doc__))
                 return;
+        if (exc(d, &mm_FECError, "mixminion._minionlib.FECError",
+                "FECError", mm_FECError__doc__))
+                return;
 
-        /* We set ob_type here so that Cygwin is happy. */
+        /* We set ob_type here so that Cygwin and Win32 are happy. */
         mm_RSA_Type.ob_type = mm_TLSContext_Type.ob_type =
-                mm_TLSSock_Type.ob_type = &PyType_Type;
+                mm_TLSSock_Type.ob_type = mm_FEC_Type.ob_type = &PyType_Type;
 
         Py_INCREF(&mm_RSA_Type);
         if (PyDict_SetItemString(d, "RSA", (PyObject*)&mm_RSA_Type) < 0)
@@ -134,6 +145,11 @@ init_minionlib(void)
         if (PyDict_SetItemString(d, "TLSSock",
                                  (PyObject*)&mm_TLSSock_Type) < 0)
                 return;
+
+        Py_INCREF(&mm_FEC_Type);
+        if (PyDict_SetItemString(d, "FEC",
+                                 (PyObject*)&mm_FEC_Type) < 0)
+                return;
 }
 
 /*
