@@ -1,5 +1,5 @@
-/* Copyright (c) 2002 Nick Mathewson.  See LICENSE for licensing information */
-/* $Id: tls.c,v 1.25 2003/06/05 18:41:41 nickm Exp $ */
+/* Copyright (c) 2002-2003 Nick Mathewson.  See LICENSE for licensing information */
+/* $Id: tls.c,v 1.31 2003/11/28 04:14:05 nickm Exp $ */
 #include "_minionlib.h"
 
 #include <time.h>
@@ -111,8 +111,7 @@ const char mm_TLSContext_new__doc__[] =
    "Allocates a new TLSContext object.  The files, if provided, are used\n"
    "contain the PEM-encoded X509 public keys, private key, and DH\n"
    "parameters for this context.\n\n"
-   "If a cert is provided, assume we're working in server mode, and allow\n"
-   "BUG:In the future, certs, pks, and dh parameters will be first-class.\n\n"
+   "If a cert is provided, assume we're working in server mode, and allow\n\n"
    "LIMITATION: We don\'t expose any more features than Mixminion needs.\n";
 
 PyObject*
@@ -123,7 +122,7 @@ mm_TLSContext_new(PyObject *self, PyObject *args, PyObject *kwargs)
         mm_RSA *rsa = NULL;
         int err = 0;
 
-        SSL_METHOD *method;
+        SSL_METHOD *method = NULL;
         SSL_CTX *ctx = NULL;
         DH *dh = NULL;
         BIO *bio = NULL;
@@ -139,10 +138,21 @@ mm_TLSContext_new(PyObject *self, PyObject *args, PyObject *kwargs)
                 return NULL;
 
         Py_BEGIN_ALLOW_THREADS;
-        method = TLSv1_method();
 
+        if (certfile) {
+                /* Accept SSL2 and SSL3 and TLS1. */
+                method = SSLv23_method();
+        } else {
+                /* Generate only TLS1. */
+                method = TLSv1_method();
+        }
+        /* Allow SSL2 and SSL3 and TLS1 */
         if (!(ctx = SSL_CTX_new(method)))
                 err = 1;
+        /* But not actually SSL2. */
+        if (certfile) {
+                SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);
+        }
         if (!err && !SSL_CTX_set_cipher_list(ctx,
                                        TLS1_TXT_DHE_RSA_WITH_AES_128_SHA))
                 err = 1;
@@ -355,7 +365,7 @@ static PyObject*
 mm_TLSSock_connect(PyObject *self, PyObject *args, PyObject *kwargs)
 {
         SSL *ssl;
-        int r, err;
+        int r;
 
         assert(mm_TLSSock_Check(self));
         FAIL_IF_ARGS();
@@ -366,7 +376,6 @@ mm_TLSSock_connect(PyObject *self, PyObject *args, PyObject *kwargs)
         r = SSL_connect(ssl);
         Py_END_ALLOW_THREADS
         if (r <= 0) {
-                err = SSL_get_error(ssl,r);
                 tls_error(ssl, r, 0);
                 return NULL;
         }
@@ -593,9 +602,9 @@ mm_TLSSock_check_cert_alive(PyObject *self, PyObject *args, PyObject *kwargs)
                 MM_TLS_ERR("Certificate has expired");
                 goto error;
         }
-        
+
         X509_free(cert);
-        
+
         Py_INCREF(Py_None);
         return Py_None;
  error:
@@ -678,7 +687,7 @@ mm_TLSSock_verify_cert_and_get_identity_pk(
 }
 
 static char mm_TLSSock_renegotiate__doc__[] =
-    "tlssock.renegotate()\n\n"
+    "tlssock.renegotiate()\n\n"
     "Mark this connection as requiring renegotiation.  No renegotiation is\n"
     "performed until do_handshake is called.  Note that renegotiation only\n"
     "works intuitively on the client side.\n";
